// Copyright 2012 MightyClient Authors. All Rights Reserved.
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//      http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS-IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

/**
 * @fileoverview Defines an abstract class for two-way communication channel
 * between the client and the server.
 */


goog.require('goog.asserts');
goog.require('goog.debug.Logger');
goog.require('goog.events.EventTarget');
goog.require('goog.string');

goog.provide('mc.Channel');



/**
 * An abstract class for a two-way communication channel.
 * @constructor
 * @extends {goog.events.EventTarget}
 */
mc.Channel = function() {
  goog.base(this);


  /**
   * Map of service names to callbacks pre-bound to context objects.
   * @type {!Object.<!function(*)>}
   * @private
   */
  this.services_ = {};
};
goog.inherits(mc.Channel, goog.events.EventTarget);


/**
 * Event types dispatched by mc.Channel implementations.
 * @enum {string}
 */
mc.Channel.EventType = {
  // Dispatched when the channel is opened.
  OPEN: 'open',

  // Dispatched when the channel is closed (it can be reopened later, for
  // example when the user regains internet connection).
  CLOSE: 'close'
};


/**
 * Logger.
 * @type {!goog.debug.Logger}
 * @protected
 */
mc.Channel.prototype.logger = goog.debug.Logger.getLogger('mc.Channel');


/**
 * Registers a service and the associated handler of incoming messages.
 * @param {string} service Service name, any unique string.
 * @param {!function(*)} messageHandler Handler for incoming messages that takes
 *     the non-serialized payload as a single argument.
 * @param {Object=} opt_obj Optional scope in which to call the handler.
 */
mc.Channel.prototype.registerService = function(service, messageHandler,
    opt_obj) {
  goog.asserts.assert(!(service in this.services_),
      'Service \'%s\' already registered.', service);
  this.services_[service] = goog.bind(messageHandler, opt_obj || null);
};


/**
 * Unregisters a handler of incoming messages for a service.
 * @param {string} service Service name.
 */
mc.Channel.prototype.unregisterService = function(service) {
  goog.asserts.assert(service in this.services_,
      'Service \'%s\' not registered.', service);
  delete this.services_[service];
};


/**
 * Should be called by subclasses when an incoming message is received. Calls
 * the handler of incoming messages that was registered for supplied service or
 * throws an assertion error if the service is not registered.
 * @param {string} service Service name.
 * @param {*} payload Non-serialized payload.
 * @protected
 */
mc.Channel.prototype.handleIncomingMessage = function(service, payload) {
  if (window.JSON && goog.DEBUG) {
    this.logger.info(goog.string.subs(
        'Received \'%s\' through service \'%s\'.',
        goog.string.truncate(window.JSON.stringify(payload), 200), service));
  } else {
    this.logger.info(goog.string.subs(
        'Received a message through service \'%s\'.', service));
  }
  if (service in this.services_) {
    this.services_[service].call(null, payload);
  } else {
    this.logger.warning(goog.string.subs(
        'Received a message for service \'%s\' which is not registered.',
        service));
  }
};


/**
 * Sends a message over the channel, should be overridden by subclasses.
 * @param {string} service Service name.
 * @param {*} payload Non-serialized payload.
 * @protected
 */
mc.Channel.prototype.sendInternal = goog.abstractMethod;


/**
 * Sends a message over the channel.
 * @param {string} service Service name.
 * @param {*} payload Non-serialized payload.
 */
mc.Channel.prototype.send = function(service, payload) {
  if (window.JSON && goog.DEBUG) {
    this.logger.info(goog.string.subs(
        'Will send \'%s\' through service \'%s\'.',
        goog.string.truncate(window.JSON.stringify(payload), 200), service));
  } else {
    this.logger.info(goog.string.subs(
        'Will send a message through service \'%s\'.', service));
  }
  this.sendInternal(service, payload);
};


/**
 * @return {boolean} Whether the channel is currently open.
 */
mc.Channel.prototype.isOpen = goog.abstractMethod;
