// Copyright 2012 MightyClient Authors. All Rights Reserved.
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//      http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS-IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

/**
 * @fileoverview Defines an API to read, write, and listen for changes in a
 * collection of JSON-serializable values where each value is located at a path
 * composed of string keys. Exposes hooks for data synchronization. Optionally,
 * a schema can be defined that will be enforced by run-time assertions in debug
 * mode and stripped by compiler in production.
 */


goog.require('goog.Disposable');
goog.require('goog.array');
goog.require('goog.asserts');
goog.require('goog.events.EventTarget');
goog.require('goog.math');
goog.require('goog.object');
goog.require('goog.string');

goog.provide('mc.ds');


/**
 * @define {boolean} Whether to strip out schema definitions and runtime checks
 *     based on them.
 */
mc.ds.ENABLE_SCHEMA = goog.asserts.ENABLE_ASSERTS;


/**
 * An array of string keys.
 * @typedef {!Array.<string>}
 */
mc.ds.Path;


/**
 * Utility function that sets a value in a nested hierarchy of objects at
 * supplied path. If the supplied value is undefined, removes the existing value
 * and the objects that will become empty as a result, if any. If the value is
 * not undefined, if necessary creates the objects that form the required path.
 * @param {!Object} tree A tree of nested objects.
 * @param {mc.ds.Path} path An array of keys. Must be non-empty. No prefix
 *     should point to a primitive value.
 * @param {*=} opt_value Value to set.
 * @return {*} The value which was at provided path originally, or undefined if
 *     there wasn't any.
 */
mc.ds.setValueByKeys = function(tree, path, opt_value) {
  goog.asserts.assert(path.length);
  if (!goog.isDef(opt_value)) {
    var removeObj = null;
    var removeKey;
    path = goog.array.clone(path);
    while (path.length) {
      var pathEl = path.shift();
      if (!(pathEl in tree)) {
        return;
      }
      var child = tree[pathEl];
      if (path.length) {
        goog.asserts.assertObject(child,
            'No prefix of the path should point to a primitive value.');
      }
      var moreThanOneElement = function(obj) {
        var foundOne = false;
        for (var key in obj) {
          if (foundOne) {
            return true;
          }
          foundOne = true;
        }
        return false;
      }
      if (path.length && moreThanOneElement(child)) {
        removeObj = null;
      } else if (goog.isNull(removeObj)) {
        removeObj = tree;
        removeKey = pathEl;
      }
      tree = child;
    }
    goog.object.remove(removeObj, removeKey);
    return tree;
  } else {
    goog.array.forEach(goog.array.slice(path, 0, -1),
        function(pathEl) {
          tree = goog.asserts.assertObject(
              goog.object.setIfUndefined(tree, pathEl, {}),
              'No prefix of the path should point to a primitive value.');
        });
    var key = path[path.length - 1];
    var oldValue = tree[key];
    tree[key] = opt_value;
    return oldValue;
  }
};


/**
 * Utility function that sets a value in a nested hierarchy of objects at
 * supplied path if none already exists, if necessary creating the objects that
 * form the required path.
 * @param {!Object} tree A tree of nested objects.
 * @param {mc.ds.Path} path An array of keys. Must be non-empty. No prefix
 *     should point to a non-object value.
 * @param {*} value Value to set, must not be undefined.
 * @return {*} The resulting value at provided path.
 */
mc.ds.setValueByKeysIfUndefined = function(tree, path, value) {
  goog.asserts.assert(path.length);
  goog.asserts.assert(goog.isDef(value));
  goog.array.forEach(goog.array.slice(path, 0, -1),
      function(pathEl) {
        tree = goog.asserts.assertObject(
            goog.object.setIfUndefined(tree, pathEl, {}),
            'No prefix of the path should point to a non-object value.');
      });
  var key = path[path.length - 1];
  if (key in tree) {
    return tree[key];
  } else {
    return tree[key] = value;
  }
};


/**
 * Calls a function for each element in a nested hierarchy of objects that
 * matches the supplied path. If any call returns true, returns true (without
 * checking the rest). If all calls return false, returns false.
 * @param {!Object} tree A tree made of nested objects.
 * @param {mc.ds.Path} path Path over which to iterate, can contain '*' (any
 *     key) wildcards.
 * @param {function(*, mc.ds.Path): boolean} f The function to call for every
 *     matching element. This function takes 2 arguments (the element
 *     and the fully qualified path to it) and should return a boolean.
 * @param {Object=} opt_obj Optional scope within which to call f.
 * @param {!Object=} opt_leafKeys If provided, will not iterate over paths
 *     that have any of the supplied object's keys as elements. Provided path
 *     must not include any of such keys.
 * @return {boolean} true if any element passes the test.
 */
mc.ds.some = function(tree, path, f, opt_obj, opt_leafKeys) {
  var some = function(treeElement, pathLeft, pathRight) {
    if (!pathRight.length) {
      return f.call(opt_obj || null, treeElement, pathLeft);
    } else if (goog.isObject(treeElement)) {
      if (pathRight[0] == '*') {
        return goog.object.some(treeElement,
            function(childElement, childElementKey) {
              if (!opt_leafKeys || !(childElementKey in opt_leafKeys)) {
                return some(childElement,
                    goog.array.concat(pathLeft, childElementKey),
                    goog.array.slice(pathRight, 1));
              }
              return false;
            });
      } else if (pathRight[0] in treeElement) {
        if (opt_leafKeys) {
          goog.asserts.assert(!(pathRight[0] in opt_leafKeys),
              'Path must not contain any of the keys in opt_leafKeys');
        }
        return some(treeElement[pathRight[0]],
            goog.array.concat(pathLeft, pathRight[0]),
            goog.array.slice(pathRight, 1));
      }
    }
    return false;
  };
  return some(tree, [], path);
};


/**
 * Calls a function for each element in a nested hierarchy of objects that
 * matches the supplied path. If all calls return true, returns true. If any
 * call returns false, returns false (without checking the rest).
 * @param {!Object} tree A tree made of nested objects.
 * @param {mc.ds.Path} path Path over which to iterate, can contain '*' (any
 *     key) wildcards.
 * @param {function(*, mc.ds.Path): boolean} f The function to call for every
 *     matching element. This function takes 2 arguments (the element
 *     and the fully qualified path to it) and should return a boolean.
 * @param {Object=} opt_obj Optional scope within which to call f.
 * @param {!Object=} opt_leafKeys If provided, will not iterate over paths
 *     that have any of the supplied object's keys as elements. Provided path
 *     must not include any of such keys.
 * @return {boolean} false if any element fails the test.
 */
mc.ds.every = function(tree, path, f, opt_obj, opt_leafKeys) {
  return !mc.ds.some(tree, path, function(treeElement, path) {
    return !f.call(opt_obj || null, treeElement, path);
  }, null, opt_leafKeys);
};


/**
 * Utility function that iterates over each element in a nested hierarchy of
 * objects that matches the supplied path.
 * @param {!Object} tree A tree made of nested objects.
 * @param {mc.ds.Path} path Path over which to iterate, can contain '*' (any
 *     key) wildcards.
 * @param {function(*, mc.ds.Path)} f The function to call for every
 *     matching element. This function takes 2 arguments (the element
 *     and the fully qualified path to it).
 * @param {Object=} opt_obj Optional scope within which to call f.
 * @param {!Object=} opt_leafKeys If provided, will not iterate over paths
 *     that have any of the supplied object's keys as elements. Provided path
 *     must not include any of such keys.
 */
mc.ds.forEach = function(tree, path, f, opt_obj, opt_leafKeys) {
  mc.ds.some(tree, path, function(treeElement, path) {
    f.call(opt_obj || null, treeElement, path);
    return false;
  }, null, opt_leafKeys);
};


/**
 * Merges one tree of nested objects into another, throwing an assertion error
 * if in the source there is a primitive value at path which exists in the
 * destination tree.
 * @param {!Object} tree Destination tree.
 * @param {!Object} otherTree Source tree.
 */
mc.ds.extendTree = function(tree, otherTree) {
  goog.object.forEach(otherTree, function(value, key) {
    if (!(key in tree)) {
      tree[key] = value;
    } else {
      mc.ds.extendTree(goog.asserts.assertObject(tree[key],
          'Conflict when merging trees'), goog.asserts.assertObject(value,
          'Conflict when merging trees'));
    }
  });
};


/**
 * Checks whether the supplied value is a valid non-wildcard key in a data tree
 * path: a non-empty string composed of letters, numbers, or underscores.
 * @param {string} pathEl Path element to check.
 * @return {boolean} Whether the path element is valid.
 */
mc.ds.isValidPathElement = function(pathEl) {
  return /^[a-zA-Z0-9_]+$/.test(pathEl);
};


/**
 * Checks whether the supplied array of strings is a valid path.
 * @param {mc.ds.Path} path Path to check.
 * @return {?mc.ds.Path} Path supplied as argument if it is valid, or null
 *     otherwise.
 */
mc.ds.isValidPath = function(path) {
  if (goog.array.every(path, function(pathEl) {
    return pathEl === '*' || mc.ds.isValidPathElement(pathEl);
  })) {
    return path;
  }
  return null;
};


/**
 * Checks the supplied path for for presence of wildcards.
 * @param {mc.ds.Path} path Path to check.
 * @return {boolean} Whether the path is fully qualified.
 */
mc.ds.isPathFullyQualified = function(path) {
  return goog.array.every(path, function(pathEl) {
    return pathEl != '*';
  });
};


/**
 * Asserts that the supplied value is a valid non-wildcard key in a data tree
 * path.
 * @param {string} pathEl Path element to check.
 */
mc.ds.assertValidPathElement = function(pathEl) {
  goog.asserts.assertString(pathEl, 'Path element must be a string.');
  if (!mc.ds.isValidPathElement(pathEl)) {
    goog.asserts.assert(false,
        goog.string.subs('Invalid path element "%s".', pathEl));
  }
};


/**
 * Asserts that the supplied value is an array of valid data tree keys or
 * wildcards.
 * @param {*} path Path to check.
 * @return {mc.ds.Path} Path supplied as argument.
 */
mc.ds.assertValidPath = function(path) {
  goog.asserts.assertArray(path, 'Path must be an array.');
  goog.array.forEach(path, function(pathEl) {
    if (!(pathEl === '*' || mc.ds.isValidPathElement(pathEl))) {
      goog.asserts.assert(false,
          goog.string.subs('Invalid path element "%s".', pathEl));
    }
  });
  return path;
};


/**
 * Asserts that the supplied path is fully qualified, i.e. does not to contain
 * any wildcards.
 * @param {mc.ds.Path} path Path to check.
 */
mc.ds.assertPathFullyQualified = function(path) {
  goog.asserts.assert(mc.ds.isPathFullyQualified(path),
      'Path is not fully qualified.');
};


/**
 * @param {mc.ds.Path} path Path to check.
 * @param {mc.ds.Path} otherPath Possible superset.
 * @return {boolean} Whether data tree elements described by path form a subset
 *     of data tree elements described by otherPath.
 */
mc.ds.isSubset = function(path, otherPath) {
  mc.ds.assertValidPath(path);
  mc.ds.assertValidPath(otherPath);
  if (path.length != otherPath.length) {
    return false;
  }
  return goog.array.every(otherPath, function(pathEl, index) {
    return path[index] == pathEl || pathEl == '*';
  });
};


/**
 * Decodes a path joined with /'s. A zero-length string represents an empty
 *     path.
 * @param {string} str Path as a string.
 * @return {mc.ds.Path} Path as an array.
 */
mc.ds.decodePath = function(str) {
  if (!str.length) { return []; }
  return str.split('/');
};


/**
 * @param {*} value Value.
 * @param {boolean=} opt_doNotClone Do not deep-clone the value.
 * @return {*} The supplied value or its deep clone if opt_doNotClone is true.
 * @private
 */
mc.ds.maybeClone_ = function(value, opt_doNotClone) {
  return opt_doNotClone ? value : goog.object.unsafeClone(value);
};



/**
 * An event with a fully or partially qualified path attached to it that can
 * be dispatched with dispatchEvent() method of a data node.
 * @param {mc.ds.Path} path Fully or partially qualified path from root.
 * @constructor
 */
mc.ds.Event = function(path) {
  mc.ds.assertValidPath(path);


  /**
   * Path associated with the event.
   * @type {mc.ds.Path}
   */
  this.path = path;
};



/**
 * An mc.ds.Event subclass for events with an event source object attached to
 * them.
 * @param {mc.ds.Path} path Fully qualified path.
 * @param {Object=} opt_source Optional event source object.
 * @constructor
 * @extends {mc.ds.Event}
 */
mc.ds.EventWithSource = function(path, opt_source) {
  goog.base(this, path);


  /**
   * Event source or null if not provided. For an object that both listens to
   * changes in data and can also modify the data, it can be useful to be able
   * to distinguish the events it has itself initiated by supplying a reference
   * to itself as source argument of set() method.
   * @type {?Object}
   */
  this.source = opt_source || null;
};
goog.inherits(mc.ds.EventWithSource, mc.ds.Event);



/**
 * An mc.ds.EventWithSource subclass for events indicating a change in a value,
 * and which have versions of the value before and after the change attached to
 * them.
 * @param {mc.ds.Path} path Fully qualified path to the changed value.
 * @param {*} oldValue Value before the change.
 * @param {*} newValue Value after the change.
 * @param {Object=} opt_source Optional event source object.
 * @constructor
 * @extends {mc.ds.EventWithSource}
 */
mc.ds.ChangeEvent = function(path, oldValue, newValue, opt_source) {
  goog.base(this, path, opt_source);


  /**
   * Value before the change.
   * @type {*}
   * @private
   */
  this.oldValue_ = oldValue;


  /**
   * Value after the change.
   * @type {*}
   * @private
   */
  this.newValue_ = newValue;
};
goog.inherits(mc.ds.ChangeEvent, mc.ds.EventWithSource);


/**
 * @param {boolean=} opt_doNotClone Do not deep-clone the value.
 * @return {*} Value after the change.
 */
mc.ds.ChangeEvent.prototype.getOldValue = function(opt_doNotClone) {
  return mc.ds.maybeClone_(this.oldValue_, opt_doNotClone);
};


/**
 * @param {boolean=} opt_doNotClone Do not deep-clone the value.
 * @return {*} Value after the change.
 */
mc.ds.ChangeEvent.prototype.getNewValue = function(opt_doNotClone) {
  return mc.ds.maybeClone_(this.newValue_, opt_doNotClone);
};



/**
 * A class which allows to easily stop listening by calling the dispose() method
 * of the listener instance. Can be used as follows:
 * <pre>
 * disposable.registerDisposable(node.listen(...));
 * </pre>
 * where disposable is any instance of goog.Disposable.
 * @param {!mc.ds.RootNode} rootNode Root data node.
 * @param {number} listenerID Listener's unique ID.
 * @param {mc.ds.Path} path Path from root which is listened to.
 * @constructor
 * @extends {goog.Disposable}
 * @private
 */
mc.ds.EventListener = function(rootNode, listenerID, path) {
  goog.base(this);


  /**
   * Root data node.
   * @type {!mc.ds.RootNode}
   * @private
   */
  this.rootNode_ = rootNode;


  /**
   * Listener's unique ID.
   * @type {number}
   * @private
   */
  this.listenerID_ = listenerID;


  /**
   * Path from root which is listened to.
   * @type {mc.ds.Path}
   * @private
   */
  this.path_ = path;
};
goog.inherits(mc.ds.EventListener, goog.Disposable);


/** @inheritDoc */
mc.ds.EventListener.prototype.disposeInternal = function() {
  goog.base(this, 'disposeInternal');
  mc.ds.setValueByKeys(this.rootNode_.listeners_,
      goog.array.concat(this.path_, mc.ds.LISTENER_KEY_, this.listenerID_));
};


/**
 * Keys used to identify leafs in the data tree.
 * @enum {string}
 * @private
 */
mc.ds.DATA_KEYS_ = {
  VALUE: '.',
  NODE: '$'
};


/**
 * String that is used in the event listeners index as a leaf key.
 * @type {string}
 * @const
 * @private
 */
mc.ds.LISTENER_KEY_ = '.';



/**
 * A class whose instances encapsulate a reference to a root node and a fully
 * qualified path from it.
 * @param {!mc.ds.Node} baseNode Any existing node.
 * @param {mc.ds.Path} path Fully qualified path from base node to the new node.
 * @constructor
 * @extends {goog.Disposable}
 */
mc.ds.Node = function(baseNode, path) {
  goog.base(this);
  mc.ds.assertValidPath(path);
  mc.ds.assertPathFullyQualified(path);
  var data = goog.asserts.assertObject(
      mc.ds.setValueByKeysIfUndefined(baseNode.data_, path, {}));
  goog.asserts.assert(!(mc.ds.DATA_KEYS_.NODE in data), goog.string.subs(
      'Node at path \'%s\' already exists.',
      baseNode.absPath(path).join('/')));
  if (mc.ds.ENABLE_SCHEMA && goog.isDef(baseNode.schemaElement_)) {
    this.schemaElement_ = mc.ds.schemaGet_(baseNode.path_,
        baseNode.schemaElement_, path);
  }
  data[mc.ds.DATA_KEYS_.NODE] = this;


  /**
   * Root node.
   * @type {!mc.ds.RootNode}
   * @private
   */
  this.rootNode_ = baseNode.rootNode_;


  /**
   * Fully qualified path to this node.
   * @type {mc.ds.Path}
   * @private
   */
  this.path_ = baseNode.absPath(path);


  /**
   * Element of the data tree at the same path as the node's path.
   * @type {!Object}
   * @private
   */
  this.data_ = data;


  /**
   * Associated element of the schema, a value type or an object (never a custom
   * type name, such names must be de-aliased).
   * @type {!Object|string|undefined}
   * @private
   */
  this.schemaElement_;
};
goog.inherits(mc.ds.Node, goog.Disposable);


/**
 * @return {!mc.ds.RootNode} The root node.
 */
mc.ds.Node.prototype.getRootNode = function() {
  return this.rootNode_;
};


/**
 * @return {mc.ds.Path} Path that this node points to.
 */
mc.ds.Node.prototype.getPath = function() {
  return this.path_;
};


/**
 * @param {mc.ds.Path} path Path relative to this node.
 * @return {mc.ds.Path} Path relative to root node.
 */
mc.ds.Node.prototype.absPath = function(path) {
  return goog.array.concat(this.path_, path);
};


/**
 * Returns a node at supplied path relative to this node. If such node doesn't
 * exist, returns null.
 * @param {mc.ds.Path} path Fully qualified path relative to this node.
 * @return {?mc.ds.Node} The node at provided relative path or null.
 */
mc.ds.Node.prototype.getNode = function(path) {
  mc.ds.assertValidPath(path);
  mc.ds.assertPathFullyQualified(path);
  if (!path.length) { return this; }
  if (mc.ds.ENABLE_SCHEMA && goog.isDef(this.schemaElement_)) {
    // Throw error if according to schema there can never be any element at
    // provided path.
    mc.ds.schemaGet_(this.path_, this.schemaElement_, path);
  }
  var node = goog.object.getValueByKeys(this.data_,
      goog.array.concat(path, mc.ds.DATA_KEYS_.NODE)) || null;
  goog.asserts.assert(goog.isNull(node) || node instanceof mc.ds.Node);
  return node;
};


/**
 * Calls a function for each node whose path matches the supplied relative path
 * from this node. If any call returns true, returns true (without checking the
 * rest). If all calls return false, returns false.
 * @param {mc.ds.Path} path Relative path to nodes, can contain wildcards.
 * @param {function(!mc.ds.Node, mc.ds.Path): boolean} f The function to call
 *     for every node. This function takes 2 arguments (the node and the fully
 *     qualified path to it relative to this node) and returns a boolean.
 * @param {Object=} opt_obj Scope within which to call f.
 * @return {boolean} Whether any node has passed the test.
 */
mc.ds.Node.prototype.someNode = function(path, f, opt_obj) {
  mc.ds.assertValidPath(path);
  if (mc.ds.ENABLE_SCHEMA && goog.isDef(this.schemaElement_)) {
    var foundSchemaElement = mc.ds.schemaForEach_(this.schemaElement_, path,
        function() {
          return true;
        }, this);
    if (!foundSchemaElement) {
      goog.asserts.assert(false,
          'According to schema, path \'%s\' will never match any nodes.',
          this.absPath(path).join('/'));
    }
  }
  var leafKeys = {};
  leafKeys[mc.ds.DATA_KEYS_.VALUE] = null;
  return mc.ds.some(this.data_, goog.array.concat(path, mc.ds.DATA_KEYS_.NODE),
      function(node, path) {
        goog.asserts.assert(node instanceof mc.ds.Node);
        return f.call(opt_obj || null, node,
            goog.array.slice(path, 0, path.length - 1));
      }, this, leafKeys);
};


/**
 * Calls a function for each node whose path matches the supplied relative path
 * from this node. If all calls return true, returns true. If any call returns
 * false, returns false (without checking the rest).
 * @param {mc.ds.Path} path Relative path to nodes, can contain wildcards.
 * @param {function(!mc.ds.Node, mc.ds.Path): boolean} f The function to call
 *     for every node. This function takes 2 arguments (the node and the fully
 *     qualified path to it relative to this node) and returns a boolean.
 * @param {Object=} opt_obj Scope within which to call f.
 * @return {boolean} Whether all nodes have passed the test.
 */
mc.ds.Node.prototype.everyNode = function(path, f, opt_obj) {
  return !this.someNode(path, function(node, path) {
    return !f.call(opt_obj || null, node, path);
  });
};


/**
 * Iterates over each node whose path matches the supplied relative path from
 * this node. Can be used to dispose nodes to free up the memory.
 * @param {mc.ds.Path} path Relative path to nodes, can contain wildcards.
 * @param {function(!mc.ds.Node, mc.ds.Path)} f The function to call for every
 *     node. This function takes 2 arguments (the node and the fully qualified
 *     path to it relative to this node).
 * @param {Object=} opt_obj Scope within which to call f.
 */
mc.ds.Node.prototype.forEachNode = function(path, f, opt_obj) {
  this.someNode(path, function(node, path) {
    f.call(opt_obj || null, node, path);
    return false;
  });
};


/**
 * Indicates that the supplied path has been read if transaction is in progress
 * and the function is not called from within a callback to a change event.
 * @param {mc.ds.Path} path Fully or partially qualified path relative to this
 *     node.
 */
mc.ds.Node.prototype.read = function(path) {
  this.rootNode_.read(this.absPath(path));
};


/**
 * Returns the value at provided path relative to this node.
 * @param {mc.ds.Path} path Fully qualified relative path.
 * @param {*=} opt_default Optional value to return if the value at this path
 *     does not exist. If this argument is omitted, when the value does not
 *     exist will return null.
 * @param {boolean=} opt_doNotClone Do not deep-clone the value.
 * @return {*} Value at provided path, the default value, or null.
 */
mc.ds.Node.prototype.get = function(path, opt_default, opt_doNotClone) {
  mc.ds.assertValidPath(path);
  mc.ds.assertPathFullyQualified(path);
  if (mc.ds.ENABLE_SCHEMA && goog.isDef(this.schemaElement_)) {
    var schemaElement = mc.ds.schemaGet_(this.path_, this.schemaElement_,
        path);
    goog.asserts.assert(schemaElement in mc.ds.primitiveSchemaElements_ &&
        schemaElement != 'EventPath',
        'Called get for path \'%s\' which according to schema does not point' +
        ' to a value.',
        this.absPath(path).join('/'));
  }
  var value = goog.object.getValueByKeys(this.data_,
      goog.array.concat(path, mc.ds.DATA_KEYS_.VALUE));
  this.read(path);
  return goog.isDef(value) ? mc.ds.maybeClone_(value, opt_doNotClone) :
      (goog.isDef(opt_default) ? opt_default : null);
};


/**
 * Calls a function for each value whose path matches the supplied relative path
 * from this node. If any call returns true, returns true (without checking the
 * rest). If all calls return false, returns false.
 * @param {mc.ds.Path} path Relative path to values, can contain wildcards.
 * @param {function(*, mc.ds.Path): boolean} f The function to call for every
 *     value. This function takes 2 arguments (the value and the fully qualified
 *     path to it relative to this node) and returns a boolean.
 * @param {Object=} opt_obj Scope within which to call f.
 * @param {boolean=} opt_doNotClone Do not deep-clone the value passed as
 *     as argument to callback.
 * @return {boolean} Whether any value has passed the test.
 */
mc.ds.Node.prototype.some = function(path, f, opt_obj, opt_doNotClone) {
  mc.ds.assertValidPath(path);
  if (mc.ds.ENABLE_SCHEMA && goog.isDef(this.schemaElement_)) {
    var foundSchemaElement = mc.ds.schemaForEach_(this.schemaElement_, path,
        function(schemaElement) {
          return goog.isString(schemaElement) && schemaElement != 'EventPath';
        }, this);
    if (!foundSchemaElement) {
      goog.asserts.assert(false,
          'According to schema, path \'%s\' will never match any values.',
          this.absPath(path).join('/'));
    }
  }
  this.read(path);
  var leafKeys = {};
  leafKeys[mc.ds.DATA_KEYS_.NODE] = null;
  return mc.ds.some(this.data_, goog.array.concat(path, mc.ds.DATA_KEYS_.VALUE),
      function(value, path) {
        return f.call(opt_obj || null, mc.ds.maybeClone_(value, opt_doNotClone),
            goog.array.slice(path, 0, path.length - 1));
      }, this, leafKeys);
};


/**
 * Calls a function for each value whose path matches the supplied relative path
 * from this node. If all calls return true, returns true. If any call returns
 * false, returns false (without checking the rest).
 * @param {mc.ds.Path} path Relative path to values, can contain wildcards.
 * @param {function(*, mc.ds.Path): boolean} f The function to call for every
 *     value. This function takes 2 arguments (the value and the fully qualified
 *     path to it relative to this node) and returns a boolean.
 * @param {Object=} opt_obj Scope within which to call f.
 * @param {boolean=} opt_doNotClone Do not deep-clone the value passed as
 *     as argument to callback.
 * @return {boolean} Whether all values have passed the test.
 */
mc.ds.Node.prototype.every = function(path, f, opt_obj, opt_doNotClone) {
  return !this.some(path, function(value, path) {
    return !f.call(opt_obj || null, value, path);
  }, null, opt_doNotClone);
};


/**
 * Iterates over each value whose path matches the supplied relative path from
 * this node.
 * @param {mc.ds.Path} path Relative path to values, can contain wildcards.
 * @param {function(*, mc.ds.Path)} f The function to call for every value.
 *     This function takes 2 arguments (the value and the fully qualified path
 *     to it relative to this node).
 * @param {Object=} opt_obj Scope within which to call f.
 * @param {boolean=} opt_doNotClone Do not deep-clone the value passed as
 *     as argument to callback.
 */
mc.ds.Node.prototype.forEach = function(path, f, opt_obj, opt_doNotClone) {
  this.some(path, function(value, path) {
    f.call(opt_obj || null, value, path);
    return false;
  }, null, opt_doNotClone);
};


/**
 * Sets a primitive value at provided path relative to his node and dispatches
 * the corresponding change event (mc.ds.ChangeEvent). Handlers of these events
 * should not themselves call this method, and if a transaction is in progress,
 * reads made by the handlers are not registered as part of the transaction.
 * @param {mc.ds.Path} path Fully qualified relative path.
 * @param {*} value Value to set.
 * @param {Object=} opt_source Optional event source object to attach to the
 *     change event.
 * @param {boolean=} opt_doNotClone Do not deep-clone the value.
 */
mc.ds.Node.prototype.set = function(path, value, opt_source, opt_doNotClone) {
  mc.ds.assertValidPath(path);
  mc.ds.assertPathFullyQualified(path);
  goog.asserts.assert(!this.rootNode_.dispatchingChangeEvent_,
      'Data cannot be modified by a handler of a change event.');
  value = mc.ds.maybeClone_(value, opt_doNotClone);
  if (mc.ds.ENABLE_SCHEMA && goog.isDef(this.schemaElement_)) {
    var schemaElement = mc.ds.schemaGet_(this.path_, this.schemaElement_,
        path);
    goog.asserts.assert(schemaElement in mc.ds.primitiveSchemaElements_ &&
        schemaElement != 'EventPath',
        'According to schema, the path \'%s\' does not point to a value.',
        this.absPath(path).join('/'));
    if (!goog.isNull(value)) {
      mc.ds.assertType_(value, goog.asserts.assertString(schemaElement),
          this.absPath(path));
    }
  }
  var oldValue = mc.ds.setValueByKeys(this.data_,
      goog.array.concat(path, mc.ds.DATA_KEYS_.VALUE),
      goog.isNull(value) ? undefined : value);
  oldValue = goog.isDef(oldValue) ? oldValue : null;
  var absPath = this.absPath(path);
  var changeEvent = new mc.ds.ChangeEvent(absPath, oldValue, value, opt_source);
  this.rootNode_.dispatchingChangeEvent_ = true;
  this.rootNode_.dispatchEvent_(changeEvent);
  this.rootNode_.dispatchingChangeEvent_ = false;
  var writeEvent = new mc.ds.WriteEvent();
  writeEvent.changeEvent = changeEvent;
  this.rootNode_.eventTarget_.dispatchEvent(writeEvent);
};


/**
 * Adds a listenter to events whose associated path is not mutually exclusive
 * with the supplied path relative to this node.
 * @param {mc.ds.Path} path Relative path, can contain wildcards.
 * @param {!Function} callback Function to call when the event is fired, takes
 *     the event as the single argument.
 * @param {!Object=} opt_obj Optional object to use as scope for callback.
 * @return {!mc.ds.EventListener} An object which should be disposed to stop
 *     listening.
 */
mc.ds.Node.prototype.listen = function(path, callback, opt_obj) {
  mc.ds.assertValidPath(path);
  if (mc.ds.ENABLE_SCHEMA && goog.isDef(this.schemaElement_)) {
    var foundSchemaElement = mc.ds.schemaForEach_(this.schemaElement_, path,
        function(schemaElement) {
          // We are looking for any schema element.
          return goog.isString(schemaElement);
        }, this);
    if (!foundSchemaElement) {
      goog.asserts.assert(false,
          'Called listen for path \'%s\' which according to schema' +
          ' will never match any events.',
          this.absPath(path).join('/'));
    }
  }
  return this.rootNode_.listen_(this.absPath(path), callback,
      opt_obj);
};


/** @inheritDoc */
mc.ds.Node.prototype.disposeInternal = function() {
  goog.base(this, 'disposeInternal');
  mc.ds.setValueByKeys(this.rootNode_.data_,
      goog.array.concat(this.path_, mc.ds.DATA_KEYS_.NODE));
};



/**
 * Class used for mc.ds.EventType.READ events.
 * @constructor
 * @extends {goog.events.Event}
 */
mc.ds.ReadEvent = function() {
  goog.base(this, mc.ds.EventType.READ);


  /**
   * Path at which the data has been read, can be partially qualified.
   * @type {mc.ds.Path}
   */
  this.path;
};
goog.inherits(mc.ds.ReadEvent, goog.events.Event);



/**
 * Class used for mc.ds.EventType.WRITE events.
 * @constructor
 * @extends {goog.events.Event}
 */
mc.ds.WriteEvent = function() {
  goog.base(this, mc.ds.EventType.WRITE);


  /**
   * An mc.ds.ChangeEvent instance that was dispatched through the internal
   * mechanism.
   * @type {!mc.ds.ChangeEvent}
   */
  this.changeEvent;
};
goog.inherits(mc.ds.WriteEvent, goog.events.Event);



/**
 * Root node of the data tree.
 * @param {!Object} data Data to initialize with. A tree structure of nested
 *     objects where leaf values are marked by '.' keys.
 * @constructor
 * @extends {mc.ds.Node}
 */
mc.ds.RootNode = function(data) {
  goog.Disposable.call(this);


  // Set superclass properties (private but defined in this module).
  this.rootNode_ = this;
  this.path_ = [];
  this.data_ = data;


  /**
   * An index of event listeners. A single listener with ID listenerID and path
   * ['pathEl1', 'pathEl2'] and will create the following index:
   * <pre>
   * {'pathEl1': {'pathEl2': {<key>: {listenerID: <callback function bound to
   *     scope object>}}}}
   * </pre>
   * where <key> = mc.ds.LISTENER_KEY_.
   * @type {!Object}
   * @private
   */
  this.listeners_ = {};


  /**
   * Counter used for listener IDs.
   * @type {number}
   * @private
   */
  this.listenerIdCounter_ = 0;


  /**
   * Set to true while dispatching change events.
   * @type {boolean}
   * @private
   */
  this.dispatchingChangeEvent_ = false;


  /**
   * Event target for dispatching events through standard mechanism.
   * @type {!goog.events.EventTarget}
   * @private
   */
  this.eventTarget_ = new goog.events.EventTarget();


  /**
   * Whether a transaction is in progress.
   * @type {boolean}
   * @private
   */
  this.transaction_ = false;
};
goog.inherits(mc.ds.RootNode, mc.ds.Node);


/**
 * Enum for events that are dispatched by a RootNode through its associated
 * EventTarget.
 * @type {Object}
 */
mc.ds.EventType = {
  // Dispatched when a transaction is started.
  TRANSACTION_START: 'transactionStart',

  // Dispatched when a transaction ends.
  TRANSACTION_END: 'transactionEnd',

  // Dispatched when a transaction is in progress and data is read using get,
  // some, every or forEach method of a node. Events are instances of
  // mc.ds.ReadEvent. Even when transaction is in progress, this event is not
  // dispatched when the read is made by a callback to a change event.
  READ: 'read',

  // Dispatched when set method of a node is called, events are instances of
  // mc.ds.WriteEvent.
  WRITE: 'write'
};


/**
 * Adds an event listener.
 * @param {mc.ds.Path} path Path from root, can contain wildcards.
 * @param {!Function} callback Function to call when the event is fired, takes
 *     the event as the single argument.
 * @param {!Object=} opt_obj Optional object to use as scope for callback.
 * @return {!mc.ds.EventListener} An object which should be disposed to stop
 *     listening.
 * @private
 */
mc.ds.RootNode.prototype.listen_ = function(path, callback, opt_obj) {
  var id = ++this.listenerIdCounter_;
  mc.ds.setValueByKeys(this.listeners_,
      goog.array.concat(path, mc.ds.LISTENER_KEY_, id),
      goog.bind(callback, opt_obj || null));
  return new mc.ds.EventListener(this, id, path);
};


/**
 * Dispatches an event.
 * @param {!mc.ds.Event} event Event to dispatch.
 * @private
 */
mc.ds.RootNode.prototype.dispatchEvent_ = function(event) {
  var dispatchEvent = function(path, listeners) {
    if (path.length) {
      if (path[0] == '*') {
        goog.object.forEach(listeners, function(value, key) {
          if (key != '.') {
            dispatchEvent(goog.array.slice(path, 1), value);
          }
        });
      } else {
        if ('*' in listeners) {
          dispatchEvent(goog.array.slice(path, 1), listeners['*']);
        }
        if (path[0] in listeners) {
          dispatchEvent(goog.array.slice(path, 1),
              listeners[path[0]]);
        }
      }
    } else {
      if (mc.ds.LISTENER_KEY_ in listeners) {
        goog.object.forEach(
            listeners[mc.ds.LISTENER_KEY_],
            function(callback) {
              callback.call(null, event);
            });
      }
    }
  };
  dispatchEvent(event.path, this.listeners_);
};


/**
 * @return {!goog.events.EventTarget} EventTarget used for dispatching events
 *     through standard mechanism.
 */
mc.ds.RootNode.prototype.getEventTarget = function() {
  return this.eventTarget_;
};


/**
 * Dispatches a custom event. When type checking is on, we check that the
 * event's path matches an 'EventPath' special element in the schema.
 * @param {!mc.ds.Event} event Event to dispatch.
 */
mc.ds.RootNode.prototype.dispatchEvent = function(event) {
  if (mc.ds.ENABLE_SCHEMA && goog.isDef(this.schemaElement_)) {
    var foundSchemaElement;
    mc.ds.schemaForEach_(this.schemaElement_, event.path,
        function(schemaElement) {
          goog.asserts.assert(schemaElement === 'EventPath',
              'Tried to dispatch a custom event for path \'%s\' which ' +
              ' according to schema can match an element other' +
              ' than \'EventPath\'',
              event.path.join('/'));
          foundSchemaElement = true;
        }, this);
    if (!foundSchemaElement) {
      goog.asserts.assert(false,
          'Tried to dispatch a custom event for path \'%s\' which ' +
          ' according to schema does not match an \'EventPath\'',
          event.path.join('/'));
    }
  }
  this.dispatchEvent_(event);
};


/**
 * For when the data is synced with other sites, indicates that if a sequence
 * of writes (calls to set method of mc.ds.Node) made by the callback is applied
 * over such consistent state of the data that the reads (calls to get and
 * forEach methods of mc.ds.Node) made by the callback would produce the same
 * results, the state of the data will also be consistent after each write is
 * applied. Note that writes should be made in such a way that the data would be
 * consistent throughout the transaction, not just at the end of it.
 *
 * A write made outside of a transaction is considered to be a transaction of
 * its own.
 *
 * If a transaction is already in progress, simply calls the callback without
 * doing anything else.
 *
 * Only reads and writes made synchronously are registered as part of the
 * transaction.
 *
 * @param {function(this: S): T} f Callback function.
 * @param {S=} opt_obj Optional context object for f.
 * @return {T} The return value of f.
 * @template T,S
 */
mc.ds.RootNode.prototype.inTransaction = function(f, opt_obj) {
  if (this.transaction_) {
    return f.call(opt_obj);
  } else {
    try {
      this.transaction_ = true;
      this.eventTarget_.dispatchEvent(mc.ds.EventType.TRANSACTION_START);
      return f.call(opt_obj);
    } finally {
      this.transaction_ = false;
      this.eventTarget_.dispatchEvent(mc.ds.EventType.TRANSACTION_END);
    }
  }
};


/** @inheritDoc */
mc.ds.RootNode.prototype.read = function(path) {
  if (this.transaction_ && !this.dispatchingChangeEvent_) {
    var readEvent = new mc.ds.ReadEvent();
    readEvent.path = path;
    this.eventTarget_.dispatchEvent(readEvent);
  }
};


/** @inheritDoc */
mc.ds.RootNode.prototype.disposeInternal = function() {
  goog.Disposable.prototype.disposeInternal.call(this);
};


/**
 * An object whose keys are type names and values are type schemas.
 * @type {!Object}
 * @private
 */
mc.ds.schema_ = {};


/**
 * The set of primitive schema elements, an object where keys are element names
 * and values are nulls.
 * @type {!Object.<null>}
 * @private
 */
mc.ds.primitiveSchemaElements_ = {
  // Numbers other than NaN and Infinity.
  'number': null,
  // Booleans.
  'boolean': null,
  // Strings.
  'string': null,
  // Values that are anything that can be serialized to JSON.
  '*': null,
  // Marks paths at which no value is stored but which can be used to dispatch
  // custom events.
  'EventPath': null
};


/**
 * Asserts that the value is of supplied type.
 * @param {*} value Value to check.
 * @param {string} type Expected type.
 * @param {mc.ds.Path} path Renamed path to value.
 * @private
 */
mc.ds.assertType_ = function(value, type, path) {
  var msg = goog.string.subs(
      'Invalid value type: expected %s at path \'%s\'.',
      type, path.join('/'));
  switch (type) {
    case 'number':
      goog.asserts.assert(goog.isNumber(value) &&
          goog.math.isFiniteNumber(value), msg);
      break;
    case 'boolean':
      goog.asserts.assertBoolean(value, msg);
      break;
    case 'string':
      goog.asserts.assertString(value, msg);
      break;
    case '*':
      break;
    default:
      goog.asserts.fail();
  }
};


/**
 * Registers schema for a data tree element type.
 * @param {string} type Type name, any unique string.
 * @param {*} schema Schema of descendant elements of the data tree. A tree of
 *     nested objects containing strings, each describing an element at a given
 *     path from any element with the type that is being defined. The string can
 *     be either a primitive schema element or a caller-defined type.
 *
 *     A '*' key is a wildcard matching any path element and should be the only
 *     key of its object.
 */
mc.ds.defineType = function(type, schema) {
  if (!mc.ds.ENABLE_SCHEMA) { return; }

  goog.asserts.assert(!(type in mc.ds.schema_),
      'Type already registered.');
  goog.asserts.assert(!(type in mc.ds.primitiveSchemaElements_),
      'Type name clashes with a primitive schema element.');
  /**
   * Checks the schema for errors.
   * @param {*} schema Schema element to check.
   * @return {*} Schema with replaced keys.
   */
  var walk = function(schema) {
    goog.asserts.assert((goog.isObject(schema) || goog.isString(schema)) &&
        !goog.isArray(schema) && !goog.isFunction(schema),
        'Schema must only contain strings or' +
        ' objects other than arrays and functions.');
    if (goog.isObject(schema)) {
      if ('*' in schema) {
        goog.asserts.assert(
            goog.object.getCount(goog.asserts.assertObject(schema)) == 1,
            'If an object in the schema contains a wildcard key,' +
            ' this object should have no other keys.');
      }
      var retval = {};
      goog.object.forEach(goog.asserts.assertObject(schema),
          function(value, key) {
            if (key !== '*') {
              mc.ds.assertValidPathElement(key);
            }
            retval[key] = walk(value);
          });
      return retval;
    } else {
      return schema;
    }
  };
  mc.ds.schema_[type] = walk(schema);
};


/**
 * Sets the type of a root node. This is required to enable schema-based checks.
 * Should be called after defining schema and before any calls to getNode method
 * of the root node.
 * @param {!mc.ds.RootNode} rootNode A root node.
 * @param {string} type Type name.
 */
mc.ds.setType = function(rootNode, type) {
  if (mc.ds.ENABLE_SCHEMA) {
    rootNode.schemaElement_ = mc.ds.schemaDealias_(type);
  }
};


/**
 * Takes an element of the schema, and if it is a custom type name, looks up an
 * object or value type that the custom type points to, possibly via
 * intermediate custom types.
 * @param {!Object|string} el Element of the schema.
 * @return {!Object|string} Element of the schema.
 * @private
 */
mc.ds.schemaDealias_ = function(el) {
  while (goog.isString(el) && !(el in mc.ds.primitiveSchemaElements_)) {
    goog.asserts.assert(el in mc.ds.schema_,
        'Schema references type \'%s\' which is not defined.', el);
    el = mc.ds.schema_[el];
  }
  return el;
};


/**
 * Key lookup for schema objects that understands wildcard key.
 * @param {!Object} obj Object in which to look.
 * @param {string} lookupKey Non-wildcard data tree key.
 * @return {!Object|string|null} Child element in the schema or null if no
 *     matching element.
 * @private
 */
mc.ds.schemaGetChild_ = function(obj, lookupKey) {
  var key;
  if ('*' in obj) {
    key = '*';
  } else {
    if (!(lookupKey in obj)) {
      return null;
    }
    key = lookupKey;
  }
  return obj[key];
};


/**
 * Returns the schema element that corresponds to data tree elements matching a
 * given path, taking as argument a type registry element that is known to
 * correspond to data tree elements matching a prefix of this path. Calling this
 * function implies expectation that such schema  element exists - if it is not
 * found, assertion error is thrown.
 * @param {mc.ds.Path} prefix Renamed path prefix, can contain wildcards.
 * @param {!Object|string} el Schema element associated with prefix, must be
 *     dealiased.
 * @param {mc.ds.Path} suffix Renamed remainder of the path, must be fully
 *     qualified.
 * @return {!Object|string} Type registry element that corresponds to a path
 *     which is concatenation of prefix and suffix. Custom type names are
 *     dealiased.
 * @private
 */
mc.ds.schemaGet_ = function(prefix, el, suffix) {
  if (!suffix.length) {
    return el;
  }

  if (goog.isString(el)) {
    goog.asserts.assert(false,
        'Invalid path \'%s\': according to schema, prefix \'%s\' points to a' +
        ' value.',
        goog.array.concat(prefix, suffix).join('/'), prefix.join('/'));
  }

  var childEl = mc.ds.schemaGetChild_(goog.asserts.assertObject(el),
      suffix[0]);
  goog.asserts.assert(!goog.isNull(childEl),
      'Invalid path \'%s\': unexpected path element \'%s\' following \'%s\'.',
      goog.array.concat(prefix, suffix).join('/'), suffix[0],
      prefix.join('/'));
  goog.asserts.assert(goog.isObject(childEl) || goog.isString(childEl));
  childEl = mc.ds.schemaDealias_(/** @type {!Object|string} */ (childEl));
  return mc.ds.schemaGet_(goog.array.concat(prefix, suffix[0]), childEl,
      goog.array.slice(suffix, 1));
};


/**
 * Iterates over all schema elements that can be associated with data tree
 * elements matching a given, possibly not fully qualified path. Takes as
 * argument the schema element that is known to correspond to data tree elements
 * matching some prefix of the path.
 * @param {!Object|string} el Schema element, must be dealiased.
 * @param {mc.ds.Path} path Renamed relative path, can contain wildcards.
 * @param {function((!Object|string)): *} f The callback function, takes the
 *     schema element (for custom types this will be type schema, not type name)
 *     as argument. Iteration stops if f returns a value that evaluates to true.
 * @param {Object=} opt_obj Optional scope in which to call f.
 * @return {*} A value that was returned by callback function if the iteration
 *     stopped early.
 * @private
 */
mc.ds.schemaForEach_ = function(el, path, f, opt_obj) {
  if (!path.length) {
    return f.call(opt_obj, el);
  }

  if (goog.isString(el)) {
    return;
  }

  if (path[0] == '*') {
    for (var key in el) {
      var childEl = mc.ds.schemaGetChild_(el, key);
      if (!goog.isNull(childEl)) {
        var retval = mc.ds.schemaForEach_(mc.ds.schemaDealias_(childEl),
            goog.array.slice(path, 1), f, opt_obj);
        if (retval) { return retval; }
      }
    }
  } else {
    var childEl = mc.ds.schemaGetChild_(el, path[0]);
    if (!goog.isNull(childEl)) {
      return mc.ds.schemaForEach_(mc.ds.schemaDealias_(childEl),
          goog.array.slice(path, 1), f, opt_obj);
    }
  }
};
