// Copyright 2012 MightyClient Authors. All Rights Reserved.
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//      http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS-IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

/**
 * @fileoverview Implementation of mc.Channel for when the server runs on
 * Google Appengine.
 *
 * For client-server communication protocol, see wiki/protocol_channel.md.
 */

goog.require('goog.Uri');
goog.require('goog.async.Delay');
goog.require('goog.debug.Logger');
goog.require('goog.json');
goog.require('goog.net.Jsonp');
goog.require('goog.net.XhrIo');
goog.require('goog.object');
goog.require('goog.ui.IdleTimer');
goog.require('mc.Channel');

goog.provide('mc.GaeChannel');



/**
 * Implementation of mc.Channel for when the server runs on Google Appengine.
 * Messages in either direction are not guaranteed to be always delivered, or
 * not to be delivered multiple times.
 *
 * Google's API is such that the GAE channel (strictly speaking, GAE channel's
 * socket) expires in two hours after you open it. To minimize channel openings,
 * we do not reopen the channel if the user was detected to be idle for some
 * time. When the user becomes active, the channel will be reopened if it isn't
 * open.
 *
 * @param {string} clientID Unique ID of this client.
 * @param {!goog.net.XhrIo=} opt_xhr XhrIo to use for forward channel
 *     requests. Exactly one of opt_xhr, opt_jsonp must be provided. The caller
 *     is responsible for disposing the XhrIo (the same does not apply to Jsonp
 *     just because Jsonp is not a disposable).
 * @param {!goog.net.Jsonp=} opt_jsonp Jsonp to use for forward channel
 *     requests. The URI passed to its constructor should be host/port +
 *     mc.GaeChannel.BASE_URL_PATH.
 * @param {!goog.ui.ActivityMonitor=} opt_activityMonitor The activity monitor
 *     keeping track of user interaction. Defaults to a default-constructed
 *     activity monitor. If a default activity monitor is used then this class
 *     will dispose of it. If an activity monitor is passed in then the caller
 *     remains responsible for disposing of it.
 * @param {boolean=} opt_forcePolling Use polling instead of GAE channel. Even
 *     when this parameter is false, we will still fall back on polling if there
 *     is an error creating a channel token on the server.
 * @constructor
 * @extends {mc.Channel}
 */
mc.GaeChannel = function(clientID, opt_xhr, opt_jsonp, opt_activityMonitor,
    opt_forcePolling) {
  goog.base(this);


  /**
   * Unique ID of this client.
   * @type {string}
   * @private
   */
  this.clientID_ = clientID;


  if (opt_jsonp) {
    goog.asserts.assert(!opt_xhr);
  } else {
    goog.asserts.assert(opt_xhr);
  }


  /**
   * XhrIo for forward channel requests.
   * @type {?goog.net.XhrIo}
   * @private
   */
  this.xhr_ = opt_xhr || null;
  if (this.xhr_) {
    this.xhr_.setTimeoutInterval(mc.GaeChannel.FORWARD_CHANNEL_TIMEOUT);
    goog.events.listen(this.xhr_, goog.net.EventType.COMPLETE,
        this.handleXhrComplete_, undefined, this);
  }


  /**
   * Jsonp for forward channel requests.
   * @type {?goog.net.Jsonp}
   * @private
   */
  this.jsonp_ = opt_jsonp || null;


  /**
   * Jsonp callback.
   * @type {!Function|undefined}
   * @private
   */
  this.jsonpReplyCallback_;


  /**
   * Jsonp error handler.
   * @type {!Function|undefined}
   * @private
   */
  this.jsonpErrorCallback_;


  if (this.jsonp_) {
    this.jsonp_.setRequestTimeout(mc.GaeChannel.FORWARD_CHANNEL_TIMEOUT);
    this.jsonpReplyCallback_ = goog.bind(this.handleJsonpReply_, this);
    this.jsonpErrorCallback_ = goog.bind(this.handleJsonpError_, this);
  }


  /**
   * The object returned by jsonp's send method, or null if jsonp is not active.
   * @type {?Object}
   * @private
   */
  this.jsonpRequest_ = null;


  /**
   * Whether to always use polling.
   * @type {boolean}
   * @private
   */
  this.forcePolling_ = opt_forcePolling || false;


  /**
   * Token used when getting server messages by polling. Non-null value means
   * that a backward channel is open with polling as delivery mechanism.
   * @type {?string}
   * @private
   */
  this.pollingToken_ = null;


  /**
   * An object where keys are service names and values are arrays of messages
   * not yet sent to the server.
   * @type {!Object.<!Array>}
   * @private
   */
  this.bufferedMessages_ = {};


  /**
   * An object where keys are service names and values are arrays of messages
   * that have been sent to the server but for which have not yet received a
   * successful response.
   * @type {!Object.<!Array>}
   * @private
   */
  this.sentMessages_ = {};


  /**
   * Number of forward channel retries so far, not counting the initial request.
   * @type {number}
   * @private
   */
  this.forwardChannelRetryCount_ = 0;


  /**
   * A timer for deferring sending of XHR request to the next cycle of brower
   * event loop.
   * @type {!goog.async.Delay}
   * @private
   */
  this.sendDelay_ = new goog.async.Delay(this.sendDelayCallback_, 0, this);
  this.registerDisposable(this.sendDelay_);


  /**
   * Set to true when this.sendDelay_ has fired but sending XHR was deferred
   * until channel XHR response is received or retry delay expires.
   * @type {boolean}
   * @private
   */
  this.sendPending_ = false;


  /**
   * A timer that fires when the earliest of canWait delays expires, restarted
   * whenever a call to send method is made.
   * @type {!goog.async.Delay}
   * @private
   */
  this.retryDelay_ = new goog.async.Delay(this.retryDelayCallback_,
      mc.GaeChannel.RETRY_DELAY, this);
  this.registerDisposable(this.retryDelay_);


  /**
   * GAE socket. A non-null value means that either the GAE channel is open,
   * or we have called the socket's open method and are waiting for the channel
   * to open.
   * @type {?appengine.Socket}
   * @private
   */
  this.socket_ = null;


  /**
   * Whether the GAE socket is open.
   * @type {boolean}
   * @private
   */
  this.socketOpen_ = false;


  /**
   * A timer that fires when channel token is not received from the server
   * within a certain timeout.
   * @type {!goog.async.Delay}
   * @private
   */
  this.tokenResponseTimeout_ = new goog.async.Delay(
      this.tokenResponseTimeoutCallback_,
      mc.GaeChannel.TOKEN_RESPONSE_TIMEOUT, this);
  this.registerDisposable(this.tokenResponseTimeout_);


  /**
   * A timer that fires when GAE socket opening times out.
   * @type {!goog.async.Delay}
   * @private
   */
  this.socketTimeout_ = new goog.async.Delay(this.socketTimeoutCallback_,
      mc.GaeChannel.SOCKET_OPEN_TIMEOUT, this);
  this.registerDisposable(this.socketTimeout_);


  /**
   * The activity monitor keeping track of user interaction.
   * @type {!goog.ui.IdleTimer}
   * @private
   */
  this.idleTimer_ = new goog.ui.IdleTimer(mc.GaeChannel.IDLE_THRESHOLD,
      opt_activityMonitor);
  var eh = new goog.events.EventHandler(this);
  this.registerDisposable(eh);
  eh.listen(this.idleTimer_, goog.ui.IdleTimer.Event.BECOME_ACTIVE,
      this.handleBecomeActive_);


  /**
   * A timer that is used to safeguard against automatically re-opening the
   * backward channel too often.
   * @type {!goog.async.Delay}
   * @private
   */
  this.autoSocketReopenDelay_ = new goog.async.Delay(function() {},
      mc.GaeChannel.AUTO_SOKET_REOPEN_DELAY);
  this.registerDisposable(this.autoSocketReopenDelay_);


  /**
   * A timer that is used for polling requests.
   * @type {!goog.async.Delay}
   * @private
   */
  this.pollingDelay_ = new goog.async.Delay(this.poll_,
      mc.GaeChannel.POLLING_TIMEOUT, this);
  this.registerDisposable(this.pollingDelay_);


  /**
   * A timer that fires when polling token expires.
   * @type {!goog.async.Delay}
   * @private
   */
  this.pollingTokenExpirationDelay_ = new goog.async.Delay(
      this.handlePollingTokenExpiration_,
      mc.GaeChannel.POLLING_TOKEN_EXPIRATION, this);
  this.registerDisposable(this.pollingTokenExpirationDelay_);


  this.registerService(mc.GaeChannel.SYSTEM_SERVICE_,
      this.handleSystemServiceMessage_, this);
};
goog.inherits(mc.GaeChannel, mc.Channel);


/**
 * The base relative path for making requests to the server.
 * @type {string}
 */
mc.GaeChannel.BASE_URL_PATH = '/mc_channel';


/**
 * Name of the service used internally by this module.
 * @type {string}
 * @private
 * @const
 */
mc.GaeChannel.SYSTEM_SERVICE_ = 'system';


/**
 * The timeout in milliseconds for a forward channel request.
 * @type {number}
 */
mc.GaeChannel.FORWARD_CHANNEL_TIMEOUT = 120 * 1000;


/**
 * Maximum number of retries for forward channel requests, not counting the
 * initial request.
 * @type {number}
 */
mc.GaeChannel.FORWARD_CHANNEL_MAX_RETRIES = 1;


/**
 * Delay in milliseconds before a retry of a forward channel request.
 * @type {number}
 */
mc.GaeChannel.RETRY_DELAY = 5 * 1000;


/**
 * The timeout in milliseconds for receiving the channel token after the message
 * requesting it has been sent.
 * @type {number}
 */
mc.GaeChannel.TOKEN_RESPONSE_TIMEOUT = 30 * 1000;


/**
 * The timeout in milliseconds for opening a GAE socket.
 * @type {number}
 */
mc.GaeChannel.SOCKET_OPEN_TIMEOUT = 30 * 1000;


/**
 * Amount of time in milliseconds at which we consider the user has gone idle.
 * @type {number}
 */
mc.GaeChannel.IDLE_THRESHOLD = 30 * 60 * 1000;


/**
 * Amount of time in milliseconds for autoSocketReopenDelay_.
 * @type {number}
 */
mc.GaeChannel.AUTO_SOKET_REOPEN_DELAY = 60 * 60 * 1000;


/**
 * Time in milliseconds between polling requests.
 * @type {number}
 */
mc.GaeChannel.POLLING_TIMEOUT = 60 * 1000;


/**
 * Time in milliseconds that it takes for a polling token to expire.
 * @type {number}
 */
mc.GaeChannel.POLLING_TOKEN_EXPIRATION = 2 * 60 * 60 * 1000;


/**
 * Event types dispatched by mc.GaeChannel. Please see wiki/channel.md for
 * details.
 * @enum {string}
 */
mc.GaeChannel.EventType = {
  // Dispatched when opening of the channel is started automatically, i.e. not
  // as a result of calling open() method.
  AUTO_OPEN_START: 'autoOpenStart',
  // Dispatched when opening of the channel has failed. Call open() method to
  // try again.
  OPEN_FAILURE: 'backwardChannelOpenFailure'
};


/**
 * Logger.
 * @type {!goog.debug.Logger}
 * @protected
 */
mc.GaeChannel.prototype.logger = goog.debug.Logger.getLogger('mc.GaeChannel');


/**
 * @return {boolean} Whether XHR/Jsonp forward channel request is active.
 * @private
 */
mc.GaeChannel.prototype.isForwardChannelRequestActive_ = function() {
  if (this.xhr_) {
    return this.xhr_.isActive();
  } else {
    return !goog.isNull(this.jsonpRequest_);
  }
};


/**
 * Moves all messages in bufferedMessages_ to sentMessages_ and then sends the
 * a forward channel request with all messages in sentMessages_.
 * @private
 */
mc.GaeChannel.prototype.sendForwardChannelRequest_ = function() {
  this.logger.fine('sendForwardChannelRequest_');
  goog.object.forEach(this.bufferedMessages_,
      function(bufferedMessagesForService, service) {
        var sentMessagesForService = goog.object.setIfUndefined(
            this.sentMessages_, service, []);
        goog.asserts.assertArray(sentMessagesForService);
        goog.array.extend(sentMessagesForService, bufferedMessagesForService);
      }, this);
  this.bufferedMessages_ = {};
  var jsonMessages = goog.json.serialize(this.sentMessages_);
  if (this.xhr_) {
    var postParams = [
      'cid=' + encodeURIComponent(this.clientID_),
      'm=' + encodeURIComponent(jsonMessages)
    ];
    var location = window.location;
    var forwardChannelUri = goog.Uri.create(location.protocol, null,
        location.hostname, location.port, mc.GaeChannel.BASE_URL_PATH);
    forwardChannelUri.makeUnique();
    this.xhr_.send(forwardChannelUri, 'POST', postParams.join('&'));
  } else {
    var getParams = {
      'cid': this.clientID_,
      'm': jsonMessages
    };
    getParams[goog.Uri.RANDOM_PARAM] = goog.string.getRandomString();
    this.jsonpRequest_ = this.jsonp_.send(getParams, this.jsonpReplyCallback_,
        this.jsonpErrorCallback_);
  }
};


/**
 * A callback for when sendDelay_ fires.
 * @private
 */
mc.GaeChannel.prototype.sendDelayCallback_ = function() {
  this.logger.fine('sendDelay_ has fired.');
  if (this.isForwardChannelRequestActive_() ||
      this.retryDelay_.isActive()) {
    this.sendPending_ = true;
    this.logger.fine('Sending forward channel request was deferred.');
  } else {
    this.sendForwardChannelRequest_();
  }
};


/**
 * A callback for when retryDelay_ fires.
 * @private
 */
mc.GaeChannel.prototype.retryDelayCallback_ = function() {
  this.logger.fine('retryDelay_ has fired.');
  this.sendPending_ = false;
  this.sendForwardChannelRequest_();
};


/**
 * Relays incoming messages to appropriate services.
 * @param {*} messages Incoming messages as an object where keys are serivice
 *     names and values are arrays of payloads.
 * @private
 */
mc.GaeChannel.prototype.relayIncomingMessages_ = function(messages) {
  goog.asserts.assertObject(messages);
  goog.object.forEach(messages, function(messagesForService, service) {
    goog.asserts.assertArray(messagesForService);
    goog.array.forEach(messagesForService, function(payload) {
      this.handleIncomingMessage(service, payload);
    }, this);
  }, this);
};


/**
 * Handles successful response to either XHR or Jsonp forward channel request.
 * @param {*} data Deserialized response data, undefined if no data was sent.
 * @private
 */
mc.GaeChannel.prototype.handleForwardChannelResponse_ = function(data) {
  if (goog.isDef(data)) {
    this.relayIncomingMessages_(data);
  }
  this.forwardChannelRetryCount_ = 0;
  this.sentMessages_ = {};
  if (this.sendPending_) {
    this.sendPending_ = false;
    // XHR must be sent asyncroniously because at this point XhrIo is still
    // active.
    this.sendDelay_.start();
  }
};


/**
 * Handler for when XHR or Jsonp completes unsuccessfully.
 * @private
 */
mc.GaeChannel.prototype.handleForwardChannelRequestError_ = function() {
  if (this.forwardChannelRetryCount_ <
      mc.GaeChannel.FORWARD_CHANNEL_MAX_RETRIES) {
    this.forwardChannelRetryCount_++;
    this.retryDelay_.start();
    this.logger.info(goog.string.subs(
        'Scheduled retry #%s of forward channel request.',
        this.forwardChannelRetryCount_));
  } else {
    this.forwardChannelRetryCount_ = 0;
    this.logger.info('Reached maximum number of retries of forward' +
        ' channel request.');
    if (!goog.isNull(this.pollingToken_)) {
      this.pollingToken_ = null;
      this.pollingTokenExpirationDelay_.stop();
      this.pollingDelay_.stop();
      this.dispatchEvent(mc.Channel.EventType.CLOSE);
      this.dispatchEvent(mc.GaeChannel.EventType.OPEN_FAILURE);
    }
  }
};


/**
 * Handles XhrIo's COMPLETE event.
 * @private
 */
mc.GaeChannel.prototype.handleXhrComplete_ = function() {
  if (this.xhr_.getLastErrorCode() == goog.net.ErrorCode.NO_ERROR) {
    this.logger.fine('Forward channel XHR completed successfully.');
  } else if (this.xhr_.getLastErrorCode() == goog.net.ErrorCode.ABORT) {
    this.logger.fine('Forward channel XHR was aborted.');
  } else if (this.xhr_.getLastErrorCode() == goog.net.ErrorCode.TIMEOUT) {
    this.logger.warning('Forward channel XHR timed out.');
  } else {
    this.logger.warning(goog.string.subs(
        'Forward channel XHR failed before timing out, ' +
        ' status code %s, status text "%s".',
        this.xhr_.getStatus(), this.xhr_.getStatusText()));
  }
  if (this.xhr_.getLastErrorCode() == goog.net.ErrorCode.NO_ERROR) {
    var data;
    var responseText = this.xhr_.getResponseText();
    if (responseText.length) {
      try {
        data = goog.json.unsafeParse(responseText);
      } catch (e) {
        this.logger.warning(
            'Could not parse response to forward channel XHR.', e);
      }
    }
    this.handleForwardChannelResponse_(data);
  } else {
    this.handleForwardChannelRequestError_();
  }
};


/**
 * Handles a reply to Jsonp request.
 * @param {*} responseData Response data, undefined if no data was sent.
 * @private
 */
mc.GaeChannel.prototype.handleJsonpReply_ = function(responseData) {
  this.logger.fine('Forward channel Jsonp request completed successfully.');
  this.handleForwardChannelResponse_(responseData);
  // Set *after* calling the generic handler just to be consistent with XhrIo.
  this.jsonpRequest_ = null;
};


/**
 * Handles a Jsonp error.
 * @private
 */
mc.GaeChannel.prototype.handleJsonpError_ = function() {
  this.logger.warning('Forward channel Jsonp request failed.');
  this.handleForwardChannelRequestError_();
  // Set *after* calling the generic handler just to be consistent with XhrIo.
  // If the error occurs inside send() method, this.jsonpRequest_ will be
  // overwritten by return value of send(), but it will also be null.
  this.jsonpRequest_ = null;
};


/**
 * Instantiates a new appengine.Channel and returns it.
 * @param {string} token GAE token.
 * @return {!appengine.Channel} New channel instance.
 * @protected
 */
mc.GaeChannel.prototype.createChannel = function(token) {
  return new appengine.Channel(token);
};


/**
 * Opens the GAE socket.
 * @param {string} token GAE token.
 * @private
 */
mc.GaeChannel.prototype.openSocket_ = function(token) {
  goog.asserts.assert(goog.isNull(this.socket_), 'Socket already open.');
  this.logger.fine('openSocket_');
  this.socket_ = this.createChannel(token).open();
  this.socket_.onopen = goog.bind(this.handleSocketOpen_, this);
  this.socket_.onmessage = goog.bind(this.handleSocketMessage_, this);
  this.socket_.onerror = goog.bind(this.handleSocketError_, this);
  this.socket_.onclose = goog.bind(this.handleSocketClose_, this);
  this.socketTimeout_.start();
};


/**
 * Handles socket message.
 * @param {string} message Text Message.
 * @private
 */
mc.GaeChannel.prototype.handleSocketMessage_ = function(message) {
  var messageText = message['data'];
  this.logger.fine('GAE socket message received.');
  if (messageText.length) {
    var messageData;
    try {
      messageData = goog.json.unsafeParse(messageText);
    } catch (e) {
      this.logger.warning('Could not parse GAE socket message.', e);
    }
    if (messageData) {
      this.relayIncomingMessages_(messageData);
    }
  } else {
    this.logger.warning('Received an empty message through GAE socket.');
  }
};


/**
 * Handles socket open event.
 * @private
 */
mc.GaeChannel.prototype.handleSocketOpen_ = function() {
  this.logger.info('GAE socket open.');
  this.socketTimeout_.stop();
  this.socketOpen_ = true;
  this.dispatchEvent(mc.Channel.EventType.OPEN);
};


/**
 * Handles socket close event.
 * @private
 */
mc.GaeChannel.prototype.handleSocketClose_ = function() {
  this.logger.info('GAE socket closed.');
};


/**
 * Handles socket error.
 * @param {!Object} error An object containing error description.
 * @private
 */
mc.GaeChannel.prototype.handleSocketError_ = function(error) {
  this.logger.info('GAE socket error: ' + error['description']);
  this.maybeDisposeSocket_();
  this.socketOpen_ = false;
  this.dispatchEvent(mc.Channel.EventType.CLOSE);
  if (!this.idleTimer_.isIdle() && !this.autoSocketReopenDelay_.isActive()) {
    this.autoSocketReopenDelay_.start();
    this.dispatchEvent(mc.GaeChannel.EventType.AUTO_OPEN_START);
    this.open_();
  } else {
    if (this.idleTimer_.isIdle()) {
      this.logger.info(
          'Will not request channel token because the user is idle.');
    } else if (this.autoSocketReopenDelay_.isActive()) {
      this.logger.info(
          'Will not request channel token because the lock that safeguards' +
          ' against too frequent automatic channel reopenings is active.');
      this.dispatchEvent(mc.GaeChannel.EventType.OPEN_FAILURE);
    }
  }
};


/**
 * Disposes the socket if this.socket_ is not null.
 * @private
 */
mc.GaeChannel.prototype.maybeDisposeSocket_ = function() {
  this.logger.fine('maybeDisposeSocket_');
  if (!goog.isNull(this.socket_)) {
    this.logger.fine('Will call this.socket_.close.');
    /** @preserveTry */
    try {
      this.socket_.close();
    } catch (e) {
      this.logger.warning('Error when closing socket.', e);
    }
    this.socket_ = null;
  }
};


/**
 * A callback for when channel token is not received from the server within
 * a certain timeout.
 * @private
 */
mc.GaeChannel.prototype.tokenResponseTimeoutCallback_ = function() {
  this.logger.warning('Token request has timed out.');
  this.dispatchEvent(mc.GaeChannel.EventType.OPEN_FAILURE);
};


/**
 * A callback for when GAE socket opening times out.
 * @private
 */
mc.GaeChannel.prototype.socketTimeoutCallback_ = function() {
  this.logger.warning('GAE socket opening has timed out.');
  this.maybeDisposeSocket_();
  this.dispatchEvent(mc.GaeChannel.EventType.OPEN_FAILURE);
};


/**
 * Polls the server and schedules the next polling request.
 * @private
 */
mc.GaeChannel.prototype.poll_ = function() {
  this.send(mc.GaeChannel.SYSTEM_SERVICE_, {'poll': this.pollingToken_});
  this.pollingDelay_.start();
};


/**
 * Handles expiration of polling token.
 * @private
 */
mc.GaeChannel.prototype.handlePollingTokenExpiration_ = function() {
  this.logger.fine('Polling token has expired');
  this.pollingToken_ = null;
  this.pollingDelay_.stop();
  this.dispatchEvent(mc.Channel.EventType.CLOSE);
  if (this.idleTimer_.isIdle()) {
    this.logger.info(
        'Will not reopen backward channel because the user is idle.');
  } else {
    this.dispatchEvent(mc.GaeChannel.EventType.AUTO_OPEN_START);
    this.open_();
  }
};


/**
 * Handles a system service message.
 * @param {*} payload Deserialized service message.
 * @private
 */
mc.GaeChannel.prototype.handleSystemServiceMessage_ = function(payload) {
  goog.asserts.assertObject(payload);
  // When we get an 'opened' message, make sure we never have a GAE Channel and
  // polling active at the same time: one or both of this.pollingToken_ and
  // this.socket_ must be null.
  if ('opened' in payload) {
    if (goog.isNull(this.pollingToken_) && goog.isNull(this.socket_)) {
      var token = payload['opened']['token'];
      if (payload['opened']['poll']) {
        this.pollingToken_ = token;
        this.pollingTokenExpirationDelay_.start();
        this.logger.fine('Will start polling.');
        this.pollingDelay_.start();
        this.dispatchEvent(mc.Channel.EventType.OPEN);
      } else {
        this.tokenResponseTimeout_.stop();
        this.openSocket_(token);
      }
    }
  } else if ('pickup' in payload) {
    this.send(mc.GaeChannel.SYSTEM_SERVICE_, {'pickup': payload['pickup']});
  } else {
    goog.asserts.assert(false, 'Unrecognized system service message.');
  }
};


/**
 * Opens the backward channel.
 * @private
 */
mc.GaeChannel.prototype.open_ = function() {
  this.logger.fine('open_');
  this.send(mc.GaeChannel.SYSTEM_SERVICE_, {'getToken': this.forcePolling_});
  this.tokenResponseTimeout_.start();
};


/** @inheritDoc */
mc.GaeChannel.prototype.isOpen = function() {
  return this.socketOpen_ || !goog.isNull(this.pollingToken_);
};


/**
 * Initiates opening of the backward channel if it is not already in progress.
 * When the channel gets opened, mc.Channel.EventType.OPEN will be dispatched.
 * Should not be called when the channel is already open.
 */
mc.GaeChannel.prototype.open = function() {
  goog.asserts.assert(!this.isOpen(), 'Channel already open.');
  if (goog.isNull(this.socket_)) {
    this.open_();
  }
};


/** @inheritDoc */
mc.GaeChannel.prototype.sendInternal = function(service, payload) {
  this.logger.fine('sendInternal');
  var messagesForService = goog.object.setIfUndefined(this.bufferedMessages_,
      service, []);
  messagesForService.push(payload);
  this.sendDelay_.start();
};


/**
 * Handles an event dispatched when an idle user becomes active.
 * @private
 */
mc.GaeChannel.prototype.handleBecomeActive_ = function() {
  this.logger.info('User became active.');
  if (!this.isOpen() && !this.autoSocketReopenDelay_.isActive()) {
    if (!this.forcePolling_) {
      this.autoSocketReopenDelay_.start();
    }
    this.dispatchEvent(mc.GaeChannel.EventType.AUTO_OPEN_START);
    this.open_();
  }
};


/** @inheritDoc */
mc.GaeChannel.prototype.disposeInternal = function() {
  this.logger.fine('disposeInternal');
  goog.base(this, 'disposeInternal');
  if (!goog.isNull(this.jsonpRequest_)) {
    this.jsonp_.cancel(this.jsonpRequest_);
  }
  this.maybeDisposeSocket_();
};
