// Copyright 2012 MightyClient Authors. All Rights Reserved.
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//      http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS-IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

/**
 * @fileoverview Defines a class for objects that keep an mc.ds data tree in
 * sync with the server.
 *
 * For client-server communication protocol, see wiki/protocol_datasource.md.
 */


goog.require('goog.async.Delay');
goog.require('goog.debug.Logger');
goog.require('goog.events.EventHandler');
goog.require('goog.events.EventTarget');
goog.require('goog.structs.LinkedMap');
goog.require('mc.Channel');
goog.require('mc.ds');

goog.provide('mc.ServerAgent');



/**
 * A class for objects that keep an mc.ds data tree in sync with the server.
 *
 * Requires that data tree not be modified (i.e. set() method should not be
 * called on any node) in the following two cases:
 * <ul>
 *   <li>When the associated channel is closed.</li>
 *   <li>In a callback to an mc.ds.Change event if the event source is
 *       mc.ServerAgent.SERVER or mc.ServerAgent.ROLLBACK.</li>
 * </ul>
 * @param {!mc.Channel} channel Channel through which to communicate with the
 *     server.
 * @param {string} service Service name to use for two-way messaging through the
 *     channel.
 * @constructor
 * @extends {goog.events.EventTarget}
 */
mc.ServerAgent = function(channel, service) {
  goog.base(this);


  /**
   * Root node of the data tree, or null before the initial state of the data
   * has been loaded.
   * @type {?mc.ds.RootNode}
   * @private
   */
  this.rootNode_ = null;


  /**
   * Channel through which to communicate with the server.
   * @type {!mc.Channel}
   * @private
   */
  this.channel_ = channel;
  channel.registerService(service, this.handleMessage_, this);


  /**
   * Event handler registered as a dependent disposable.
   * @type {!goog.events.EventHandler}
   * @private
   */
  this.eventHandler_ = new goog.events.EventHandler(this);
  this.registerDisposable(this.eventHandler_);


  this.eventHandler_.listen(this.channel_, mc.Channel.EventType.OPEN,
      this.handleChannelOpen_);
  this.eventHandler_.listen(this.channel_, mc.Channel.EventType.CLOSE,
      this.handleChannelClose_);


  /**
   * Service name to use for two-way messaging through the channel.
   * @type {string}
   * @private
   */
  this.service_ = service;


  /**
   * Random token that remains unchanged for as long as the channel is open.
   * Reset when we first connect and before sending a request to reconnect.
   * @type {string}
   * @private
   */
  this.connectionToken_;


  /**
   * The token of the in-memory state of the data on the server. This is null
   * initially and becomes non-null when we are connected. If the channel closes
   * and we get temporarily disconnected, this becomes null again until we are
   * reconnected.
   * @type {?string}
   * @private
   */
  this.inMemoryStateToken_ = null;


  /**
   * An array of messages other than 'init_success' and 'init_failure' received
   * before the RootNode has been instantiated.
   * @type {!Array}
   * @private
   */
  this.bufferedServerMessages_ = [];


  /**
   * A version of the data such that all updates up to this version, possibly
   * including those only committed in-memory, have been applied on this client.
   * @type {number}
   * @private
   */
  this.version_;


  /**
   * The last version of the data that is known to be committed on disk.
   * @type {number}
   * @private
   */
  this.diskVersion_;


  /**
   * Reads and writes made inside a transaction if one is in progress, null
   * otherwise.
   * @type {?mc.ServerAgent.Transaction}
   * @private
   */
  this.transaction_ = null;


  /**
   * A counter used for internal IDs assigned to transactions for indexing.
   * @type {number}
   * @private
   */
  this.transactionIDCounter_ = 0;


  /**
   * A LinkedMap of transactions that have not yet been sent to the server.
   * Keys are transaction IDs, values are mc.ServerAgent.Transaction's.
   * @type {!goog.structs.LinkedMap}
   * @private
   */
  this.bufferedUpdates_ = new goog.structs.LinkedMap();


  /**
   * A LinkedMap of transactions that have been sent to the server but for
   * which the outcomes have not yet been received. Keys are transaction IDs,
   * values are mc.ServerAgent.Transaction's.
   * @type {!goog.structs.LinkedMap}
   * @private
   */
  this.sentUpdates_ = new goog.structs.LinkedMap();


  /**
   * A LinkedMap of transactions that are known to be committed on the server
   * in-memory and which are later than this.version_. Keys are transaction IDs,
   * values are mc.ServerAgent.Transaction's.
   * @type {!goog.structs.LinkedMap}
   * @private
   */
  this.committedUpdates_ = new goog.structs.LinkedMap();


  /**
   * The resulting version of the latest transaction added to
   * this.committedUpdates_ (null if no such transactions yet),
   * used for determining whether local changes have been saved to disk.
   * @type {?number}
   * @private
   */
  this.lastCommittedVersion_ = null;


  /**
   * Data on values modified by this.transaction_, this.bufferedUpdates_,
   * this.sentUpdates_, or this.committedUpdates_. Keys are '/'-joined paths and
   * values are objects {initial: <value before any of the updates>, updates: <a
   * LinkedMap whose keys are transaction IDs and values are arrays of values
   * set by the transaction in the order that they were set>}.
   * @type {!Object.<(!{initial: *, updates: !goog.structs.LinkedMap})>}
   * @private
   */
  this.history_ = {};


  /**
   * An object where values are updates not originating from this client that
   * have been received from the server but not yet applied, and keys are
   * versions after the corresponding update.
   * @type {!Object.<{encodedPath: string, newValue: *}>}
   * @private
   */
  this.receivedUpdates_ = {};


  /**
   * The return value of areChangesSaved_(), last time it was called. This is
   * used to only dispatch SAVED_STATUS_CHANGE when the status actually changes.
   * @type {boolean}
   * @private
   */
  this.lastChangesSavedStatus_ = true;


  /**
   * A timer for making a call to send_ method in the next cycle of browser
   * event loop.
   * @type {!goog.async.Delay}
   * @private
   */
  this.sendDelay_ = new goog.async.Delay(this.send_, 0, this);
  this.registerDisposable(this.sendDelay_);


  /**
   * A timer used to detect the timeout of 'connect' requests.
   * @type {!goog.async.Delay}
   * @private
   */
  this.connectResponseTimeoutDelay_ = new goog.async.Delay(
      this.responseTimeoutCallback_,
      mc.ServerAgent.CONNECT_RESPONSE_TIMEOUT, this);
  this.registerDisposable(this.connectResponseTimeoutDelay_);


  /**
   * A timer used to detect the timeout of 'reconnect' requests.
   * @type {!goog.async.Delay}
   * @private
   */
  this.reconnectResponseTimeoutDelay_ = new goog.async.Delay(
      this.responseTimeoutCallback_,
      mc.ServerAgent.RECONNECT_RESPONSE_TIMEOUT, this);
  this.registerDisposable(this.reconnectResponseTimeoutDelay_);


  /**
   * A timer started or re-started whenever updates are sent to the server and
   * stopped when areChangesSaved_ starts returning true.
   * @type {!goog.async.Delay}
   * @private
   */
  this.updatesResponseTimeoutDelay_ = new goog.async.Delay(
      this.responseTimeoutCallback_,
      mc.ServerAgent.UPDATES_RESPONSE_TIMEOUT, this);
  this.registerDisposable(this.updatesResponseTimeoutDelay_);


  if (this.channel_.isOpen()) {
    this.sendConnectRequest_();
  }
};
goog.inherits(mc.ServerAgent, goog.events.EventTarget);


/**
 * A typedef (created for documentation purposes) representing an object tree
 * that contains information on reads that were made inside a transaction. For
 * each read, the tree contains a value of null at the corresponding path (which
 * can contain wildcards) plus LEAF_KEY_.
 * @typedef {!Object}
 */
mc.ServerAgent.Reads;


/**
 * A typedef for a data structure describing a transaction. writes property is
 * an array of '/'-joined paths where the transaction has set a value, in the
 * order that the values were set. version property is added if and when the
 * outcome of the transaction has been received from the server, it is the
 * version of the data after the last of the updates that the transaction has
 * made.
 * @typedef {!{reads: mc.ServerAgent.Reads, writes: !Array.<string>,
 *     version: (string|undefined)}}
 */
mc.ServerAgent.Transaction;


/**
 * Token object used as event source for updates not originating from this
 * client.
 * @type {Object}
 * @const
 */
mc.ServerAgent.SERVER = {};


/**
 * Token object used as event source for rollbacks of changes made by this
 * client.
 * @type {Object}
 * @const
 */
mc.ServerAgent.ROLLBACK = {};


/**
 * Types of events dispatched by a ServerAgent.
 * @enum {string}
 */
mc.ServerAgent.EventType = {
  // Dispatched when the datasource has been initialized, i.e. the initial state
  // of the data has been loaded and the root node of the data tree has been
  // instantiated and is accessible with getRootNode().
  INITIALIZED: 'initialized',

  // Dispatched when return value of areChangesSaved() method changes.
  SAVED_STATUS_CHANGE: 'savedStatusChange',

  // Dispatched when some changes made locally have been rolled back.
  ROLLBACK: 'rollback',

  // Dispatched when an error notification is received from the server or when
  // a response is expected from the server but is not received within a
  // certain period of time despite the channel being open all this time. When
  // this event is dispatched, ServerAgent, RootNode, and descenadant nodes have
  // to be re-instantiated, probably by reloading the page.
  ERROR: 'error'
};


/**
 * A string guaranteed not to clash with any element of a data tree path, and
 * used as a key pointing to a leaf in object trees.
 * @type {string}
 * @const
 * @private
 */
mc.ServerAgent.LEAF_KEY_ = '.';


/**
 * The timeout in milliseconds for receiving response from the server to
 * 'connect' request.
 * @type {number}
 */
mc.ServerAgent.CONNECT_RESPONSE_TIMEOUT = 30 * 1000;


/**
 * The timeout in milliseconds for receiving response from the server to
 * 'reconnect' request.
 * @type {number}
 */
mc.ServerAgent.RECONNECT_RESPONSE_TIMEOUT = 30 * 1000;


/**
 * The timeout in milliseconds for receiving response from the server to
 * 'updates' message.
 * @type {number}
 */
mc.ServerAgent.UPDATES_RESPONSE_TIMEOUT = 120 * 1000;


/**
 * @param {mc.ds.Path} path A path to check.
 * @param {mc.ServerAgent.Reads} reads The reads property of an
 *     mc.ServerAgent.Transaction.
 * @return {boolean} Whether the path matches the supplied reads.
 * @private
 */
mc.ServerAgent.isPathRead_ = function(path, reads) {
  if (!path.length) {
    return mc.ServerAgent.LEAF_KEY_ in reads;
  }
  if (path[0] in reads &&
      mc.ServerAgent.isPathRead_(goog.array.slice(path, 1), reads[path[0]])) {
    return true;
  }
  if ('*' in reads &&
      mc.ServerAgent.isPathRead_(goog.array.slice(path, 1), reads['*'])) {
    return true;
  }
  return false;
};


/**
 * Logger.
 * @type {!goog.debug.Logger}
 * @protected
 */
mc.ServerAgent.prototype.logger = goog.debug.Logger.getLogger('mc.ServerAgent');


/**
 * @return {?mc.ds.RootNode} Root node of the data tree. Before initial state of
 *     the data has been loaded, returns null.
 */
mc.ServerAgent.prototype.getRootNode = function() {
  return this.rootNode_;
};


/**
 * Buffers an update before sending to the server.
 * @param {mc.ds.Path} path Path to changed value.
 * @param {*} oldValue Value before the change.
 * @param {*} newValue Value after the change.
 * @private
 */
mc.ServerAgent.prototype.buffer_ = function(path, oldValue, newValue) {
  var transaction;
  var transactionID;
  if (!goog.isNull(this.transaction_)) {
    transaction = this.transaction_;
    transactionID = this.transactionIDCounter_.toString();
  } else {
    transaction = {reads: {}, writes: []};
    transactionID = (++this.transactionIDCounter_).toString();
    this.bufferedUpdates_.set(transactionID, transaction);
  }
  var encodedPath = path.join('/');
  var historyForPath;
  if (encodedPath in this.history_) {
    historyForPath = this.history_[encodedPath];
  } else {
    historyForPath = {initial: oldValue, updates: new goog.structs.LinkedMap()};
    this.history_[encodedPath] = historyForPath;
  }
  var historyForTransaction = historyForPath.updates.get(transactionID);
  if (!goog.isDef(historyForTransaction)) {
    historyForTransaction = [];
    historyForPath.updates.set(transactionID, historyForTransaction);
  }

  transaction.writes.push(encodedPath);
  historyForTransaction.push(newValue);

  if (this.lastChangesSavedStatus_) {
    this.lastChangesSavedStatus_ = false;
    this.dispatchEvent(
        mc.ServerAgent.EventType.SAVED_STATUS_CHANGE);
    this.updatesResponseTimeoutDelay_.start();
  }
  if (goog.isNull(this.transaction_)) {
    this.startOrStopSendDelay_();
  }
};


/**
 * Handles a READ event dispatched by mc.ds.RootNode.
 * @param {mc.ds.ReadEvent} e Read event.
 * @private
 */
mc.ServerAgent.prototype.handleRead_ = function(e) {
  // READ events are only dispatched when a transaction is in progress,
  // so this.transaction_ is not null.
  goog.asserts.assert(!goog.isNull(this.transaction_));

  if (!mc.ServerAgent.isPathRead_(e.path, this.transaction_.reads)) {

    // First, for the case when e.path contains wildcards, remove any paths
    // already in reads that match this wildcard path.
    var pathsToRemove = [];
    var leafKeys = {};
    leafKeys[mc.ServerAgent.LEAF_KEY_] = null;
    mc.ds.forEach(this.transaction_.reads, e.path,
        function(value, path) {
          if (mc.ServerAgent.LEAF_KEY_ in value) {
            pathsToRemove.push(path);
          }
        }, this, leafKeys);
    goog.array.forEach(pathsToRemove, function(path) {
      mc.ds.setValueByKeys(this.transaction_.reads,
          goog.array.concat(path, mc.ServerAgent.LEAF_KEY_));
    }, this);

    mc.ds.setValueByKeys(this.transaction_.reads,
        goog.array.concat(e.path, mc.ServerAgent.LEAF_KEY_), null);
  }
};


/**
 * @return {boolean} The same as areChangesSaved, but re-evaluates the status
 *     instead of returning the cached value.
 * @private
 */
mc.ServerAgent.prototype.areChangesSaved_ = function() {
  var retval = (
      (goog.isNull(this.transaction_) || !this.transaction_.writes.length) &&
      this.bufferedUpdates_.isEmpty() && this.sentUpdates_.isEmpty() &&
      (goog.isNull(this.lastCommittedVersion_) ||
      this.lastCommittedVersion_ <= this.diskVersion_));
  this.lastChangesSavedStatus_ = retval;
  return retval;
};


/**
 * @return {boolean} Whether the changes made locally, if any, have been saved
 *     to disk on the server.
 */
mc.ServerAgent.prototype.areChangesSaved = function() {
  return this.lastChangesSavedStatus_;
};


/**
 * Starts or stops this.sendDelay_ timer depending on whether all of the
 * following holds: (1) there are buffered updates to send, (2) the first of the
 * buffered updates is made over this.version_, i.e. sentUpdates_ and
 * committedUpdates_are empty, (3) channel is open and connected/reconnected
 * message was received from the server, i.e. this.inMemoryStateToken_ is not
 * null.
 * @private
 */
mc.ServerAgent.prototype.startOrStopSendDelay_ = function() {
  if (!this.bufferedUpdates_.isEmpty() && this.sentUpdates_.isEmpty() &&
      this.committedUpdates_.isEmpty() &&
      !goog.isNull(this.inMemoryStateToken_)) {
    if (this.sendDelay_.isActive()) {
      this.logger.fine('sendDelay_ already started.');
    } else {
      this.logger.fine('Will start sendDelay_.');
    }
    this.sendDelay_.start();
  } else {
    if (!this.sendDelay_.isActive()) {
      this.logger.fine('sendDelay_ already inactive.');
    } else {
      this.logger.fine('Will stop sendDelay_.');
    }
    this.sendDelay_.stop();
  }
};


/**
 * Sends 'connect' request to the server.
 * @private
 */
mc.ServerAgent.prototype.sendConnectRequest_ = function() {
  this.connectionToken_ = goog.string.getRandomString();
  this.logger.fine('Will send \'connect\' request.');
  this.channel_.send(this.service_, {'connect': this.connectionToken_});
  this.connectResponseTimeoutDelay_.start();
};


/**
 * Sends 'reconnect' request to the server.
 * @private
 */
mc.ServerAgent.prototype.sendReconnectRequest_ = function() {
  this.logger.fine('Will send \'reconnect\' request.');
  this.channel_.send(this.service_, {
    'reconnect': [
      this.version_,
      this.connectionToken_
    ]
  });
  this.reconnectResponseTimeoutDelay_.start();
};


/**
 * Handles the channel OPEN event: either loads the state for the first time,
 * or checks for missed updates.
 * @private
 */
mc.ServerAgent.prototype.handleChannelOpen_ = function() {
  if (goog.isNull(this.rootNode_)) {
    this.sendConnectRequest_();
  } else {
    this.sendReconnectRequest_();
  }
};


/**
 * Handles the channel CLOSE event: rolls back any unsaved updates and stops
 * the response timeout timers.
 * @private
 */
mc.ServerAgent.prototype.handleChannelClose_ = function() {
  this.connectResponseTimeoutDelay_.stop();
  this.reconnectResponseTimeoutDelay_.stop();
  this.updatesResponseTimeoutDelay_.stop();
  if (!goog.isNull(this.rootNode_)) {
    this.flush_();
  }
  this.inMemoryStateToken_ = null;
};


/**
 * Rolls back a transaction.
 * @param {mc.ServerAgent.Transaction} transaction Transaction to roll back.
 * @param {string} transactionID ID of the transaction to roll back.
 * @private
 */
mc.ServerAgent.prototype.rollBackTransaction_ = function(transaction,
    transactionID) {
  goog.array.forEachRight(transaction.writes, function(encodedPath) {
    var historyForPath = this.history_[encodedPath];
    var updates = historyForPath.updates;
    var historyForTransactionNode = updates.map_.get(transactionID);
    var historyForTransaction = historyForTransactionNode.value;
    var historyForPrevTransaction = historyForTransactionNode.prev.value;
    var historyForNextTransaction = historyForTransactionNode.next.value;
    historyForTransaction.pop();
    var oldValue;
    if (historyForTransaction.length) {
      oldValue = historyForTransaction[historyForTransaction.length - 1];
    } else {
      updates.remove(transactionID);
      if (goog.isDef(historyForPrevTransaction)) {
        oldValue = historyForPrevTransaction[
            historyForPrevTransaction.length - 1];
      } else {
        oldValue = historyForPath.initial;
        if (updates.isEmpty()) {
          delete this.history_[encodedPath];
        }
      }
    }
    if (!goog.isDef(historyForNextTransaction)) {
      this.rootNode_.set(mc.ds.decodePath(encodedPath), oldValue,
          mc.ServerAgent.ROLLBACK, true);
    }
  }, this);
};


/**
 * Rolls back each transaction in bufferedUpdates_ that reads a value at one
 * of the supplied paths, and any dependent transactions (the ones reading
 * values that will change as a result of rollback of a previous transaction).
 * @param {!Object.<null>} encodedPaths An set of '/'-joined paths.
 * @return {boolean} Whether any transactions were rolled back.
 * @private
 */
mc.ServerAgent.prototype.selectivelyRollbackBufferedUpdates_ =
    function(encodedPaths) {
  var transactionsToRollback = [];
  var newBufferedUpdates = new goog.structs.LinkedMap();
  encodedPaths = goog.object.clone(encodedPaths);

  this.bufferedUpdates_.forEach(function(transaction, transactionID) {
    if (goog.object.some(encodedPaths, function(value, encodedPath) {
      return mc.ServerAgent.isPathRead_(mc.ds.decodePath(encodedPath),
          transaction.reads);
    }, this)) {
      transactionsToRollback.push(transactionID);
      goog.array.forEach(transaction.writes, function(encodedPath) {
        encodedPaths[encodedPath] = null;
      }, this);
    } else {
      // The values that are overwritten will not change from the point of view
      // of subsequent transactions.
      goog.array.forEach(transaction.writes, function(encodedPath) {
        goog.object.remove(encodedPaths, encodedPath);
      });
      newBufferedUpdates.set(transactionID, transaction);
    }
  }, this);

  if (transactionsToRollback.length) {
    goog.array.forEachRight(transactionsToRollback, function(transactionID) {
      this.rollBackTransaction_(/** @type {mc.ServerAgent.Transaction} */
          (this.bufferedUpdates_.get(transactionID)), transactionID);
    }, this);
    this.bufferedUpdates_ = newBufferedUpdates;
    return true;
  }
  return false;
};


/**
 * Sends bufferedUpdates_ to the server.
 * @private
 */
mc.ServerAgent.prototype.send_ = function() {
  // version_ should reflect the state of the data before the first of the
  // buffered updates, thus the two asserts below.
  goog.asserts.assert(this.sentUpdates_.isEmpty());
  goog.asserts.assert(this.committedUpdates_.isEmpty());
  this.logger.fine('Will send buffered updates to the server.');
  this.channel_.send(this.service_, {'updates': [
    this.connectionToken_,
    (this.version_ <= this.diskVersion_) ? null : this.inMemoryStateToken_,
    this.version_,
    this.bufferedUpdates_.map(function(transaction, transactionID) {
      // This object will have encoded paths as keys, and is used to keep track
      // of how many updates setting value at a given path have been processed.
      var updateIndexes = {};
      var writes = [];
      goog.array.forEach(transaction.writes, function(encodedPath) {
        if (encodedPath in updateIndexes) {
          updateIndexes[encodedPath]++;
        } else {
          updateIndexes[encodedPath] = 0;
        }
        var historyForPath = this.history_[encodedPath];
        var historyForTransaction = historyForPath.updates.get(transactionID);
        writes.push([encodedPath,
          historyForTransaction[updateIndexes[encodedPath]]]);
      }, this);
      return [transaction.reads, writes];
    }, this)
  ]});
  this.sentUpdates_ = this.bufferedUpdates_;
  this.bufferedUpdates_ = new goog.structs.LinkedMap();
};


/**
 * Advances this.version_ as far as is possible without encountering a version
 * absent from both this.receivedUpdates_ and this.committedUpdates_ and without
 * having to apply a server update while sentUpdates_ is not empty.
 * @return {boolean} Whether any bufferedUpdates_ were rolled back in order
 *     to be able to apply a server update.
 * @private
 */
mc.ServerAgent.prototype.advanceVersion_ = function() {
  var rollback = false;
  while (true) {
    if ((this.version_ + 1) in this.receivedUpdates_ &&
        this.sentUpdates_.isEmpty()) {
      this.version_++;
      var update = this.receivedUpdates_[this.version_];
      delete this.receivedUpdates_[this.version_];
      var rollbackSet = {};
      rollbackSet[update.encodedPath] = null;
      rollback = this.selectivelyRollbackBufferedUpdates_(rollbackSet) ||
          rollback;
      if (!(update.encodedPath in this.history_)) {
        this.rootNode_.set(mc.ds.decodePath(update.encodedPath),
            update.newValue, mc.ServerAgent.SERVER, true);
      } else {
        this.history_[update.encodedPath].initial = update.newValue;
      }
    } else {
      var firstCommittedTransaction = this.committedUpdates_.peek();
      if (!goog.isDef(firstCommittedTransaction) ||
          firstCommittedTransaction.version -
          firstCommittedTransaction.writes.length != this.version_) {
        break;
      }
      // firstCommittedTransaction.writes converted to a set to avoid
      // duplicates.
      var writesSet = {};
      goog.array.forEach(firstCommittedTransaction.writes,
          function(encodedPath) {
            writesSet[encodedPath] = null;
          }, this);
      goog.object.forEach(writesSet, function(value, encodedPath) {
        var historyForPath = this.history_[encodedPath];
        var historyForTransaction = historyForPath.updates.shift();
        if (historyForPath.updates.isEmpty()) {
          delete this.history_[encodedPath];
        } else {
          historyForPath.initial = historyForTransaction[
              historyForTransaction.length - 1];
        }
      }, this);
      this.committedUpdates_.shift();
      this.version_ += firstCommittedTransaction.writes.length;
    }
  }
  return rollback;
};


/**
 * Rolls back all updates in this.sentUpdates_ and this.bufferedUpdates_.
 * @private
 */
mc.ServerAgent.prototype.flush_ = function() {
  this.logger.fine('flush_');
  var rollback = false;
  goog.array.forEachRight(this.bufferedUpdates_.getKeys(),
      function(transactionID) {
        rollback = true;
        this.rollBackTransaction_(/** @type {mc.ServerAgent.Transaction} */
            (this.bufferedUpdates_.get(transactionID)), transactionID);
      }, this);
  goog.array.forEachRight(this.sentUpdates_.getKeys(),
      function(transactionID) {
        rollback = true;
        this.rollBackTransaction_(/** @type {mc.ServerAgent.Transaction} */
            (this.sentUpdates_.get(transactionID)), transactionID);
      }, this);
  this.bufferedUpdates_ = new goog.structs.LinkedMap();
  this.sentUpdates_ = new goog.structs.LinkedMap();
  this.startOrStopSendDelay_();
  if (rollback) {
    this.dispatchEvent(mc.ServerAgent.EventType.ROLLBACK);
  }
  if (!this.lastChangesSavedStatus_ && this.areChangesSaved_()) {
    this.dispatchEvent(mc.ServerAgent.EventType.SAVED_STATUS_CHANGE);
  }
};


/**
 * A factory function used to instantiate a root node, called when 'connected'
 * message is received from the server. The default implementation expects the
 * argument to be formatted as a tree of nested objects where at each path plus
 * a '.' key there is the value stored at that path.  Please see section
 * "Response to request to connect" in the datasource client-server
 * communication protocol for more details.
 * @param {*} payload Initial state as sent from the server.
 * @return {!mc.ds.RootNode} Root node.
 * @protected
 */
mc.ServerAgent.prototype.createRootNode = function(payload) {
  return new mc.ds.RootNode(goog.asserts.assertObject(payload));
};


/**
 * Handles 'connected' message from the server.
 * @param {*} payload Deserialized payload.
 * @private
 */
mc.ServerAgent.prototype.handleConnectedMessage_ = function(payload) {
  this.logger.fine('handleConnectedMessage_');
  if (!this.connectResponseTimeoutDelay_.isActive() ||
      this.connectionToken_ != payload['connected'][0]) {
    return;
  }
  this.connectResponseTimeoutDelay_.stop();

  var inMemoryStateToken = goog.asserts.assertString(
      payload['connected'][1]);
  var diskVersion = goog.asserts.assertNumber(payload['connected'][2]);
  goog.asserts.assert(goog.isNull(this.rootNode_));
  this.inMemoryStateToken_ = inMemoryStateToken;
  this.version_ = diskVersion;
  this.diskVersion_ = diskVersion;
  this.rootNode_ = this.createRootNode(payload['connected'][3]);

  this.eventHandler_.listen(this.rootNode_.getEventTarget(),
      mc.ds.EventType.WRITE,
      function(e) {
        var changeEvent = e.changeEvent;
        if (changeEvent.source != mc.ServerAgent.SERVER &&
            changeEvent.source != mc.ServerAgent.ROLLBACK) {
          goog.asserts.assert(this.channel_.isOpen(),
              'Tried to set value when the channel is closed.');
          this.buffer_(changeEvent.path, changeEvent.getOldValue(true),
              changeEvent.getNewValue(true));
        }
      });

  this.eventHandler_.listen(this.rootNode_.getEventTarget(),
      mc.ds.EventType.TRANSACTION_START,
      function() {
        this.transactionIDCounter_++;
        this.transaction_ = {reads: {}, writes: []};
      });

  this.eventHandler_.listen(this.rootNode_.getEventTarget(),
      mc.ds.EventType.READ, this.handleRead_);

  this.eventHandler_.listen(this.rootNode_.getEventTarget(),
      mc.ds.EventType.TRANSACTION_END,
      function() {
        if (this.transaction_.writes.length) {
          this.bufferedUpdates_.set(this.transactionIDCounter_.toString(),
              this.transaction_);
          this.startOrStopSendDelay_();
        }
        this.transaction_ = null;
      });

  this.dispatchEvent(mc.ServerAgent.EventType.INITIALIZED);

  goog.array.forEach(this.bufferedServerMessages_, function(payload) {
    this.handleMessage_(payload);
  }, this);
};


/**
 * Handles 'reconnected' message from the server.
 * @param {*} payload Deserialized payload.
 * @private
 */
mc.ServerAgent.prototype.handleReconnectedMessage_ = function(payload) {
  var connectionToken = goog.asserts.assertString(payload['reconnected'][0]);
  if (!this.reconnectResponseTimeoutDelay_.isActive() ||
      connectionToken != this.connectionToken_) {
    return;
  }
  var inMemoryStateToken = goog.asserts.assertString(payload['reconnected'][1]);
  this.inMemoryStateToken_ = inMemoryStateToken;
  this.reconnectResponseTimeoutDelay_.stop();

  goog.array.forEach(this.bufferedServerMessages_, function(payload) {
    this.handleMessage_(payload);
  }, this);

  // Because this.inMemoryStateToken_ has changed.
  this.startOrStopSendDelay_();
};


/**
 * Handles a message from the server with outcomes of updates previously sent
 * from this client.
 * @param {*} payload Deserialized payload.
 * @private
 */
mc.ServerAgent.prototype.handleOutcomesMessage_ = function(payload) {
  this.logger.fine('handleOutcomesMessage_');
  var connectionToken = goog.asserts.assertString(payload['outcomes'][0]);
  var inMemoryStateToken = goog.asserts.assertString(payload['outcomes'][1]);
  var version = goog.asserts.assertNumber(payload['outcomes'][2]);
  if (!this.sentUpdates_.getCount() ||
      version != this.version_ ||
      connectionToken != this.connectionToken_) {
    return;
  }
  this.inMemoryStateToken_ = inMemoryStateToken;
  var outcomes = goog.asserts.assertArray(payload['outcomes'][3]);
  goog.asserts.assert(outcomes.length == this.sentUpdates_.getCount());
  var rollback = false;
  // Array where values are [<transaction ID>, <transaction>] - transactions
  // that need to be added to committedUpdates_, in the order that they should
  // be added.
  var newCommittedUpdates = [];
  goog.array.forEachRight(outcomes, function(outcome) {
    var transactionID = this.sentUpdates_.head_.prev.key;
    var transaction = /** @type {mc.ServerAgent.Transaction} */
        (this.sentUpdates_.pop());
    if (goog.isNull(outcome)) {
      rollback = true;
      var writesSet = {};
      goog.array.forEach(transaction.writes, function(encodedPath) {
        writesSet[encodedPath] = null;
      }, this);
      this.selectivelyRollbackBufferedUpdates_(writesSet);
      this.rollBackTransaction_(transaction, transactionID);
    } else {
      transaction.version = goog.asserts.assertNumber(outcome);
      newCommittedUpdates.unshift([transactionID, transaction]);
    }
  }, this);
  goog.array.forEach(newCommittedUpdates, function(value) {
    this.committedUpdates_.set(value[0], value[1]);
    this.lastCommittedVersion_ = value[1].version;
  }, this);
  rollback = this.advanceVersion_() || rollback;

  if (rollback) {
    this.dispatchEvent(mc.ServerAgent.EventType.ROLLBACK);
  }
  goog.asserts.assert(!this.lastChangesSavedStatus_);
  if (this.areChangesSaved_()) {
    this.updatesResponseTimeoutDelay_.stop();
    this.dispatchEvent(mc.ServerAgent.EventType.SAVED_STATUS_CHANGE);
  }
  this.startOrStopSendDelay_();
};


/**
 * Handles a message from the server with updates not originating from this
 * client.
 * @param {*} payload Deserialized payload.
 * @private
 */
mc.ServerAgent.prototype.handleUpdatesMessage_ = function(payload) {
  this.logger.fine('handleUpdatesMessage_');
  var token = goog.asserts.assertString(payload['updates'][0]);
  if (token != this.inMemoryStateToken_) {
    return;
  }
  var version = goog.asserts.assertNumber(payload['updates'][1]);
  var updates = goog.asserts.assertArray(payload['updates'][2]);
  var startingVersion = version - updates.length;
  goog.array.forEach(updates, function(update, index) {
    var updateVersion = startingVersion + index + 1;
    if (updateVersion > this.version_) {
      var encodedPath = goog.asserts.assertString(update[0]);
      var newValue = update[1];
      this.receivedUpdates_[updateVersion] =
          {encodedPath: encodedPath, newValue: newValue};
    }
  }, this);
  var rollback = this.advanceVersion_();
  if (rollback) {
    this.dispatchEvent(mc.ServerAgent.EventType.ROLLBACK);
  }
  // Condition in startOrStopSendDelay_ depends on committedUpdates_, which
  // can be modified by advanceVersion_.
  this.startOrStopSendDelay_();
};


/**
 * Handles a 'saved' notification from the server.
 * @param {*} payload Deserialized payload.
 * @private
 */
mc.ServerAgent.prototype.handleSavedMessage_ = function(payload) {
  this.logger.fine('handleSavedMessage_');
  var token = goog.asserts.assertString(payload['saved'][0]);
  if (token != this.inMemoryStateToken_) {
    return;
  }
  var version = goog.asserts.assertNumber(payload['saved'][1]);
  if (version > this.diskVersion_) {
    this.diskVersion_ = version;
    if (!this.lastChangesSavedStatus_ && this.areChangesSaved_()) {
      this.updatesResponseTimeoutDelay_.stop();
      this.dispatchEvent(mc.ServerAgent.EventType.SAVED_STATUS_CHANGE);
    }
  }
};


/**
 * Handles an error notification from the server.
 * @param {*} payload Deserialized payload.
 * @private
 */
mc.ServerAgent.prototype.handleErrorMessage_ = function(payload) {
  this.logger.fine('handleErrorMessage_');
  this.dispatchEvent(mc.ServerAgent.EventType.ERROR);
};


/**
 * Handles a message from the server.
 * @param {*} payload Deserialized payload.
 * @private
 */
mc.ServerAgent.prototype.handleMessage_ = function(payload) {
  goog.asserts.assertObject(payload);
  if ('connected' in payload) {
    this.handleConnectedMessage_(payload);
  } else if ('reconnected' in payload) {
    this.handleReconnectedMessage_(payload);
  } else if ('error' in payload) {
    this.handleErrorMessage_(payload);
  } else {
    if (goog.isNull(this.inMemoryStateToken_)) {
      this.bufferedServerMessages_.push(payload);
    } else {
      if ('outcomes' in payload) {
        this.handleOutcomesMessage_(payload);
      } else if ('updates' in payload) {
        this.handleUpdatesMessage_(payload);
      } else if ('saved' in payload) {
        this.handleSavedMessage_(payload);
      } else {
        goog.asserts.fail();
      }
    }
  }
};


/**
 * A callback for when timeouts for server responses expire.
 * @private
 */
mc.ServerAgent.prototype.responseTimeoutCallback_ = function() {
  this.logger.warning('Response from server timed out.');
  this.dispatchEvent(mc.ServerAgent.EventType.ERROR);
};


/** @inheritDoc */
mc.ServerAgent.prototype.disposeInternal = function() {
  goog.base(this, 'disposeInternal');
  if (this.channel_.isOpen()) {
    this.logger.fine('Will send disconnect message.');
    this.channel_.send(this.service_, {'disconnect': null});
  }
  this.channel_.unregisterService(this.service_);
};
