// Copyright 2012 MightyClient Authors. All Rights Reserved.
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//      http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS-IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

/**
 * @fileoverview The script for client page.
 */

goog.require('goog.debug.DivConsole');
goog.require('mc.GaeChannel');
goog.require('mc.ServerAgent');


var rootNode;
var serverAgent;


var main = function(datasourceID) {
  var divConsole = new goog.debug.DivConsole(
      document.getElementById('console'));
  divConsole.setCapturing(true);


  var clientID = goog.string.getRandomString();
  var channel = new mc.GaeChannel(clientID, new goog.net.XhrIo());
  channel.open();
  serverAgent = new mc.ServerAgent(channel, 'demo/' + datasourceID);
  var rollback;


  /**
   * Either enqueues callback or calls is immediately, depending on existence of
   * header frame and the value parent['header']['queue'] (not null means
   * buffering is on).
   * @param {!Function} callback Callback.
   * @return {boolean} Whether callback was enqueued.
   */
  var maybeEnqueue = function(callback) {
    if (parent['header'] && !goog.isNull(parent['header']['queue'])) {
      parent['header']['queue'].push(callback);
      parent['header'].document.getElementById('hold-normal').style.display =
          'none';
      parent['header'].document.getElementById('queue').innerHTML =
          parent['header']['queue'].length +
          ' callback(s) enqueued. Uncheck to fire.';
      return true;
    } else {
      callback();
      return false;
    }
  };


  /**
   * Runs the callback in transaction depending on existence of header frame and
   * the value of 'use transactions' checkbox. If no header frame, uses
   * transaction.
   * @param {!Function} callback Callback.
   */
  var maybeInTransaction = function(callback) {
    if (parent['header'] &&
        !parent['header'].document.getElementById('transaction').checked) {
      callback();
    } else {
      rootNode.inTransaction(callback);
    }
  };


  var updateCheckboxesEnabled = function() {
    goog.dom.classes.enable(document.getElementById('a'),
        'goog-checkbox-disabled', !channel.isOpen());
    goog.dom.classes.enable(document.getElementById('b'),
        'goog-checkbox-disabled', !channel.isOpen() ||
        !goog.dom.classes.has(document.getElementById('a'),
        'goog-checkbox-checked'));
  };


  var updateCheckboxes = function() {
    goog.dom.classes.enable(document.getElementById('a'),
        'goog-checkbox-checked', rootNode.get(['a'], false));
    goog.dom.classes.enable(document.getElementById('b'),
        'goog-checkbox-checked', rootNode.get(['b'], false));
    updateCheckboxesEnabled();
  };


  goog.events.listen(document.getElementById('pa'), goog.events.EventType.CLICK,
      function() {
        maybeEnqueue(function() {
          maybeInTransaction(function() {
            // The reason why we need transaction here is that we only want to
            // set A to false if we have previously set B to false. The order is
            // important because unlike usual database transactions, state has
            // to be consistent after each write, not just at the end.
            if (goog.dom.classes.has(document.getElementById('a'),
                'goog-checkbox-disabled')) {
              return;
            }
            var newValue = !rootNode.get(['a']);
            // "&& rootNode.get(['b'])" is optional here.
            if (!newValue && rootNode.get(['b'])) {
              rootNode.set(['b'], false);
            }
            rootNode.set(['a'], newValue);
          });
          updateCheckboxes();
        });
      }, true);


  goog.events.listen(document.getElementById('pb'), goog.events.EventType.CLICK,
      function() {
        maybeEnqueue(function() {
          maybeInTransaction(function() {
            if (goog.dom.classes.has(document.getElementById('b'),
                'goog-checkbox-disabled')) {
              return;
            }
            var newValue = !rootNode.get(['b']);
            if (newValue) {
              // This ensures that the transaction will only be applied on the
              // other client(s) if there value A is the same - i.e. true.
              // The reason we are sure that it's true on this client is that
              // the checkbox B is disabled whenever A is unchecked.
              var a = rootNode.get(['a']);
              // goog.asserts.assert(rootNode.get(['a'])) would not have worked:
              // compiler would assume that rootNode.get(['a']) has no side
              // effects.
              goog.asserts.assert(a);
            }
            rootNode.set(['b'], newValue);
          });
          updateCheckboxes();
        });
      }, true);


  goog.events.listen(serverAgent,
      mc.ServerAgent.EventType.INITIALIZED,
      function() {
        rootNode = serverAgent.getRootNode();
        rootNode.listen(['*'], updateCheckboxes);
        updateCheckboxes();
        document.getElementById('status').innerHTML = 'Initialized.';
      });


  goog.events.listen(serverAgent,
      mc.ServerAgent.EventType.ROLLBACK,
      function() {
        rollback = true;
      });


  goog.events.listen(serverAgent,
      mc.ServerAgent.EventType.SAVED_STATUS_CHANGE,
      function() {
        if (!serverAgent.areChangesSaved()) {
          rollback = false;
          document.getElementById('status').innerHTML = 'Saving changes...';
        } else {
          if (rollback) {
            document.getElementById('status').innerHTML =
                'Some changes were rolled back.';
          } else {
            document.getElementById('status').innerHTML =
                'All changes have been saved.';
          }
        }
      });


  goog.events.listen(serverAgent,
      mc.ServerAgent.EventType.ERROR,
      function() {
        document.getElementById('main').innerHTML =
            '<p>Server error, try <a href="">reloading the page</a>.</p>';
      });


  goog.events.listen(channel,
      mc.GaeChannel.EventType.OPEN_FAILURE,
      function() {
        document.getElementById('no-connection').style.display = 'block';
        document.getElementById('reconnecting').style.display = 'none';
      });


  goog.events.listen(channel,
      mc.Channel.EventType.OPEN,
      function() {
        updateCheckboxesEnabled();
        document.getElementById('no-connection').style.display = 'none';
        document.getElementById('reconnecting').style.display = 'none';
      });


  goog.events.listen(channel,
      mc.Channel.EventType.CLOSE,
      function() {
        updateCheckboxesEnabled();
      });


  goog.events.listen(document.getElementById('reconnect'),
      goog.events.EventType.CLICK,
      function() {
        channel.open();
        document.getElementById('no-connection').style.display = 'none';
        document.getElementById('reconnecting').style.display = 'block';
      }, true);


  goog.events.listen(document.getElementById('clear-console'),
      goog.events.EventType.CLICK,
      function() {
        divConsole.clear();
      }, true);
};
goog.exportSymbol('main', main);
