# Copyright 2012 MightyClient Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS-IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""Implements AbstractStorage using GAE datastore.
"""

from collections import defaultdict
import json
import logging
import sys

from google.appengine.api import lib_config
from google.appengine.ext import ndb

from lib.server.mc.abstract_storage import AbstractStorage
from lib.server.mc import json_utils


class _ConfigDefaults(object):
  """Configurable constants.

  Attributes:
      NAMESPACE: a string to be prepended to datastore kinds used by this
            module.
  """

  NAMESPACE = 'mc_ds_'


_config = lib_config.register('mc_datastore_storage', _ConfigDefaults.__dict__)


# A factory function for infinitely nested defaultdicts, takes no arguments.
get_nested_defaultdict = lambda: defaultdict(get_nested_defaultdict)


class _Version(ndb.Model):
  """A model for entities that store the current version of the data.

  Properties:
      version: version number.
  """

  version = ndb.IntegerProperty('n', indexed=False)

  _use_cache = False
  _use_memcache = False

  @classmethod
  def _get_kind(cls):
    """Overrides the superclass static method.
    """
    return _config.NAMESPACE + 'N'


class _Value(ndb.Model):
  """A model for entities that store the current value at a given path.

  Properties:
      value: the JSON-encoded value.
  """

  value = ndb.TextProperty('v')

  _use_cache = False
  _use_memcache = False

  @classmethod
  def _get_kind(cls):
    """Overrides the superclass static method.
    """
    return _config.NAMESPACE + 'V'


class DatastoreStorage(AbstractStorage):
  """Implements AbstractStorage using GAE datastore.
  """
  def __init__(self, datastore_key):
    """Args:
        datastore_key: a datastore key used as parent key for all entities
            containing data for the datasource.
    """
    self.__datastore_key = datastore_key


  def get_datasource_key(self):
    """Implements the abstract method of the superclass.
    """
    return str(hash(self.__datastore_key))


  def get_disk_version(self):
    """Implements the abstract method of the superclass.
    """
    version_key = ndb.Key(_Version, 1, parent=self.__datastore_key)
    version_entity = version_key.get()
    if version_entity is None:
      return 0
    else:
      return version_entity.version


  def get_disk_state(self):
    """Implements the abstract method of the superclass.
    """
    version_key = ndb.Key(_Version, 1, parent=self.__datastore_key)
    values_query = _Value.query(ancestor=self.__datastore_key)
    def txn():
      return version_key.get(), values_query.fetch()
    version_entity, value_entities = ndb.transaction(txn)
    if version_entity is None:
      version = 0
      data = {}
    else:
      version = version_entity.version
      data = get_nested_defaultdict()
      for valueEntity in value_entities:
        if valueEntity.key.id() == '/': path = []
        else: path = valueEntity.key.id().split('/')
        data_el = data
        for path_el in path: data_el = data_el[path_el]
        data_el['.'] = json_utils.loads(valueEntity.value)
    return version, json_utils.dumps(data)


  def save_updates(self, updates, version):
    """Implements the abstract method of the superclass.
    """
    def txn():
      version_key = ndb.Key(_Version, 1, parent=self.__datastore_key)
      version_entity = version_key.get()
      if version_entity is None:
        version_entity = _Version(version=0, id=1,
            parent=self.__datastore_key)
      if version_entity.version != version - len(updates):
        return False
      version_entity.version = version
      entities_put = {version_entity.key: version_entity}
      keys_delete = set()
      for update in updates:
        encoded_path, value = update
        if len(encoded_path) == 0: id = '/'
        else: id = encoded_path
        if value is not None:
          entity = _Value(id=id, parent=self.__datastore_key,
              value=json_utils.dumps(value))
          entities_put[entity.key] = entity
          if entity.key in keys_delete:
            keys_delete.remove(entity.key)
        else:
          key = ndb.Key(_Value, id, parent=self.__datastore_key)
          if key in entities_put:
            del entities_put[key]
          keys_delete.add(key)
      ndb.put_multi(entities_put.values())
      ndb.delete_multi(list(keys_delete))
      return True

    # Should not throw
    # ndb.google_imports.datastore_errors.TransactionFailedError because we
    # make sure to run one transaction per entity group at a time.
    try:
      retval = ndb.transaction(txn, retries=0)
    except Exception:
      logging.error('Error while saving data.', exc_info=sys.exc_info())
      return False
    return retval
