#!/usr/bin/python
#
# Copyright 2012 MightyClient Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS-IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import json
import unittest

from google.appengine.datastore import datastore_stub_util
from google.appengine.ext import ndb
from google.appengine.ext import testbed

import datastore_storage


class TestCase(unittest.TestCase):

  def setUp(self):
    self.testbed = testbed.Testbed()
    self.testbed.activate()
    self.policy = datastore_stub_util.PseudoRandomHRConsistencyPolicy(
        probability=0)
    self.testbed.init_datastore_v3_stub(consistency_policy=self.policy)
    self.testbed.init_memcache_stub()
    self.testbed.init_channel_stub()
    self.datastore_key = ndb.Key('datasource', 1)
    self.storage = datastore_storage.DatastoreStorage(self.datastore_key)


  def tearDown(self):
    self.testbed.deactivate()


  def test_get_disk_version_and_state(self):
    self.assertEquals(0, self.storage.get_disk_version())
    self.assertEquals((0, '{}'), self.storage.get_disk_state())
    datastore_storage._Version(parent=self.datastore_key, id=1, version=3).put()
    datastore_storage._Value(parent=self.datastore_key, id='/',
        value='"v1"').put()
    datastore_storage._Value(parent=self.datastore_key, id='a0/a1',
        value='"v2"').put()
    self.assertEquals(3, self.storage.get_disk_version())
    version, disk_state = self.storage.get_disk_state()
    self.assertEquals(3, version)
    self.assertEquals({
      '.': 'v1',
      'a0': {'a1': {'.': 'v2'}}
    }, json.loads(disk_state))


  def test_save_updates(self):
    retval = self.storage.save_updates([
      ['', 'v1'],
      ['b', None],
      ['b', 'v2'],
      ['d', 'v'],
      ['d', None],
    ], 5)
    self.assertEquals(5, ndb.Key(datastore_storage._Version, 1,
        parent=self.datastore_key).get().version)
    self.assertEquals('"v1"', ndb.Key(datastore_storage._Value, '/',
        parent=self.datastore_key).get().value)
    self.assertEquals('"v2"', ndb.Key(datastore_storage._Value, 'b',
        parent=self.datastore_key).get().value)
    self.assertIs(None, ndb.Key(datastore_storage._Value, 'd',
        parent=self.datastore_key).get())
    self.assertIs(retval, True)

    # Try supplying incorrect version.
    retval = self.storage.save_updates([
      ['b', 'v2'],
    ], 5)
    self.assertIs(retval, False)
    self.assertEquals(5, ndb.Key(datastore_storage._Version, 1,
        parent=self.datastore_key).get().version)
    self.assertEquals('"v2"', ndb.Key(datastore_storage._Value, 'b',
        parent=self.datastore_key).get().value)


    # Modify what is stored on disk for paths '' and 'b' to make sure
    # the first two updates do not get saved again.
    datastore_storage._Value(parent=self.datastore_key, id='/',
        value='"other_v1"').put()
    datastore_storage._Value(parent=self.datastore_key, id='b',
        value='"other_v2"').put()
    datastore_storage._Value(parent=self.datastore_key, id='c',
        value='"other_v3"').put()
    self.storage.save_updates([
      ['c',  None],
    ], 6)
    self.assertEquals(6, ndb.Key(datastore_storage._Version, 1,
        parent=self.datastore_key).get().version)
    self.assertEquals('"other_v1"', ndb.Key(datastore_storage._Value, '/',
        parent=self.datastore_key).get().value)
    self.assertEquals('"other_v2"', ndb.Key(datastore_storage._Value, 'b',
        parent=self.datastore_key).get().value)
    self.assertIs(None, ndb.Key(datastore_storage._Value, 'c',
        parent=self.datastore_key).get())
