# Copyright 2012 MightyClient Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS-IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""Defines an abstract base class for objects to which datasource delegates the
tasks of persistently storing and retrieving the data.
"""

class AbstractStorage(object):
  """An abstract class for objects responsible for persisting the data.
  """

  def get_datasource_key(self):
    """Should return a string that is unique to the set of data that datasource
    exposes, and which the datasource module uses to build memcache keys.
    """
    raise NotImplementedError()


  def get_disk_version(self):
    """Reads the current version of the data on disk.

    Returns: version number.
    """
    raise NotImplementedError()


  def get_disk_state(self):
    """Reads the current state of the data on disk.

    Returns:
        A tuple (<version>, <state>), where <state> is a JSON-enconded string
        that contains the state of the data on disk. Please see section
        "Response to request to connect" in the datasource client-server
        communication protocol for more details on the format of <state>.
    """
    raise NotImplementedError()


  def save_updates(self, updates, version):
    """Saves supplied updates to disk.

    Should save only if the original version of the data is supplied version
    minus len(updates).

    Calls to this method are made one at a time per datasource: while saving
    is in progress, a memcache-based lock prevents additional calls except when
    the in-memory state of the datasource is evicted from memcache.

    Args:
        updates: a list of updates in the order they were applied, each update
            incrementing version of the data by 1, in the format
            [<'/'-joined path>, <new value>].
        version: the version after the last of the updates.

    Returns:
        A boolean indicating whether saving succeeded.
    """
    raise NotImplementedError()
