# Copyright 2012 MightyClient Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS-IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""GAE side of two-way communication channel between the client and the server.

For client-server communication protocol, see wiki/protocol_channel.md.
"""


from collections import defaultdict, OrderedDict
import datetime
import logging
import random
import re
import string
import sys
import threading
import time

from google.appengine.api import channel
from google.appengine.api import lib_config
from google.appengine.api import memcache
import webapp2

import json_utils
import memcache_utils
from service import SerializedPayload, Service


class _ConfigDefaults(object):
  """Configurable constants.

  Sample appengine_config.py:
  --------
  from mc.service import Service


  class EchoService(Service):
    def handle_message(self, payload):
      # Echoes the message to the client who sent it.
      self.send(payload)


  class HelloService(Service):
    def handle_message(self, payload):
      # When one client sends another's ID, says hello to that other client.
      self.send('Hello!', to=payload)


  mc_channel_SERVICES = [
    ('echo', EchoService),
    ('hello', HelloService),
  ]
  --------

  Attributes:
      SERVICES: a list of touples (regex, service) where regex is a regular
          expression used to match service name, and service is a subclass of
          service.Service. ^ and $ are added to the regular expession if absent.
          When an incoming message is received, the first matching service in
          the list will be instantiated, and the message will be passed to its
          handle_message method. None of the regular expressions should match
          the value of _SYSTEM_SERVICE constant.

      NAMESPACE: a string to be prepended to memcache keys used by this module.
  """

  SERVICES = []

  NAMESPACE = 'mc_channel_'


_config = lib_config.register('mc_channel', _ConfigDefaults.__dict__)


_logger = logging.getLogger(__name__)


# Name of the service used internally by this module.
_SYSTEM_SERVICE = 'system'


# How long in seconds to keep in memcache the messages to be picked up by a
# client. A message can be left for pickup either because it is too large to
# send using GAE Channel or when the client uses polling to receive messages
# originating from the server. We choose this value based on assumption that
# polling frequency is at least once a minute.
_PICKUP_TIMEOUT = 3 * 60


# The lifespan of a connection token (GAE Channel token or, for polling,
# internal one), used for expiration of memcache items.
_CHANNEL_LIFESPAN = 120 * 60


# A lock used when accessing _open_requests.
_open_requests_lock = threading.Lock()


# A set of client IDs for which this server instance is currently handling
# forward channel requests.
_open_requests = set()


# A lock used when accessing _outgoing_messages.
_outgoing_messages_lock = threading.Lock()


# Messages to be sent in responces to forward channel requests. A two-level
# defaultdict, first-level keys are client IDs, second-level keys are service
# names, and values are lists of outgoing messages (messages are in serialized
# form).
_outgoing_messages = defaultdict(lambda: defaultdict(list))


# Alias for time.time that allows using mock clock in unit tests.
_gettime = time.time


# Data stored in memcache (all keys are prefixed with NAMESPACE):
#
#     pickup_<pickup ID>: A message waiting to be picked up by a client using a
#         forward channel request. Used when the message is large or when the
#         client uses polling. A dictionary {'service': <service over which the
#         message was sent>, 'message': <JSON-encoded message passed through
#         memcache_utils.maybe_split>}. Expiration time: _PICKUP_TIMEOUT.
#
#     connected_<client ID>: this memcache item is used to indicate whether the
#         client gets server messages by GAE channel or by polling. If this item
#         is absent for a given client ID, we assume that the client is not
#         connected (if it is present, this does not necessarily mean that the
#         client is connected). The value is a dictionary {'poll': <a boolean
#         indicating whether the client gets updates by polling>, 'token'
#         (present if poll=True): <an internally generated connection token (not
#         GAE channel token)>}. Expiration time: _CHANNEL_LIFESPAN.
#
#     poll_<connection token>: a list of pickup IDs of items waiting to be
#         picked up by a client that uses polling. Expiration time:
#         _CHANNEL_LIFESPAN.


def _get_random_string():
  """Returns a random string.
  """
  return ''.join(random.choice(string.ascii_letters +
        string.digits) for x in range(12))


def _store_message_for_pickup(service, message):
  """Stores a message in memcache to be picked up later by _pickup_message.

  Args:
      service: service over which the message was sent.
      message: JSON-encoded message.

  Returns:
      A pickup ID, a unique string randomly generated for each stored message.
  """
  pickup_id = _get_random_string()
  pickup_key = _config.NAMESPACE + 'pickup_' + pickup_id;
  result_set_pickup = memcache.set(pickup_key, {'service': service,
      'message': memcache_utils.maybe_split(message, time=_PICKUP_TIMEOUT)},
      time=_PICKUP_TIMEOUT)
  if not result_set_pickup:
    raise memcache_utils.MemcacheWriteFailure()
  return pickup_id


def _pickup_message(pickup_id):
  """Picks up a message stored in memcache.

  Args:
      pickup_id: ID returnd by _store_message_for_pickup.

  Returns:
      The touple (service, message) originally passed to
      _store_message_for_pickup.
  """
  memcache_key = _config.NAMESPACE + 'pickup_' + pickup_id
  memcache_value = memcache.get(memcache_key)
  if memcache_value is None: raise memcache_utils.MemcacheReadFailure()
  service = memcache_value['service']
  message = memcache_utils.maybe_merge(memcache_value['message'])
  return service, message


def _encode_messages(messages):
  """Encodes server-to-client messages into a string that can be sent to the
  client.

  Args:
      messages: A dictionary whose keys are service names, and values are lists
          of messages. Each message is an already JSON-encoded payload.

  Returns:
      Encoded string.
  """
  def ser_messages(messages):
    for service, messages_for_service in messages.iteritems():
      yield (json_utils.dumps(service) + ':[' +
          ','.join(messages_for_service) + ']')

  return '{' + ','.join(ser_messages(messages)) + '}'


def _send_message_by_polling(token, pickup_id):
  """Adds a pickup ID to a list stored in memcache under key 'poll_...' for the
  next time the client polls the server for new messages.

  Args:
      token: internally generated connection token.
      pickup_id: pickup ID as returned by _store_message_for_pickup.
  """
  memcache_client = memcache.Client()
  memcache_key = _config.NAMESPACE + 'poll_' + token
  attempt, max_attempts = 1, 1 + 10
  while attempt <= max_attempts:
    memcache_value = memcache_client.get(memcache_key, for_cas=True)
    if memcache_value is None:
      memcache_client.add(memcache_key, [], time=_CHANNEL_LIFESPAN)
      continue
    memcache_value.append(pickup_id)
    # We use the _async version because unlike the others it has different
    # return value for when cas failed because the value has changed and when
    # it failed because of RPC error.
    results_cas = memcache_client.cas_multi_async(
        {memcache_key: memcache_value},
        time=_PICKUP_TIMEOUT).get_result()
    if (results_cas is None or results_cas[memcache_key] == memcache.ERROR):
      raise memcache_utils.MemcacheWriteFailure()
    elif (results_cas[memcache_key] == memcache.STORED):
      break
    attempt += 1
  else:
    raise memcache_utils.MemcacheContentionError()


def _get_service_instance(service, client_id, request):
  """Creates a Service instance.

  Args:
      service: service name.
      client_id: client ID.
      request: forward channel request.

  Returns:
      A new instance of Service or None of no services matched supplied name.
  """
  for pattern, service_cls in _config.SERVICES:
    if not pattern.startswith('^'):
      pattern = '^' + pattern
    if not pattern.endswith('$'):
      pattern += '$'
    if re.match(pattern, service):
      return service_cls(service, client_id, request)
  _logger.warning('No services match service name "%s".'%service)


class _SystemService(Service):
  """A service that handles messages internal to this module.
  """

  def _handle_get_token_message(self, payload):
    """Handles a getToken message from the client.

    Args:
        payload: message sent by the client.
    """
    force_polling = payload['getToken']
    connected_key = _config.NAMESPACE + 'connected_' + self.client_id
    token = None
    if not force_polling:
      try:
        token = channel.create_channel(self.client_id)
      except:
        # Most likely this is apiproxy_errors.OverQuotaError.
        _logger.warning(
            'Exception when creating channel, will fall back to polling.',
            exc_info=sys.exc_info())
      else:
        connected_value = {'poll': False}
    if token is None:
      token = _get_random_string()
      connected_value = {'poll': True, 'token': token}
    self.send({'opened': {'token': token, 'poll': connected_value['poll']}})
    result_set_connected = memcache.set(connected_key, connected_value,
        time=_CHANNEL_LIFESPAN)
    if not result_set_connected:
      raise memcache_utils.MemcacheWriteFailure()


  def _handle_pickup_message(self, payload):
    """Handles a 'pickup' message from the client.

    Args:
        payload: message sent by the client.
    """
    pickup_id = payload['pickup']
    service, message = _pickup_message(pickup_id)
    self.send(SerializedPayload(message), service=service)


  def _handle_poll_message(self, payload):
    """Handles a 'poll' message from the client.

    Args:
        payload: message sent by the client.
    """
    token = payload['poll']
    memcache_client = memcache.Client()
    memcache_key = _config.NAMESPACE + 'poll_' + token

    attempt, max_attempts = 1, 1 + 10
    while attempt <= max_attempts:
      memcache_value = memcache_client.get(memcache_key, for_cas=True)
      if memcache_value is None:
        break
      # We use the _async version because unlike the others it has different
      # return value for when cas failed because the value has changed and
      # when it failed because of RPC error.
      results_cas = memcache_client.cas_multi_async(
          {memcache_key: []},
          time=_CHANNEL_LIFESPAN).get_result()
      if (results_cas is None or results_cas[memcache_key] == memcache.ERROR):
        raise memcache_utils.MemcacheWriteFailure()
      elif (results_cas[memcache_key] == memcache.STORED):
        break
      attempt += 1
    else:
      raise memcache_utils.MemcacheContentionError()

    if memcache_value is not None:
      for pickup_id in memcache_value:
        service, message = _pickup_message(pickup_id)
        self.send(SerializedPayload(message), service=service)


  def handle_message(self, payload):
    """Implements the abstract method of the superclass.
    """
    assert isinstance(payload, dict)
    if 'getToken' in payload:
      _logger.debug('Will handle "getToken" message')
      self._handle_get_token_message(payload)
    elif 'pickup' in payload:
      _logger.debug('Will handle "pickup" message')
      self._handle_pickup_message(payload)
    elif 'poll' in payload:
      _logger.debug('Will handle "poll" message')
      self._handle_poll_message(payload)
    else: _logger.error('Unrecognized system service message.')


for pattern, service_cls in _config.SERVICES:
  if re.match(pattern, _SYSTEM_SERVICE):
    assert False, ('Service name pattern "%s" matches the name of the'
        ' system service ("%s").'%(pattern, _SYSTEM_SERVICE))
_config.SERVICES.append((_SYSTEM_SERVICE, _SystemService))


def send(payload, service, to):
  """Sends a message to one or more clients.

  Args:
        payload: payload that can be serialized to JSON or a SerializedPayload.
        service: name of the service.
        to: ID of the client to which the message should be sent, or a list of
            IDs to send the same message to multiple clients.
  """
  if isinstance(to, basestring): to = [to]

  if isinstance(payload, SerializedPayload):
    ser_payload = payload.payload
  else:
    ser_payload = json_utils.dumps(payload)

  # For clients for whom the server is currently processing fwd channel
  # requests, add the message to _outgoing_messages, and add other clients to
  # client_ids.
  client_ids = []
  for client_id in to:
    with _open_requests_lock, _outgoing_messages_lock:
      if client_id in _open_requests:
        _outgoing_messages[client_id][service].append(ser_payload)
      else:
        client_ids.append(client_id)

  # If client_ids is empty, we're done.
  if not client_ids: return

  # Define a function that lazily stores the message for pickup.
  pickup_id_list = [None]
  def get_pickup_id():
    if pickup_id_list[0] is None:
      pickup_id_list[0] = _store_message_for_pickup(service, ser_payload)
    return pickup_id_list[0]

  # Retrieve presence information from memcache.
  results_presence = memcache.get_multi(client_ids,
      key_prefix = _config.NAMESPACE + 'connected_')

  # Depending on whether the client is connected and how it is connected (GAE
  # channel or polling), send message by GAE
  # channel, or leave message for pickup through polling.
  for client_id in client_ids:
    presence = results_presence.get(client_id)
    if presence is not None:
      if not presence['poll']:
        # Send the message or pickup instruction by GAE Channel.
        channel_message = _encode_messages({service: [ser_payload]})
        if len(channel_message) >= 32 * 1024:
          channel_message = _encode_messages({_SYSTEM_SERVICE:
              [json_utils.dumps({'pickup': get_pickup_id()})]})
        try:
          channel.send_message(client_id, channel_message)
        except:
          _logger.debug('Exception when sending message via Channel API',
              exc_info=sys.exc_info())
      else:
        # Leave message for pickup through polling.
        _send_message_by_polling(presence['token'], get_pickup_id())


class ForwardChannelHandler(webapp2.RequestHandler):
  """Request handler class for forward channel requests.
  """

  def post(self):
    """Handles forward channel requests made via XHR.
    """
    client_id = self.request.get('cid', default_value=None)
    if client_id is None: webapp2.abort(400)
    messages_str = self.request.get('m')
    response_str = self.__handle_incoming_messages(client_id, messages_str)
    self.response.write(response_str)


  def get(self):
    """Handles forward channel requests made via Jsonp.
    """
    client_id = self.request.get('cid', default_value=None)
    if client_id is None: webapp2.abort(400)
    messages_str = self.request.get('m')

    response_str = self.__handle_incoming_messages(client_id, messages_str)
    callback = self.request.get('callback', default_value=None)
    if callback is None: webapp2.abort(400)
    self.response.headers['Content-Type'] = 'text/javascript; charset=UTF-8'
    self.response.write(callback + '(' + response_str + ')')


  def __handle_incoming_messages(self, client_id, messages_str):
    """Handles messages that were sent in forward channel request.

    Args:
        client_id: ID of the client that sent the request.
        messages_str: a string containing encoded messages.

    Returns:
        A json string to be sent in the response or an empty string if nothing
        to send.
    """
    try: messages = json_utils.loads(messages_str)
    except: webapp2.abort(400)

    _logger.debug('Started handling fwd channel request from client {},'
        ' _open_requests == {},'
        ' _outgoing_messages.keys() == {}.'.format(
        client_id, _open_requests, _outgoing_messages.keys()))

    with _open_requests_lock:
      # For simplicity, make sure that we are always handling one request from
      # a particular client at a time on this server instance.
      if client_id in _open_requests: webapp2.abort(400)
      _open_requests.add(client_id)

    # Use try/finally to make sure that even if an exception is raised,
    # client_id is removed from _open_requests.
    try:
      # Call handlers of incoming messages.
      for service, messages_for_service in messages.iteritems():
        if not isinstance(messages_for_service, list): webapp2.abort(400)
        service_instance = _get_service_instance(service, client_id,
            self.request)
        if service_instance is not None:
          for message in messages_for_service:
            service_instance.handle_message(message)

    finally:
      # Pick up all messages for this client from _outgoing_messages.
      with _outgoing_messages_lock, _open_requests_lock:
        _open_requests.remove(client_id)
        response_messages = _outgoing_messages.pop(client_id, None);

      _logger.debug('Finished handling fwd channel request from client {},'
          ' _open_requests == {},'
          ' _outgoing_messages.keys() == {}.'.format(
          client_id, _open_requests, _outgoing_messages.keys()))

    # Send messages in the response to forward channel request.
    if response_messages:
      return _encode_messages(response_messages)
    return ''
