# Copyright 2012 MightyClient Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS-IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Defines base class for datasource services.

For client-server communication protocol, see wiki/protocol_datasource.md.
"""

from collections import defaultdict
import logging
import os
import random
import sys
import string
import time

from google.appengine.api import lib_config
from google.appengine.api import memcache

import json_utils
import memcache_utils
from service import SerializedPayload, Service


class _ConfigDefaults(object):
  """Configurable constants.

  Attributes:
      NAMESPACE: a string to be prepended to memcache keys used by this module.

      STATE_EXPIRATION: expiration time in seconds for memcache items that
          contain the in-memory state.

      UPDATES_EXPIRATION: expiration time in seconds for memcache items that
          contain incremental changes applied to the data. For a client to
          be able to reconnect without reloading the state, all updates made
          while the client was not connected have to be present in memcache.
  """

  NAMESPACE = 'mc_ds_'

  STATE_EXPIRATION = 36 * 60 * 60

  UPDATES_EXPIRATION = 36 * 60 * 60


_config = lib_config.register('mc_datasource', _ConfigDefaults.__dict__)


_logger = logging.getLogger(__name__)


# Timeout in seconds for saving data to disk. After this timeout expires,
# transaction is assumed to have failed.
_SAVE_TIMEOUT = float(30)


# A delay in seconds that determines frequency at which we poll memcache to
# check if a transaction which saves data to disk has completed.
_SAVE_POLL_DELAY = float(0.5)


# How long in seconds to keep sending updates to a client after it has last
# connected or reconnected. This timeout ensures that if a client fails to send
# 'disconnect' message, we do not end up forever trying to send server updates
# to it. Set to be equal to lifespan of a channel token (GAE Channel token or
# polling token).
_CONNECTION_TIMEOUT = float(2 * 60 * 60)


# Number of retries, not including the first attempt, to compare-and-set the
# state in memcache.
_STATE_CAS_RETRIES = 9


# A string key guaranteed not to clash with any element of a data tree path.
_LEAF_KEY = '.'


# Number of transactions that will trigger handling of client updates with
# __large_update.
_LARGE_UPDATE_TRANSACTIONS = 10


# Number of modified values that will trigger handling of client updates with
# __large_update.
_LARGE_UPDATE_VALUES = 1000


# Data stored in memcache (all keys are prefixed with
# '<NAMESPACE><datasource key>_'):
#
#     state: in-memory state of the datasource, a JSON-encoded structure as
#         follows:
#
#         {
#           'token': <a random string which is reset whenever the state is
#               added to memcache>,
#           'version': <version of the in-memory state of the data>,
#           # Last transaction (may or may not be present)
#           'last_transaction': [
#             # Update 1
#             [<'/'-joined path>, <new value>],
#             # Update 2
#             [...],
#             ...
#           ]
#           # The set of connected clients.
#           'clients': {<client ID 1>: <time when client last connected or
#               reconnected, for the purposes of _CONNECTION_TIMEOUT>,
#               <client ID 2>: ..., ...},
#           'saving': <a boolean indicating if the operation that
#               writes updates to disk is in progress>,
#           'saving_start': <the time the saving started, absent initially>,
#           'disk_version': <version of the data on disk, updated when 'saving'
#               is set to False or when initializing this memcache value.>
#         }
#
#     update_<version>: the update resulting in the version embedded in the key.
#         A JSON-encoded structure as follows:
#
#         [<'/'-joined path>, <new value>]
#
#         If the in-memory state is evicted, the updates between disk version
#         and the in-memory version would become invalid, but it is guaranteed
#         that when the in-memory state is present in memcache, all updates
#         earlier than in-memory version except the ones in last_transaction
#         are valid.


def decode_path(path_str):
  """Decodes a '/'-joined path.

  Args:
      path_str: path as a string.

  Returns:
      Path as a tuple.
  """
  if not path_str: return ()
  else: return tuple(path_str.split('/'))


def _get_random_string():
  """Returns a random string.
  """
  return ''.join(random.choice(string.ascii_letters +
        string.digits) for x in range(12))


def _is_large_update(transactions):
  """Checks whether updates should be saved using __large_update.

  Args:
      transactions: transactions as set by the client.

  Returns:
      Whether to use __large_update.
  """
  if len(transactions) >= _LARGE_UPDATE_TRANSACTIONS: return True
  value_count = 0
  for reads, updates in transactions:
    for update in updates:
      value_count += 1
      if value_count >= _LARGE_UPDATE_VALUES: return True
  return False


# Alias for time.time that allows using mock clock in unit tests.
_gettime = time.time


class _MissingUpdates(memcache_utils.MemcacheError):
  """Some updates could not be retrieved from memcache either because they were
  evicted or due to service outage.
  """


class _InMemoryStateTokenMismatch(memcache_utils.MemcacheError):
  """In-memory state token sent by the client does not match current token.
  """


class _MidAirCollision(Exception):
  """A transaction sent by the client conflicted with a concurrent write.
  """


class _MaybeCorruptedState(Exception):
  """Exception raised when some operation could not be completed with a
  possible reason being the corrupted in-memory state, including updates
  missing from memcache. In this case we will try to flush the state.
  """


class _InsufficientPrivileges(Exception):
  """Exception raised when the client does not have access mode required for the
  operation it has requested.
  """


class _CancelCas(Exception):
  """Exception class used by __update_state method of DatasourceService.
  """


class DatasourceService(Service):
  """Base class for services that server agents on the client can connect to.
  """
  def __init__(self, name, client_id, request):
    super(DatasourceService, self).__init__(name, client_id, request)

    # Defined on instance instead of class to make sure there are no memory
    # leaks.
    self.__memcache_client = memcache.Client()

    # Associated instance of AbstractStorage.
    self.__storage = self.get_storage()

    # A prefix to be added to memcache keys specific to the datasource.
    self.__memcache_prefix = (_config.NAMESPACE +
        self.__storage.get_datasource_key() + '_')


  # Cached value of the in-memory state. Always retrieved with for_cas = True.
  __state = None


  def get_storage(self):
    """Must be overridden to return an instance of a class that implements
    AbstractStorage.
    """
    raise NotImplementedError()


  def get_access_mode(self):
    """Can be overridden to return access mode granted to the client.

    The return value can be 'r' (read-only), 'rw' (read-write), or None (no
    access).

    By default, the access mode is read-write.

    Changes in the return value of this method become effective immediately
    except when a client's read access is revoked while it is connected to the
    server and is receiving notifications of changes in the data - in this case
    it may take up to two hours for these notifications to stop.
    """
    return 'rw'


  def __add_state(self, disk_version):
    """memcache.add's the initial value of in-memory state.

    Args:
        disk_version: version of the data on disk to initialize with.

    Returns: True if the value was added, False if it was already in memcache.
    """
    state_key = self.__memcache_prefix + 'state'
    # We use _async because unlike the others its return value is different for
    # when the value is already in memcache and when there is an RPC error.
    results_add_state = self.__memcache_client.add_multi_async(
        {state_key: memcache_utils.maybe_split(json_utils.dumps({
          'token': _get_random_string(),
          'version': disk_version,
          'clients': {self.client_id: _gettime()},
          'saving': False,
          'disk_version': disk_version,
        }), time=_config.STATE_EXPIRATION)},
        time=_config.STATE_EXPIRATION).get_result()
    if (results_add_state is None or results_add_state[state_key] ==
        memcache.ERROR):
      raise memcache_utils.MemcacheWriteFailure()
    return (results_add_state[state_key] == memcache.STORED)


  def __maybe_initialize_state(self):
    """Checks if the state is present in memcache, and if it is absent,
    calls self.__storage.get_disk_version and then self.__add_state.

    Returns: Passes value returned by self.__add_state.
    """
    state_key = self.__memcache_prefix + 'state'
    results_get_state = self.__memcache_client.get_multi(
        [state_key], for_cas=True)
    state = results_get_state.get(state_key)
    if state is None:
      self.__add_state(self.__storage.get_disk_version())
    else:
      self.__state = json_utils.loads(memcache_utils.maybe_merge(state))


  def __maybe_get_state(self):
    """Loads this.__state if current value of this property is None. If state is
    not in memcache, an exception will be raised.
    """
    state_key = self.__memcache_prefix + 'state'
    results_get_state = self.__memcache_client.get_multi(
        [state_key], for_cas=True)
    state = results_get_state.get(state_key)
    if state is None: raise memcache_utils.MemcacheReadFailure()
    self.__state = json_utils.loads(memcache_utils.maybe_merge(state))


  def __update_state(self, callback):
    """Loads state from memcache into self.__state with for_cas option set to
    True (unless already loaded), calls the supplied function that modifies
    self.__state, and compare-and-sets self.__state back into memcache, if
    necessary making a few retries.

    State must be present in memcache, otherwise MemcacheReadFailure will be
    raised.

    Args:
        callback: a function to run during each cas attempt. Can
            raise _CancelCas exception, in which case this exception will be
            caught and __update_state will return immediately.

    Returns: the return value of the callback the last time it was run.
    """
    state_key = self.__memcache_prefix + 'state'
    attempt, max_attempts = 1, 1 + _STATE_CAS_RETRIES
    while attempt <= max_attempts:
      _logger.debug('Attempt {} to update state'.format(attempt))
      self.__maybe_get_state()
      try:
        retval = callback()
      except _CancelCas:
        return
      state = memcache_utils.maybe_split(json_utils.dumps(self.__state),
          time=_config.STATE_EXPIRATION)
      # We use the _async version because unlike the others it has different
      # return value for when cas failed because the value has changed and when
      # it failed because of RPC error.
      results_cas_state = self.__memcache_client.cas_multi_async(
          {state_key: state}, time=_config.STATE_EXPIRATION).get_result()
      if (results_cas_state is None or results_cas_state[state_key] ==
          memcache.ERROR):
        raise memcache_utils.MemcacheWriteFailure()
      elif (results_cas_state[state_key] == memcache.STORED):
        break
      # There was a concurrent write.
      self.__state = None
      attempt += 1
    else:
      raise memcache_utils.MemcacheContentionError()

    _logger.debug('Updated state succeessfully')
    return retval


  def __delete_state(self):
    """Deletes the in-memory state from memcache.
    """
    state_key = self.__memcache_prefix + 'state'
    result_delete_state = memcache.delete(state_key)
    # 2 == DELETE_SUCCESSFUL according to appengine docs.
    if result_delete_state != 2:
      raise memcache_utils.MemcacheWriteFailure()


  def __get_updates(self, start):
    """Gets all updates to in-memory state after supplied version.

    Args:
        start: the version before the first update.

    Returns:
        A list of updates in the order they were applied.
    """
    updates = []
    end = self.__state['version']
    if 'last_transaction' in self.__state:
      end -= len(self.__state['last_transaction'])
    if start < end:
      version_range = lambda: xrange(start + 1, end + 1)
      results_get_updates = memcache.get_multi(
            [str(x) for x in version_range()],
            key_prefix=(self.__memcache_prefix + 'update_'))
      for version in version_range():
        update = results_get_updates.get(str(version))
        if update is None: raise _MissingUpdates()
        updates.append(json_utils.loads(update))
    if 'last_transaction' in self.__state:
      version = end
      for update in self.__state['last_transaction']:
        version += 1
        if version > start:
          updates.append(update)
    return updates


  def __save_updates_to_memcache(self, updates, version):
    """Saves updates as individual memcache items.

    Overwrites any existing updates with the same versions.

    Args:
        updates: A list of updates.
        version: version after the last of the updates.
    """
    updates_dict = {}
    for index, update in enumerate(updates):
      updates_dict[str(version - len(updates) + index + 1)] = (
          json_utils.dumps(update))
    results_set_updates = memcache.set_multi(updates_dict,
        key_prefix=self.__memcache_prefix + 'update_',
        time=_config.UPDATES_EXPIRATION)
    if len(results_set_updates):
      raise memcache_utils.MemcacheWriteFailure()


  def __maybe_lighten_state(self):
    """If 'last_transaction' is present in self.__state, removes it and saves
    the updates from that transaction as individual memcache items.
    """
    updates = self.__state.pop('last_transaction', None)
    if updates is not None:
      self.__save_updates_to_memcache(updates, self.__state['version'])


  def __acquire_saving_lock(self, token, should_cancel=None):
    """Acquires a lock on saving operation.

    So that only one saving operation would be performed at a time, we  acquire
    a lock by setting 'saving' property of the in-memory state to True. If the
    lock is held by a concurrent call, the function sleeps the number of seconds
    determined by _SAVE_POLL_DELAY constant and then checks the memcache value
    again, repeating the cycle indefinitely.  The cycle is broken either when
    'saving' property is False and we succeessfully acquire the lock, or when
    locking is cancelled via the hook provided as argument. To handle the
    situation when the saving operation fails before releasing the lock, we also
    store saving_start timestamp in the in-memory state, and consider the lock
    to be released after _SAVE_TIMEOUT expires even if 'saving' == True.

    Args:
        token: token of the in-memory state that needs to be saved.
        should_cancel: optional function that doesn't take any arguments and
            returns a boolean indicating whether to cancel the locking.

    Returns: False if cancelled, True if lock was acquired.
    """
    class Cancelled(Exception): pass
    class Busy(Exception): pass

    def txn():
      if self.__state['token'] != token:
        raise _InMemoryStateTokenMismatch()
      if should_cancel is not None and should_cancel():
        raise Cancelled()
      if (not self.__state['saving'] or
          _gettime() - self.__state['saving_start'] > _SAVE_TIMEOUT):
        self.__state['saving'] = True
        self.__state['saving_start'] = _gettime()
        self.__maybe_lighten_state()
      else:
        raise Busy()

    _logger.debug('Will acquire lock on saving operation')
    while True:
      try:
        self.__update_state(txn)
      except Cancelled:
        _logger.debug('Locking cancelled')
        return False
      except Busy:
        _logger.debug('Will wait')
        time.sleep(_SAVE_POLL_DELAY)
        # Clear cached state to force it to be reloaded by __update_state.
        self.__state = None
      else:
        _logger.debug('Lock acquired')
        break

    return True


  def __handle_connect(self, payload):
    """Handles request to connect to datasource.

    Raises:
        _MaybeCorruptedState: if some updates may be missing from memcache.
        _InsufficientPrivileges: if the user does not have read or write access.
    """
    _logger.debug('__handle_connect')

    if self.get_access_mode() not in ['r', 'rw']:
      raise _InsufficientPrivileges()

    connection_token = payload['connect']

    disk_version, init_data = self.__storage.get_disk_state()

    if not self.__add_state(disk_version):

      def txn():
        self.__state['clients'][self.client_id] = _gettime()
        self.__maybe_lighten_state()

      self.__update_state(txn)

    self.__maybe_get_state()

    self.send(SerializedPayload('{{"connected":['
          '{connection_token},'
          '{state_token},'
          '{disk_version},'
          '{init_data}'
        ']}}'.format(connection_token=json_utils.dumps(connection_token),
        state_token=json_utils.dumps(self.__state['token']),
        disk_version=json_utils.dumps(disk_version),
        init_data=init_data)))

    try:
      updates = self.__get_updates(disk_version)
    except _MissingUpdates:
      raise _MaybeCorruptedState()
    if updates:
      self.send({'updates': [self.__state['token'], self.__state['version'],
          updates]})


  def __handle_reconnect(self, payload):
    """Handles request to reconnect to datasource.

    Raises:
        _InsufficientPrivileges: if the user does not have read or write access.
    """
    _logger.debug('__handle_reconnect')

    if self.get_access_mode() not in ['r', 'rw']:
      raise _InsufficientPrivileges()

    base_version, connection_token = payload['reconnect']

    # If the state was evicted from memcache, restore it. We cannot restore
    # the updates (which if state was evicted, are also probably evicted), but
    # it's likely that the client has the same version as disk state, and in
    # this case error will be avoided.
    if not self.__maybe_initialize_state():

      def txn():
        self.__state['clients'][self.client_id] = _gettime()
        self.__maybe_lighten_state()

      self.__update_state(txn)

    self.__maybe_get_state()

    self.send({'reconnected': [connection_token, self.__state['token']]})

    # If some updates are missing from memcache, this will throw an exception.
    # We let the exception bubble up so that client would be sent an error
    # message. The reason why the updates may be missing is a very old state
    # of the data on the client. The client can in this case flush its state
    # and try connecting via 'connect' request.
    updates = self.__get_updates(base_version)
    if updates:
      self.send({'updates': [self.__state['token'], self.__state['version'],
          updates]})

    self.send({'saved': [self.__state['token'], self.__state['disk_version']]})


  def __apply_transaction(self, reads, updates, cache, state_token):
    """Applies a transaction to the in-memory state of the data and notifies
    other clients.

    Args:
        reads: reads made inside the transaction, in the format as they were
            sent by the client.
        updates: a list of writes made inside the transaction.
        cache: a dictionary used to cache data between calls for a list of
            transactions sent by the client, for description of items see
            the body of __handle_updates().
        state_token: in-memory state token sent by the client.

    Returns:
        The version of the data after the transaction was applied.

    Raises:
        _MidAirCollision: transaction conflicted with another concurrent
            transaction.
    """
    def txn():
      if self.__state['token'] != state_token:
        raise _InMemoryStateTokenMismatch()

      # Bring diverged_value_paths up to date.
      concurrent_updates = self.__get_updates(
          cache['last_checked_version'])
      for update in concurrent_updates:
        cache['diverged_value_paths'].add(decode_path(update[0]))
        cache['last_checked_version'] += 1

      # Check if any of the diverged_value_paths match the reads.
      def path_matches_reads(path, reads_el):
        if not path:
          return _LEAF_KEY in reads_el
        if path[0] in reads_el and path_matches_reads(path[1:],
            reads_el[path[0]]):
          return True;
        if '*' in reads_el and path_matches_reads(path[1:], reads_el['*']):
          return True
        return False

      if any((path_matches_reads(path, reads) for path in
          cache['diverged_value_paths'])):
        # The fact that the transaction isn't going to be applied causes in-
        # memory state to diverge from the client state, so we need to add the
        # affected paths to diverged_value_paths.
        for update in updates:
          cache['diverged_value_paths'].add(decode_path(update[0]))
        raise _MidAirCollision()

      # Update in-memory state.
      previous_transaction = self.__state.get('last_transaction')
      if previous_transaction is not None:
        self.__save_updates_to_memcache(previous_transaction,
            self.__state['version'])
      self.__state['last_transaction'] = updates
      version = cache['last_checked_version'] + len(updates)
      self.__state['version'] = version
      return version

    version = self.__update_state(txn)

    # We set this only after cas has succeeded.
    cache['last_checked_version'] = version

    return version


  def __save(self, version, token):
    """Makes sure that in-memory updates up to provided version have been saved
    to disk.

    Args:
        version: version up to which the updates need to be saved.
        token: token of the in-memory state that needs to be saved.

    Raises:
        _MaybeCorruptedState: see comment in the body.
    """
    # Acquire the lock. If a concurrent call has saved the necessary data,
    # cancel locking and return without doing anything.
    if not self.__acquire_saving_lock(token,
        lambda: self.__state['disk_version'] >= version):
      return

    # Now the lock has been acquired and we need to perform the saving
    # operation.
    updates = self.__get_updates(self.__state['disk_version'])
    new_disk_version = self.__state['version']
    _logger.debug('Will start saving')
    saved = self.__storage.save_updates(updates, new_disk_version)
    if not saved:
      # The data might not be saved because some previous saving operation did
      # complete but this was never recorded in the in-memory state, so the
      # check performed while saving that makes sure the data is written over
      # correct state has failed. This would mean the state is corrupted and we
      # need to flush it. This is also a precaution against infinite retries
      # if the saving operation does not like the data we're trying to save for
      # some other reason, e.g. data too large.
      raise _MaybeCorruptedState()

    # Release the lock.
    def txn():
      if self.__state['token'] != token:
        raise _InMemoryStateTokenMismatch()
      self.__state['saving'] = False
      del self.__state['saving_start']
      self.__state['disk_version'] = new_disk_version
      self.__maybe_lighten_state()

    _logger.debug('Finished saving, will release lock on saving operation')
    self.__update_state(txn)


  def __large_update(self, connection_token, state_token, base_version,
      transactions):
    """Saves client updates directly to disk, bypassing in-memory state.

    When a large set updates is handled in the normal way, this can cause
    deadline exceeded error, running out of memory because of large usage by
    json module, and issues with values being evicted from memcache. By not
    saving the updates to memcache, we can handle a larger operation, but if
    there are any other clients connected to the datasource, they will err and
    will have to reload the state.

    In the current implementation of this method, we roll back all the updates
    if the state of the data on disk is different from the state on the client
    before the updates.

    Args:
        connection_token: passed from the client's message.
        state_token: token to check when updating state.
        base_version: passed from the client's message.
        transactions: passed from the client's message.

    Raises:
        _MaybeCorruptedState: can be raised while attempting to write data to
            disk.
    """
    self.__acquire_saving_lock(state_token)

    if self.__state['disk_version'] != base_version:
      def txn_release_lock():
        if self.__state['token'] != state_token:
          raise _InMemoryStateTokenMismatch()
        self.__state['saving'] = False
        del self.__state['saving_start']
        self.__maybe_lighten_state()
      self.__update_state(txn_release_lock)
      self.send({'outcomes': [connection_token, state_token, base_version,
          [None for transaction in transactions]]})
      return

    updates = []
    outcomes = []
    version = base_version
    for reads, transaction_updates in transactions:
      updates.extend(transaction_updates)
      version += len(transaction_updates)
      outcomes.append(version)
    new_disk_version = base_version + len(updates)
    _logger.debug('Will start saving (__large_update)')
    saved = self.__storage.save_updates(updates, new_disk_version)
    if not saved:
      # See comment in the body of __save.
      raise _MaybeCorruptedState()

    def txn_increment_memory_version():
      if self.__state['token'] != state_token:
        raise _InMemoryStateTokenMismatch()

      last_update_version = self.__state['version']
      if 'last_transaction' in self.__state:
        last_update_version -= len(self.__state['last_transaction'])
        del self.__state['last_transaction']

      if base_version < last_update_version:
        version_range = lambda: xrange(base_version + 1,
            last_update_version + 1)
        if self.__memcache_client.delete_multi_async(
            [str(x) for x in version_range()],
            key_prefix=(self.__memcache_prefix + 'update_')
            ).get_result() is None:
          raise memcache_utils.MemcacheWriteFailure()

      self.__state['version'] = new_disk_version
      self.__state['disk_version'] = new_disk_version
      self.__state['saving'] = False
      del self.__state['saving_start']

    _logger.debug('Finished saving, will update state')
    self.__update_state(txn_increment_memory_version)
    other_clients = [client_id for client_id in self.__state['clients']
        if client_id != self.client_id]
    if other_clients:
      self.send({'error': None}, to=other_clients)
    self.send({'outcomes': [connection_token, state_token, base_version,
        outcomes]})
    self.send({'saved': [self.__state['token'], new_disk_version]})


  def __handle_updates(self, payload):
    """Handles updates sent by a client.

    Args:
        payload: message received from the client.

    Raises:
        _MaybeCorruptedState: can be raised while attempting to write data to
            disk.
        _InsufficientPrivileges: if the user does not have write access.
    """
    _logger.debug('__handle_updates')

    if self.get_access_mode() != 'rw':
      raise _InsufficientPrivileges()

    (connection_token, state_token, base_version,
        transactions) = payload['updates']

    self.__maybe_initialize_state()
    self.__maybe_get_state()
    if state_token is None:
      state_token = self.__state['token']

    if _is_large_update(transactions):
      self.__large_update(connection_token, state_token, base_version,
          transactions)
      return

    cache = {
      # Such version number that any concurrent updates between base_version and
      # this version have been added to diverged_value_paths.
      'last_checked_version': base_version,
      # A set of paths (as tuples) at which values may be different than on the
      # client when the updates were made there. That can be caused by two
      # things: either there was some update by another client that hasn't yet
      # been applied on the first client, or some prior transaction from the
      # first client was not applied on the server.
      'diverged_value_paths': set()
    }
    # Updates that were committed, merged from committed transactions.
    committed_updates = []
    outcomes = []
    for reads, updates in transactions:
      try:
        version = self.__apply_transaction(reads, updates, cache, state_token)
        outcomes.append(version)
        committed_updates.extend(updates)
      except _MidAirCollision:
        outcomes.append(None)
    self.send({'outcomes': [connection_token, state_token, base_version,
        outcomes]})
    if committed_updates:
      updates_message = SerializedPayload(json_utils.dumps(
          {'updates': [state_token, version, committed_updates]}))

      # Clean up disconnected clients.
      def txn():
        disconnected_clients = [client_id for client_id, time_connected in
            self.__state['clients'].iteritems()
            if client_id != self.client_id and
            time_connected < _gettime() - _CONNECTION_TIMEOUT]
        if not disconnected_clients:
          raise _CancelCas()
        for client_id in disconnected_clients:
          del self.__state['clients'][client_id]
        self.__maybe_lighten_state()

      self.__update_state(txn)

      other_clients = [client_id for client_id in self.__state['clients']
          if client_id != self.client_id]
      if other_clients:
        self.send(updates_message, to=other_clients)

      self.__save(version, state_token)
      self.send({'saved': [state_token, self.__state['disk_version']]})


  def __handle_disconnect(self):
    """Handles are request to disconnect.
    """
    _logger.debug('__handle_disconnect')
    def txn():
      if self.client_id not in self.__state['clients']:
        raise _CancelCas()
      del self.__state['clients'][self.client_id]
      self.__maybe_lighten_state()

    self.__update_state(txn)


  def handle_message(self, payload):
    """Implements the abstract method of the superclass.
    """
    try:
      if 'connect' in payload:
        self.__handle_connect(payload)
      elif 'reconnect' in payload:
        self.__handle_reconnect(payload)
      elif 'updates' in payload:
        self.__handle_updates(payload)
      elif 'disconnect' in payload:
        self.__handle_disconnect()
      else:
        _logger.warning('Datasource service %s received unrecognized message.'%
            self.name)
    except memcache_utils.MemcacheError:
      _logger.warning('Memcache-related error handling incoming message.',
          exc_info=sys.exc_info())
      self.send({'error': None})
    except _InsufficientPrivileges:
      _logger.warning('The client does not have access mode required for the'
          ' operation it has requested.', exc_info=sys.exc_info())
      self.send({'error': None})
    except _MaybeCorruptedState:
      _logger.warning(
          'Possibly corrupted in-memory state, will attempt to flush.',
          exc_info=sys.exc_info())
      self.__delete_state()
      self.send({'error': None}, to=self.__state['clients'].keys())
    except Exception:
      _logger.error('Error handling incoming message.',
          exc_info=sys.exc_info())
      self.send({'error': None})
