# Copyright 2012 MightyClient Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS-IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utilities for working with memcache.
"""

import logging
import random
import string

from google.appengine.api import memcache


_logger = logging.getLogger(__name__)


# Maximum # of bytes to use for a single memcache value.
_VALUE_LIMIT = 999000


class MemcacheError(Exception):
  """Base class for errors occuring when working with memcache.
  """


class MemcacheReadFailure(MemcacheError):
  """Memcache read failure.
  """


class MemcacheWriteFailure(MemcacheError):
  """Memcache write failure.
  """


class MemcacheContentionError(MemcacheError):
  """Compare-and-set failed because of high contention.
  """


def maybe_split(value, time=0):
  """Utility function for saving in memcache a string that's possibly larger
  than the size limit for a memcache value, works in tandem with maybe_merge.

  If the string is below the limit, returns it as is without doing anything.

  Otherwise, splits it into smaller chunks, saves each chunk in memcache, and
  returns a list of keys of the chunks. This list can then be stored in memcache
  in place of the large string.

  Chunks are saved in a single memcache call, so max call size limit applies.

  Args:
      value: the value that has to be saved in memcache, must be an instance of
          str (not unicode).
      time: optinal expiration time.

  Returns:
      The string supplied as argument or a list of keys of the chunks.
  """
  assert isinstance(value, str)
  if len(value) < _VALUE_LIMIT: return value
  _logger.debug('String length {}, will split.'.format(len(value)))
  chunks = []
  keys = []
  for index in xrange(0, len(value), _VALUE_LIMIT - 1):
    chunks.append(value[index : index + _VALUE_LIMIT - 1])
    keys.append(''.join(random.choice(string.ascii_letters + string.digits)
        for x in xrange(12)))
  results = memcache.set_multi(dict(zip(keys, chunks)), time=time)
  if len(results):
    raise MemcacheWriteFailure()
  return keys


def maybe_merge(value):
  """Utility function for loading from memcache a string that's possibly larger
  than the size limit for a memcache value, works in tandem with maybe_split.

  If the supplied value is an str instance, returns it as is without doing
  anything.

  Otherwise, it is assumed that the value is a list returned by maybe_split. In
  this case we load the chunks from memcache and return the joined string.

  Args:
      value: a string or a list of memcache keys for the chunks that need to be
          joined.

  Returns:
      The string passed as argument if the argument was an str instance, or the
      joined chunks otherwise.
  """
  if isinstance(value, str): return value
  results = memcache.get_multi(value)
  if results is None: raise MemcacheReadFailure()
  retval = ''
  for key in value:
    if key not in results: raise MemcacheReadFailure()
    retval += results[key]
  _logger.debug('Has merged, resulting length {}.'.format(len(retval)))
  return retval
