#!/usr/bin/python
#
# Copyright 2012 MightyClient Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS-IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from collections import defaultdict
import json_utils
import unittest

from google.appengine.api import channel
from google.appengine.api import memcache
from google.appengine.ext import testbed
import webapp2
import webtest

import channel as mcchannel
from service import SerializedPayload, Service


original_services = mcchannel._config.SERVICES


class TestCase(unittest.TestCase):

  def setUp(self):
    app = webapp2.WSGIApplication([
      ('/mc_channel', mcchannel.ForwardChannelHandler),
    ], debug=True)
    self.testapp = webtest.TestApp(app)
    self.testbed = testbed.Testbed()
    self.testbed.activate()
    self.time = 0.
    mcchannel._gettime = lambda: self.time
    stub = memcache.memcache_stub.MemcacheServiceStub(
        gettime=(lambda: self.time))
    self.testbed._register_stub(testbed.MEMCACHE_SERVICE_NAME, stub)
    self.testbed.init_channel_stub()
    # Allows tests to modify SERVICES.
    mcchannel._config.SERVICES = list(original_services)


  def tearDown(self):
    self.testbed.deactivate()


  def forward_channel_post(self, client_id, messages):
    """Simulates a post request from the client.

    Args:
        client_id: client ID.
        messages: dictionary where keys are service names and values are lists
            of non-encoded messages.

    Returns: passes the retval of self.testapp.post.
    """
    return self.testapp.post('/mc_channel' + '?zx=abc',
        {'cid': client_id, 'm': json_utils.dumps(messages)},
        {'Content-Type': 'application/x-www-form-urlencoded;char set=utf-8'})


  def test_store_message_for_pickup(self):
    message = 'message1'
    service = 'service1'
    pickup_id = mcchannel._store_message_for_pickup(service, message)
    memcache_key = mcchannel._config.NAMESPACE + 'pickup_' + pickup_id
    self.assertEquals({'service': service, 'message': message},
        memcache.get(memcache_key))
    self.assertEquals((service, message), mcchannel._pickup_message(pickup_id))
    self.time += mcchannel._PICKUP_TIMEOUT - 1
    self.assertTrue(memcache.get(memcache_key) is not None)
    self.time += 2
    self.assertTrue(memcache.get(memcache_key) is None)


  def test_encode_messages(self):
    self.assertEquals(json_utils.dumps({
      'service1': [[1], [2]],
      'service2': ['b']
    }), mcchannel._encode_messages({
      'service1': ['[1]', '[2]'],
      'service2': ['"b"']
    }))


  def test_send_message_by_polling(self):
    token = 'token1'
    pickup_id1 = 'pickup_id1'
    memcache_key = mcchannel._config.NAMESPACE + 'poll_' + token;
    mcchannel._send_message_by_polling(token, pickup_id1)
    self.assertEquals([pickup_id1], memcache.get(memcache_key))

    pickup_id2 = 'pickup_id2'
    mcchannel._send_message_by_polling(token, pickup_id2)
    self.assertEquals([pickup_id1, pickup_id2], memcache.get(memcache_key))

    self.time += mcchannel._PICKUP_TIMEOUT - 1
    self.assertTrue(memcache.get(memcache_key) is not None)
    self.time += 2
    self.assertTrue(memcache.get(memcache_key) is None)


  def test_get_service_instance(self):
    class Service1(Service): pass
    class Service2(Service): pass
    mcchannel._config.SERVICES.insert(0, ('aa.', Service1))
    mcchannel._config.SERVICES.insert(0, ('^bbb$', Service2))
    request = object()
    service = mcchannel._get_service_instance('aax', '1', request)
    self.assertIsInstance(service, Service1)
    self.assertEquals('1', service.client_id)
    self.assertEquals('aax', service.name)
    self.assertIs(request, service.request)
    service = mcchannel._get_service_instance('bbb', '1', request)
    self.assertIsInstance(service, Service2)
    self.assertIs(None, mcchannel._get_service_instance('aaxy', '1', request))
    self.assertIs(None, mcchannel._get_service_instance('yaax', '1', request))


  def test_handle_get_token_message(self):
    channel_stub = self.testbed.get_stub('channel')
    client_id = '1'

    # Test creation of GAE Channel.
    response = self.forward_channel_post(client_id,
        {mcchannel._SYSTEM_SERVICE: [{'getToken': False}]})
    token_message = json_utils.loads(response.body)[
        mcchannel._SYSTEM_SERVICE][-1]
    self.assertIs(False, token_message['opened']['poll'])
    token = token_message['opened']['token']
    self.assertEquals('1', channel_stub.client_id_from_token(token))
    self.assertEquals({'poll': False}, memcache.get(
        mcchannel._config.NAMESPACE + 'connected_1'))
    self.time += mcchannel._CHANNEL_LIFESPAN
    self.assertIs(None, memcache.get(
        mcchannel._config.NAMESPACE + 'connected_1'))

    # Test behaviour when creating channel throws an error.
    original_create_channel = channel.create_channel
    def raise_error(client_id): raise Exception()
    channel.create_channel = raise_error
    try:
      response = self.forward_channel_post(client_id,
          {mcchannel._SYSTEM_SERVICE: [{'getToken': False}]})
      token_message = json_utils.loads(response.body)[
          mcchannel._SYSTEM_SERVICE][-1]
      token = token_message['opened']['token']
      self.assertEquals({'poll': True, 'token': token}, memcache.get(
          mcchannel._config.NAMESPACE + 'connected_1'))
      self.time += mcchannel._CHANNEL_LIFESPAN
      self.assertIs(None, memcache.get(
          mcchannel._config.NAMESPACE + 'connected_1'))
    finally:
      channel.create_channel = original_create_channel

    # Test behaviour when polling is forced.
    response = self.forward_channel_post(client_id,
        {mcchannel._SYSTEM_SERVICE: [{'getToken': True}]})
    token_message = json_utils.loads(response.body)[
        mcchannel._SYSTEM_SERVICE][-1]
    token = token_message['opened']['token']
    self.assertEquals({'poll': True, 'token': token}, memcache.get(
        mcchannel._config.NAMESPACE + 'connected_1'))
    self.time += mcchannel._CHANNEL_LIFESPAN
    self.assertIs(None, memcache.get(
        mcchannel._config.NAMESPACE + 'connected_1'))


  def test_handle_pickup_message(self):
    client_id = '1'
    memcache.set(mcchannel._config.NAMESPACE + 'pickup_' + 'abc',
        {'service': 'service1', 'message': '"message1"'})
    response = self.forward_channel_post(client_id,
        {mcchannel._SYSTEM_SERVICE: [{'pickup': 'abc'}]})
    self.assertEquals({'service1': ['message1']},
        json_utils.loads(response.body))


  def test_handle_poll_message(self):
    client_id = '1'
    token = 'token1'
    memcache_key = mcchannel._config.NAMESPACE + 'poll_' + token
    service = 'service1'
    message = 'message1'
    pickup_id = mcchannel._store_message_for_pickup(service,
        json_utils.dumps(message))
    mcchannel._send_message_by_polling(token, pickup_id)
    response = self.forward_channel_post(client_id,
        {mcchannel._SYSTEM_SERVICE: [{'poll': token}]})
    self.assertEquals({service: [message]}, json_utils.loads(response.body))
    self.assertEquals([], memcache.get(memcache_key))
    self.time += mcchannel._CHANNEL_LIFESPAN - 1
    self.assertFalse(memcache.get(memcache_key) is None)
    self.time += 2
    self.assertTrue(memcache.get(memcache_key) is None)


  def test_send(self):
    message = 'message1'
    large_message = 'a'*(32 * 1024 + 1)
    service = 'service1'
    channel_stub = self.testbed.get_stub('channel')

    # Client 1 will not be connected but we will call send() during processing
    # of a fwd channel request from this client.

    # Connect client 2 via GAE Channel.
    response = self.forward_channel_post('2',
        {mcchannel._SYSTEM_SERVICE: [{'getToken': False}]})
    token_client_2 = json_utils.loads(response.body)[
        mcchannel._SYSTEM_SERVICE][-1]['opened']['token']
    channel_stub.connect_channel(token_client_2)

    # Connect client 3 via polling.
    response = self.forward_channel_post('3',
        {mcchannel._SYSTEM_SERVICE: [{'getToken': True}]})
    token_client_3 = json_utils.loads(response.body)[
        mcchannel._SYSTEM_SERVICE][-1]['opened']['token']

    # Client 4 will not be connected.

    # Client 5 will be connected via GAE Channel and will be used to test
    # sending message with size exceeding GAE Channel api limit.
    response = self.forward_channel_post('5',
        {mcchannel._SYSTEM_SERVICE: [{'getToken': False}]})
    token_client_5 = json_utils.loads(response.body)[
        mcchannel._SYSTEM_SERVICE][-1]['opened']['token']
    channel_stub.connect_channel(token_client_5)

    class TestService(Service):

      def handle_message(self, payload):
        assert self.client_id == '1'
        self.send(message, service=service, to=['1', '2', '3', '4'])
        self.send(large_message, service=service, to='5')

    mcchannel._config.SERVICES.insert(0, (service, TestService))

    response = self.forward_channel_post('1', {service: ['']})
    self.assertEquals({service: [message]}, json_utils.loads(response.body))

    messages_client_2 = channel_stub.get_channel_messages(token_client_2)
    self.assertEquals(1, len(messages_client_2))
    self.assertEquals({service: [message]},
        json_utils.loads(messages_client_2[0]))

    response = self.forward_channel_post('3',
        {mcchannel._SYSTEM_SERVICE: [{'poll': token_client_3}]})
    self.assertEquals({service: [message]},
        json_utils.loads(response.body))

    messages_client_5 = channel_stub.get_channel_messages(token_client_5)
    self.assertEquals(1, len(messages_client_5))
    message_client_5 = json_utils.loads(messages_client_5[0])
    pickup_id = message_client_5[mcchannel._SYSTEM_SERVICE][0]['pickup']
    response = self.forward_channel_post('5',
        {mcchannel._SYSTEM_SERVICE: [{'pickup': pickup_id}]})
    self.assertEquals({service: [large_message]},
        json_utils.loads(response.body))

    self.assertEquals({}, mcchannel._outgoing_messages)


  def test_forward_channel_handler(self):
    # Tests functionality of POST request handler and
    # __handle_incoming_messages, except sending messages in responce to fwd
    # channel request, which is tested in test_send.

    channel_stub = self.testbed.get_stub('channel')

    handle_message_calls = {'service1': [], 'service2': []}

    class TestService(Service):
      def handle_message(testService, payload):
        handle_message_calls[testService.name].append(
            {'client_id': testService.client_id, 'payload': payload})
        self.assertIsInstance(testService.request, webapp2.Request)

    mcchannel._config.SERVICES.insert(0, ('service2', TestService))
    mcchannel._config.SERVICES.insert(0, ('service1', TestService))

    messages = {'service1': ['message1'],
      'service2': ['message2', 'message3']}

    self.forward_channel_post('1', messages)
    self.assertEquals({
      'service1': [
        {'client_id': '1', 'payload': 'message1'},
      ],
      'service2': [
        {'client_id': '1', 'payload': 'message2'},
        {'client_id': '1', 'payload': 'message3'},
      ]}, handle_message_calls)
    self.assertEquals(set(), mcchannel._open_requests)


  def test_forward_channel_handler_jsonp(self):
    def stub(handler, client_id, messages_str):
      self.assertEquals('1', client_id)
      self.assertEquals('msg', messages_str)
      return '"json response"'
    original_method = (mcchannel.ForwardChannelHandler.
        _ForwardChannelHandler__handle_incoming_messages)
    try:
      mcchannel.ForwardChannelHandler.\
          _ForwardChannelHandler__handle_incoming_messages = stub
      response = self.testapp.get('/mc_channel' +
          '?zx=abc&cid=1&m=msg&callback=my_callback')
      self.assertEquals('my_callback("json response")', response.body)
    finally:
      mcchannel.ForwardChannelHandler.\
          _ForwardChannelHandler__handle_incoming_messages = original_method


if __name__ == '__main__':
  unittest.main()
