#!/usr/bin/python
#
# Copyright 2012 MightyClient Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS-IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from collections import defaultdict
import json_utils
import threading
import time
import unittest

from google.appengine.api import channel
from google.appengine.api import memcache
from google.appengine.ext import testbed
import webtest

from abstract_storage import AbstractStorage
import channel as mcchannel
import datasource
import memcache_utils
from service import SerializedPayload


class MockStorage(AbstractStorage):
  def __init__(self):
    self.save_updates_calls = []
    self.save_updates_retval = True


  def get_datasource_key(self):
    return 'datasource_key'


  def get_disk_version(self):
    return self.version


  def get_disk_state(self):
    return self.version, json_utils.dumps(self.state)


  def save_updates(self, *args):
    self.save_updates_calls.append(args)
    return self.save_updates_retval



class TestCase(unittest.TestCase):

  def setUp(self):
    self.testbed = testbed.Testbed()
    self.testbed.activate()
    self.testbed.init_memcache_stub()
    self.testbed.init_channel_stub()
    self.memcache_prefix = (datasource._config.NAMESPACE + 'datasource_key_')

    class TestService(datasource.DatasourceService):
      def get_storage(self_service):
        return MockStorage()

    self.service = TestService('service1', '1', object())
    datasource._get_random_string = lambda: 'random_string'
    mcchannel._open_requests = set(['1'])


  def tearDown(self):
    self.testbed.deactivate()
    mcchannel._open_requests = set()
    mcchannel._outgoing_messages = defaultdict(lambda: defaultdict(list))


  def set_state(self, state):
    memcache.set(self.memcache_prefix + 'state',
        memcache_utils.maybe_split(json_utils.dumps(state)))


  def get_state(self):
    state = memcache.get(self.memcache_prefix + 'state')
    if state is None: return None
    return json_utils.loads(memcache_utils.maybe_merge(state))


  def test_decode_path(self):
    self.assertEquals((), datasource.decode_path(''))
    self.assertEquals(('a',), datasource.decode_path('a'))
    self.assertEquals(('a', 'b'), datasource.decode_path('a/b'))


  def test_add_state(self):
    original_gettime = datasource._gettime
    datasource._gettime = lambda: 0.
    try:
      self.assertTrue(self.service._DatasourceService__add_state(3))
      state = self.get_state()
      self.assertEquals({
        'token': 'random_string',
        'version': 3,
        'clients': {'1': 0.},
        'saving': False,
        'disk_version': 3,
      },  state)
      self.assertFalse(self.service._DatasourceService__add_state(3))
    finally:
      datasource._gettime = original_gettime


  def test_update_state(self):
    self.set_state('a')
    def txn():
      self.assertEquals('a', self.service._DatasourceService__state)
      self.service._DatasourceService__state = 'b'
      return 1
    retval = self.service._DatasourceService__update_state(txn)
    self.assertEquals(1, retval)
    self.assertEquals('b', self.get_state())

    def txn():
      self.assertEquals('b', self.service._DatasourceService__state)
      self.service._DatasourceService__state = 'c'
      return 1
    retval = self.service._DatasourceService__update_state(txn)
    self.assertEquals(1, retval)
    self.assertEquals('c', self.get_state())

    def txn():
      self.assertEquals('c', self.service._DatasourceService__state)
      self.service._DatasourceService__state = 'd'
      raise datasource._CancelCas()
    self.service._DatasourceService__update_state(txn)
    self.assertEquals('c', self.get_state())


  def test_delete_state(self):
    self.set_state('a')
    self.service._DatasourceService__delete_state()
    self.assertIs(None, self.get_state())


  def test_get_updates(self):
    memcache_prefix = self.memcache_prefix + 'update_'
    memcache.set(memcache_prefix + '2', json_utils.dumps(['a', 'v1']))
    memcache.set(memcache_prefix + '3', json_utils.dumps(['b', 'v2']))
    self.service._DatasourceService__state = {
      'version': 5,
      'last_transaction': [
        ['c', 'v3'],
        ['d', 'v4'],
      ]
    }
    self.assertEquals([
      ['a', 'v1'],
      ['b', 'v2'],
      ['c', 'v3'],
      ['d', 'v4'],
    ], self.service._DatasourceService__get_updates(1))
    with self.assertRaises(datasource._MissingUpdates):
      self.service._DatasourceService__get_updates(0)


  def test_save_updates_to_memcache(self):
    self.service._DatasourceService__save_updates_to_memcache([
      ['a', 'v1'],
      ['b', 'v2'],
    ], 3)
    self.assertEquals(json_utils.dumps(['a', 'v1']), memcache.get(
        self.memcache_prefix + 'update_2'))
    self.assertEquals(json_utils.dumps(['b', 'v2']), memcache.get(
        self.memcache_prefix + 'update_3'))


  def test_handle_connect(self):
    original_gettime = datasource._gettime
    datasource._gettime = lambda: 0.
    try:
      self.service.get_access_mode = lambda: 'r'
      self.service._DatasourceService__storage.version = 1
      self.service._DatasourceService__storage.state = {'a': {'.': 'v1'}}
      memcache.set(self.memcache_prefix + 'update_' + '2',
          json_utils.dumps(['b', 'v2']))
      self.set_state({
        'token': 'state_token_1',
        'version': 2,
        'clients': {},
        'disk_version': 1
      })

      self.service._DatasourceService__handle_connect({'connect': 'abc'})

      self.assertEquals({'1': {'service1': [
        json_utils.dumps({'connected': [
          'abc',
          'state_token_1',
          1,
          {'a': {'.': 'v1'}},
        ]}),
        json_utils.dumps({'updates': [
          'state_token_1',
          2,
          [
            ['b', 'v2'],
          ]
        ]}),
      ]}}, mcchannel._outgoing_messages)
      self.assertEquals({'1': 0}, self.get_state()['clients'])

      memcache.delete(self.memcache_prefix + 'update_' + '2')
      with self.assertRaises(datasource._MaybeCorruptedState):
        self.service._DatasourceService__handle_connect({'connect': 'abc'})

      with self.assertRaises(datasource._InsufficientPrivileges):
        self.service.get_access_mode = lambda: None
        self.service._DatasourceService__handle_connect({'connect': 'abc'})

    finally:
      datasource._gettime = original_gettime


  def test_handle_reconnect(self):
    original_gettime = datasource._gettime
    datasource._gettime = lambda: 0.
    try:
      self.service.get_access_mode = lambda: 'r'
      self.set_state({
        'token': 'state_token_1',
        'version': 2,
        'last_transaction': [
          ['b', 'v2'],
        ],
        'clients': {},
        'disk_version': 1
      })

      self.service._DatasourceService__handle_reconnect({'reconnect': [
        1,
        'connection_token_1',
      ]})

      self.assertEquals({'1': {'service1': [
        json_utils.dumps({'reconnected':
            ['connection_token_1', 'state_token_1']}),
        json_utils.dumps({'updates': [
          'state_token_1',
          2,
          [
            ['b', 'v2'],
          ]
        ]}),
        json_utils.dumps({'saved': ['state_token_1', 1]})
      ]}}, mcchannel._outgoing_messages)
      self.assertEquals({'1': 0}, self.get_state()['clients'])

      memcache.delete(self.memcache_prefix + 'state')
      self.service._DatasourceService__storage.version = 1
      mcchannel._outgoing_messages = defaultdict(lambda: defaultdict(list))
      self.service._DatasourceService__handle_reconnect({'reconnect': [
        1,
        'connection_token_1',
      ]})
      self.assertEquals({'1': {'service1': [
        json_utils.dumps({'reconnected':
            ['connection_token_1', self.get_state()['token']]}),
        json_utils.dumps({'saved': [self.get_state()['token'], 1]})
      ]}}, mcchannel._outgoing_messages)

      with self.assertRaises(datasource._InsufficientPrivileges):
        self.service.get_access_mode = lambda: None
        self.service._DatasourceService__handle_reconnect({'reconnect': [
          1,
          'connection_token_1',
        ]})

    finally:
      datasource._gettime = original_gettime


  def test_apply_transaction(self):
    memcache.set(self.memcache_prefix + 'update_' + '2',
        json_utils.dumps(['a', 'v1']))
    self.set_state({
      'token': 'state_token_1',
      'version': 3,
      'last_transaction': [
        ['b', 'v2'],
      ],
      'clients': {'1': None, '2': None}
    })
    cache = {
      'last_checked_version': 1,
      'diverged_value_paths': set()
    }

    with self.assertRaises(datasource._InMemoryStateTokenMismatch):
      self.service._DatasourceService__apply_transaction({}, [
        ['c', 'v3']
      ], cache, 'state_token_2')

    retval = self.service._DatasourceService__apply_transaction({}, [
      ['c', 'v3']
    ], cache, 'state_token_1')
    self.assertEquals({
      'last_checked_version': 4,
      'diverged_value_paths': set([('a',), ('b',)])
    }, cache)
    self.assertEquals(json_utils.dumps(['b', 'v2']),
        memcache.get(self.memcache_prefix + 'update_' + '3'))
    self.assertEquals({
      'token': 'state_token_1',
      'version': 4,
      'last_transaction': [
        ['c', 'v3'],
      ],
      'clients': {'1': None, '2': None}
    }, self.get_state())
    self.assertEquals(4, retval)

    with self.assertRaises(datasource._MidAirCollision):
      self.service._DatasourceService__apply_transaction({'a': {'.': None}}, [
        ['d', 'v4'],
        ['e', 'v5'],
      ], cache, 'state_token_1')
    self.assertEquals({
      'last_checked_version': 4,
      'diverged_value_paths': set([('a',), ('b',), ('d',), ('e',)])
    }, cache)


  def test_save(self):
    original_save_poll_delay = datasource._SAVE_POLL_DELAY
    datasource._SAVE_POLL_DELAY = 0.2
    try:
      self.set_state({'token': 'state_token_1'})
      with self.assertRaises(datasource._InMemoryStateTokenMismatch):
        self.service._DatasourceService__save(3, 'state_token_2')

      for saving in [True, False]:
        # Clear cached state.
        self.service._DatasourceService__state = None
        self.set_state(
            {'token': 'state_token_1', 'disk_version': 3, 'saving': saving})
        self.service._DatasourceService__save(3, 'state_token_1')
        self.assertEquals([],
            self.service._DatasourceService__storage.save_updates_calls)

      memcache.set(self.memcache_prefix + 'update_' + '1',
          json_utils.dumps(['', 'v1']))
      init_state = {
        'token': 'state_token_1',
        'version': 2,
        'last_transaction': [
          ['b', 'v2'],
        ],
        'clients': {'1': None},
        'saving': False,
        'disk_version': 0,
      }
      # Clear cached state.
      self.service._DatasourceService__state = None
      self.set_state(init_state)
      self.service._DatasourceService__save(1, 'state_token_1')
      self.assertEquals(False,
          self.get_state()['saving'])
      self.assertEquals(2,
          self.get_state()['disk_version'])
      self.assertEquals([([
        ['', 'v1'],
        ['b', 'v2']
      ], 2)], self.service._DatasourceService__storage.save_updates_calls)

      # Clear cached state.
      self.service._DatasourceService__state = None
      self.set_state(init_state)
      self.service._DatasourceService__storage.save_updates_retval = False
      with self.assertRaises(datasource._MaybeCorruptedState):
        self.service._DatasourceService__save(1, 'state_token_1')
      self.service._DatasourceService__storage.save_updates_retval = True

      # Clear cached state.
      self.service._DatasourceService__state = None
      self.set_state({
        'token': 'state_token_1',
        'version': 2,
        'last_transaction': [
          ['b', 'v2'],
        ],
        'clients': {'1': None},
        'saving': True,
        'disk_version': 0,
        'saving_start': time.time()
      })
      self.service._DatasourceService__storage.save_updates_calls = []

      def concurrent_call():
        time.sleep(0.1)
        state = self.get_state()
        state['disk_version'] = 1
        state['saving'] = False
        self.set_state(state)

      thread1 = threading.Thread(target=lambda:
          self.service._DatasourceService__save(2, 'state_token_1'))
      thread2 = threading.Thread(target=concurrent_call)
      thread1.start()
      thread2.start()
      thread1.join()
      thread2.join()
      self.assertEquals(False,
          self.get_state()['saving'])
      self.assertEquals(2,
          self.get_state()['disk_version'])
      self.assertEquals([([
        ['b', 'v2'],
      ], 2)], self.service._DatasourceService__storage.save_updates_calls)

    finally:
      datasource._SAVE_POLL_DELAY = original_save_poll_delay


  def test_handle_updates(self):
    original_gettime = datasource._gettime
    datasource._gettime = lambda: 0.
    try:
      self.set_state({
        'token': 'state_token_1',
        'version': 1,
        'last_transaction': [
          ['a', 'v1'],
        ],
        'clients': {'1': 0., '2': 0.,
            '3': - datasource._CONNECTION_TIMEOUT - 1.},
        'saving': False,
        'disk_version': 1
      })
      channel_stub = self.testbed.get_stub('channel')
      token1 = channel.create_channel('1')
      channel_stub.connect_channel(token1)
      token2 = channel.create_channel('2')
      memcache.set(mcchannel._config.NAMESPACE + 'connected_' + '2',
          {'poll': False})
      channel_stub.connect_channel(token2)
      payload = {'updates': [
        'connection_token_1',
        'state_token_1',
        0,
        [
          [{}, [['b', 'v2']]],
          [{'a': {'.': None}}, [['c', 'v3']]],
          [{'b': {'.': None}}, [['d', 'v4']]],
          [{'c': {'.': None}}, [['e', 'v5']]],
        ]
      ]}

      with self.assertRaises(datasource._InsufficientPrivileges):
        self.service.get_access_mode = lambda: 'r'
        self.service._DatasourceService__handle_updates(payload)

      self.service.get_access_mode = lambda: 'rw'
      self.service._DatasourceService__handle_updates(payload)
      self.assertEquals({'1': {'service1': [
        json_utils.dumps({'outcomes': [
          'connection_token_1',
          'state_token_1',
          0,
          [2, None, 3, None]
        ]}),
        json_utils.dumps({'saved': [
          'state_token_1',
          3
        ]})
      ]}}, mcchannel._outgoing_messages)
      mcchannel._outgoing_messages = defaultdict(lambda: defaultdict(list))
      self.assertEquals([], channel_stub.get_channel_messages(token1))
      self.assertEquals([json_utils.dumps({'service1': [
        {'updates': [
          'state_token_1',
          3,
          [
            ['b', 'v2'],
            ['d', 'v4'],
          ]
        ]},
      ]})], channel_stub.get_channel_messages(token2))
      self.assertEquals([([
        ['b', 'v2'],
        ['d', 'v4'],
      ], 3)], self.service._DatasourceService__storage.save_updates_calls)
      self.assertEquals({'1': 0., '2': 0.}, self.get_state()['clients'])

    finally:
      datasource._gettime = original_gettime


  def test_large_update(self):
    original_LARGE_UPDATE_VALUES = datasource._LARGE_UPDATE_VALUES
    original_LARGE_UPDATE_TRANSACTIONS = datasource._LARGE_UPDATE_TRANSACTIONS
    try:
      datasource._LARGE_UPDATE_VALUES = 5
      datasource._LARGE_UPDATE_TRANSACTIONS = 3
      self.assertEquals(False, datasource._is_large_update([
        [{}, [['a', 1]]],
        [{}, [['a', 1], ['a', 1], ['a', 1]]],
      ]))
      self.assertEquals(True, datasource._is_large_update([
        [{}, [['a', 1]]],
        [{}, [['a', 1], ['a', 1], ['a', 1], ['a', 1]]],
      ]))
      self.assertEquals(True, datasource._is_large_update([
        [{}, [['a', 1]]],
        [{}, [['a', 1]]],
        [{}, [['a', 1]]],
      ]))

      datasource._LARGE_UPDATE_VALUES = 0
      self.set_state({
        'token': 'state_token_1',
        'version': 4,
        'last_transaction': [
          ['a', 'v1'],
        ],
        'clients': {'1': 0., '2': 0.},
        'saving': False,
        'disk_version': 2
      })
      memcache.set(self.memcache_prefix + 'update_2',
          json_utils.dumps(['b', 'v2']))
      memcache.set(self.memcache_prefix + 'update_3',
          json_utils.dumps(['b', 'v3']))
      channel_stub = self.testbed.get_stub('channel')
      token1 = channel.create_channel('1')
      channel_stub.connect_channel(token1)
      token2 = channel.create_channel('2')
      memcache.set(mcchannel._config.NAMESPACE + 'connected_' + '2',
          {'poll': False})
      channel_stub.connect_channel(token2)
      payload = {'updates': [
        'connection_token_1',
        'state_token_1',
        1,
        [
          [{}, [['a', 1]]],
          [{}, [['b', 2], ['c', 3]]],
        ]
      ]}

      self.service.get_access_mode = lambda: 'rw'
      self.service._DatasourceService__handle_updates(payload)
      self.assertEquals({'1': {'service1': [
        json_utils.dumps({'outcomes': [
          'connection_token_1',
          'state_token_1',
          1,
          [None, None],
        ]})
      ]}}, mcchannel._outgoing_messages)
      mcchannel._outgoing_messages = defaultdict(lambda: defaultdict(list))
      self.assertEquals(False, self.get_state()['saving'])
      self.assertTrue(
          memcache.get(self.memcache_prefix + 'update_2') is not None)
      self.assertTrue(
          memcache.get(self.memcache_prefix + 'update_3') is not None)

      payload = {'updates': [
        'connection_token_1',
        'state_token_1',
        2,
        [
          [{}, [['a', 1]]],
          [{}, [['b', 2], ['c', 3]]],
        ]
      ]}
      self.service._DatasourceService__handle_updates(payload)
      self.assertEquals({'1': {'service1': [
        json_utils.dumps({'outcomes': [
          'connection_token_1',
          'state_token_1',
          2,
          [3, 5]
        ]}),
        json_utils.dumps({'saved': [
          'state_token_1',
          5
        ]})
      ]}}, mcchannel._outgoing_messages)
      self.assertEquals([], channel_stub.get_channel_messages(token1))
      self.assertEquals([json_utils.dumps({'service1': [
        {'error': None}]})], channel_stub.get_channel_messages(token2))
      self.assertEquals([([
        ['a', 1],
        ['b', 2],
        ['c', 3],
      ], 5)], self.service._DatasourceService__storage.save_updates_calls)
      self.assertEquals({
        'token': 'state_token_1',
        'version': 5,
        'clients': {'1': 0., '2': 0.},
        'saving': False,
        'disk_version': 5,
      }, self.get_state())
      self.assertTrue(
          memcache.get(self.memcache_prefix + 'update_2') is not None)
      self.assertTrue(
          memcache.get(self.memcache_prefix + 'update_3') is None)

    finally:
      datasource._LARGE_UPDATE_VALUES = original_LARGE_UPDATE_VALUES
      datasource._LARGE_UPDATE_TRANSACTIONS = original_LARGE_UPDATE_TRANSACTIONS


  def test_handle_disconnect(self):
    self.set_state({'clients': {'1': None, '2': None}})
    self.service._DatasourceService__handle_disconnect()
    self.assertEquals({'clients': {'2': None}}, self.get_state())
    self.service._DatasourceService__handle_disconnect()
    self.assertEquals({'clients': {'2': None}}, self.get_state())
