# Channel #

Modules `client/mc/channel.js`, `client/mc/gaechannel.js`, `server/mc/channel.py`, and `server/mc/service.py` extend App Engine's channel API with the following functionality:

- The "forward" (client-to-server) part of the channel.

- A GAE channel closes when 2 hour limit is reached or if internet connection is lost, and cannot be reopened. The client-side module defines a channel that can re-open after being closed, and does so automatically if the GAE channel token expires while the user is still active.

- GAE Channel API imposes a limit on the size of messages that can be sent to the client, and when one has to send a larger message, this has to be done by instructing the client to pick up data from the server via a regular request. This flow is executed automatically when the message is large enough.

- The notion of services: all messages sent back and forth over the channel have a string attached to them that identifies a "service", so that multiple components of the application can share the channel by using different service names. You would always have at least two services: the system service that is used internally by this module (e.g. to instruct the client to pick up data from the server), and the service used by you.

- Support for polling, either as a fallback for when GAE Channel token cannot be created (e.g. because of 60-openings-a-minute quota) or for use cases when it's fine for server updates to arrive with say a minute delay.


## Client-side API ##

Channel is represented on the client by `mc.GaeChannel` class that implements `mc.Channel`, an abstract class that defines the following functionality:

- `send`: a method for sending messages.

- `registerService`, `unregisterService`: methods for registering/unregistering a handler of incoming messages sent through a specific service.

- `mc.Channel.EventType.OPEN`, `mc.Channel.EventType.CLOSE` events: dispatched when the channel is closed or opened.

- `isOpen`: a method for checking if the channel is currently open.

For backward (server-to-client) channel `mc.GaeChannel` uses App Engine's client-side API (which works cross-domain) or polling, and for the forward (client-to-server) channel it uses either a `goog.net.XhrIo` or `goog.net.Jsonp` depending on what you pass to `mc.GaeChannel` constructor. Unfortunately, the use of Jsonp is limited by the 2048 character costraint on the size of the URL in Appengine and IE 6/7. For backward channel, by default we use GAE Channel with fallback on polling, but polling can be forced using an optional argument to `mc.GaeChannel` constructor.

Internally the messages can be delivered from server to client in two ways: by sending a GAE Channel message or by sending the data in the response to a forward channel request. We use the last method if on a given server, we are currently processing a forward channel request from the client to whom the message is sent (with polling, that's the only method available).

Even when on a given client `mc.GaeChannel` is closed, messages can be sent over the forward channel, and on a server that is currently processing the client's forward channel request, over the backward channel. But to always be able to send messages over backward channel, `mc.GaeChannel` needs to be opened by calling its `open` method, which will either open the GAE Channel or initiate polling. A little later after you call that method, one of two things will happen: either the channel will open, dispatching the `OPEN` event, or if the channel could not be opened, e.g. because there is no internet connection, `mc.GaeChannel.EventType.OPEN_FAILURE` event will be dispatched.

The channel will stay open until one of the following happens:

- GAE Channel is used for backward channel and the GAE API throws an error. This will happen when internet connection is lost or when the 2 hour channel lifespan ends.

- Polling is used for backward channel and maximum number of retries of the forward channel request is reached. This usually indicates that internet connection has been lost.

- Polling is used for backward channel and 2 hours elapse since the channel was opened.

When the channel closes, we dispatch a `CLOSE` event and, except in the cases described below, dispatch `mc.GaeChannel.EventType.AUTO_OPEN_START` and automatically attempt to re-open the channel.

The three cases when will not automatically attempt to re-open it are as follows:

- The user is detected to be idle by a `goog.ui.ActivityMonitor` (you can either pass an instance of this monitor to the `mc.GaeChannel` constructor or a default instance will be created). This allows to avoid opening a GAE channel every two hours during long periods of inactivity. The channel will be re-opened automatically once the user becomes active.

- A safeguard against too frequent channel openings kicks in. This safeguard prevents a situation when we obtain a channel token, then try to open the channel, then immediately the channel closes due to some error, and then we try to open it again, repeating the cycle. This can rapidly consume App Engine's quota on token requests. To prevent this, after an automatic reopening of the channel, we put a lock on further automatic reopenings for a certain period of time. If the channel closes before the lock expires, instead of trying to reopen it we will dispatch `OPEN_FAILURE` event. The lock does not prevent you from opening the channel manually by calling `open`.

- The channel was closed because polling is used for backward channel and maximum number of retries of the forward channel request has been reached.

When an `OPEN_FAILURE` event is dispatched, the channel will stay closed until you call `open` again. This [demo app](http://mightyclient.appspot.com/) displays 'Reconnect' button when `OPEN_FAILURE` is dispatched (which happens a minute or so after the internet connection is lost), and if the user clicks this button, we attempt to reconnect by calling `open`.

When you send messages to the server, the actual HTTP request is postponed until the next batch of browser event processing, so messages sent in the same batch will be chunked together. There is one HTTP request active at a time: messages sent while we're waiting for the request to complete are buffered.

Under the hood `mc.GaeChannel` dog-foods its own functionality by using a special service whose name is `'system'` to exchange system messages with the server, such as getting GAE Channel token.


## Server-side API ##

You need to do the following on the server to set up the channel.

First, make sure python version is 2.7, and webapp2 library is enabled:

    libraries:
    - name: webapp2
      version: "2.5.1"

Second, you need to set up `channel.ForwardChannelHandler` to handle requests to a URL which is stored in the constant `mc.GaeChannel.BASE_URL_PATH` in the client code, '/mc_channel' by default.

Finally, you need to define one or more subclasses of `service.Service` class in `appengine_config.py`, the standard App Engine module configuration file. For full details, see `channel._ConfigDefaults` class in `channel.py`. Here is a sample `appengine_config.py`:

    from mc.service import Service


    class EchoService(Service):
      def handle_message(self, payload):
        # Echoes the message to the client who sent it.
        self.send(payload)


    class HelloService(Service):
      def handle_message(self, payload):
        # When one client sends another's ID,
        # says hello to that other client.
        self.send('Hello!', to=payload)


    mc_channel_SERVICES = [
      ('echo', EchoService),
      ('hello', HelloService),
    ]

Much like it happens with HTTP request handlers, when a message is received from a client, we instantiate the corresponding `Service` (with service name being analogous here to URL - we choose the first service in the services list whose associated regular expression matches the service name) and call its `handle_message` method like we would call `get` method of a request handler. The `Service` instance has properties `client_id` and `name` that allow handlers to access the ID of the client and the service name.

Messages can be sent to a client by calling the `send` method of a `Service` instance or by calling the function `channel.send`. These two ways are equivalent, except that the instance method provides default values for arguments `client_id` and `service`. A message can be any JSON-encodable payload, or alternatively you can provide the message as an already JSON-encoded string so that the channel module would not have to encode it, by wrapping the string into an instance of `service.SerializedPayload`.

There are a few reasons why some message may not be delivered or may be delivered multiple times: something internal to Channel API, eviction of items from memcache, early termination of request processing. So although normally messages are delivered correctly, one has to keep this limitation in mind.
