# Datasource #

`client/mc/datasource.js` defines an API to read, write, and listen for changes in a collection of JSON-serializable values where each value is located at a path composed of string keys. It also exposes hooks for data synchronization. Optionally, a schema can be defined that will be enforced by run-time assertions in debug mode and stripped by compiler in production.


## Data access ##

The data is structured as a tree whose leafs are any JSON-encodable values: either a primitive (number, boolean or string) or a more complex structure. Each such value is stored at a path composed of string keys (a key can contain letters, numbers and underscores). In some contexts a special `'*'` wildcard key can be used, signifying "any key". At the root of the tree there is an `mc.ds.RootNode` instance.

We also use a concept of a data "node" (class `mc.ds.Node`, of which `mc.ds.RootNode` is a subclass), which is simply a combination of a reference to a root node and a fully qualified path from it. A node does not need to point to any existing data and disposing it will not affect the data. Node instances are created by instantiating either `mc.ds.Node` or a caller-defined subclass. Once a node has been instantiated, it is accessible by calling `getNode` method of an ancestor node. Nodes are used to conveniently reference data model objects associated with data tree paths and they also slightly optimize performance: when reading a node's descendant values and nodes, the path from root to the node does not have to be traversed.

Data is read using the node's methods `get`, `some`, `every`, and `forEach` which do what you would expect. To set a primitive value, call `set` method. You delete a value by setting it to `null`.

When a primitive value changes, an `mc.ds.ChangeEvent` event is dispatched (will discuss how to listen to these events in a moment). In addition to the path, old value and new value, the event object will have a property `source`, which is the optional `opt_source` argument passed to `set` method or null if this argument was not provided. The `source` property is useful for objects such as user interface components that both listen to changes in a value and can modify the value. By providing a reference to itself as source when setting a value, and inspecting the source property when handling an event, the object can distinguish the events that it has itself triggered.

To listen to events (either `mc.ds.ChangeEvent` or custom events discussed in the corresponding section below), call a node's `listen` method which returns an object that should be disposed to stop listening. It is convenient to attach those objects to disposables with `registerDisposable` method.

By default, `get`, `set`, `some`, `every`, `forEach` and accessor methods for old and new values attached to an instance of `mc.ds.ChangeEvent` perform deep cloning of values that makes sure that mutable values stored internally would not be accidentally modified by the caller: a value that is being read is deep-cloned before passing it to the caller, and a value that is being set is deep-cloned before storing it. You can avoid the overhead of cloning by passing `true` as an optional argument `opt_doNotClone` to any of the above-mentioned methods, but you'll need to make sure not to modify the mutable values.


## Data synchronization and transactions ##

`client/mc/datasource.js` does not contain any data synchronization logic, but it does expose the hooks that allow such logic to be implemented externally. Thus the process that performs data synchronization accesses the data tree on equal rights with any other application logic. You can have a data tree which is not being synchronized with anything, and this is convenient in unit tests: you can play around with data models and UI components that are connected to a data tree without connecting this data tree to a server.

The root node has a method `inTransaction` that allows to declare a restriction on how the data synchronization process can modify data on this and other sites. This restriction is formulated in terms of what changes can be applied to the data so that if the state was originally consistent, it would remain consistent, with syncing process then having to make sure the state is always consistent on all sites. Running a callback in transaction by passing it to `inTransaction` method indicates that if a sequence of writes (calls to `set` method of `mc.ds.Node`) made by the callback is applied over such consistent state of the data that the reads (calls to `get`, `some`, `every` and `forEach` methods of `mc.ds.Node`) made by the callback would produce the same results, the state of the data will also be consistent after each write is applied. In contrast, if you set a value outside of a transaction, this implies that this write will not break consistency without any restrictions on prior writes or results of reads.

To consider an example, suppose you have a data structure with strings at paths `['1']`, `['2']`, and so on:

    '1'
        <string>
    '2'
        <string>
    ...

The data is considered consistent only if all the strings are different. Suppose the user can modify the strings. We can put in a check that will only allow the user to save a string if it isn't equal to some other string - the code would look something like this, supposing that we're setting string at path `[id]` to `newValue`, and the root node is `rootNode`:

    if (rootNode.some(['*'], function(value, path) {
      return path[0] != id && value == newValue;
    })) {
      alert('The string is the same as some other string.');
    } else {
      rootNode.set([id], value);
    }

The problem is that even if we do this check on all clients, we can still end up with inconsistent data, because two clients can at the same time set identical strings, one say at path `['1']` and the other at path `['2']`, and these two changes will be merged.

The way to prevent it is simply to enclose the above code in `rootNode.inTransaction(function() {...})`. This will signify to the synchronization process that the write that we made can only be applied on any other site if on that site the read that we made (the call to `some` method) would produce the same results as it actually has produced when we ran our transaction, in other words, if there are no conflicting strings on the other site as well.

In the case of `some`, `every` and `forEach`, when we say that the read produces the same results, this simply means that all values matching the wildcard path supplied as argument are the same, and naturally, in the case of `get` this means that the value at provided path is the same.

Sometimes it's necessary to register a read but there's no need to actually read the data - in this case you can use `read` method of a node. The same method is called internally by methods like `get`.

Note that unlike normal database transactions, the state should be consistent after *each* write, not just in the end of the transaction. This is related to the way events work: after each write a ChangeEvent is dispatched, which can trigger event callbacks relying on consistent state of the data.

A common use case is storing a data model object that has multiple datasource values associated with it. But if the state has to be consistent after each individual value is set, how do you initialize such object? The answer is that data model should be designed in such a way that setting a single value would be enough to initialize an object - for other required values there should be methods (it is convenient to define them on an `mc.ds.Node` subclass) that return some default value if none is present in the datasource. To ensure that objects are deleted cleanly, you would delete an object by applying a transaction that nulls all values associated with it and as a last step, nulls the value that initialized it. When setting one of the associated values during the object's life, you have to do this in a transaction that starts with checking for the object's existence.

Calls to `get`, `some`, `every` and `forEach` that are made by handlers of change events are not registered as reads made by the transaction.

Be careful when using `goog.asserts` in conjunction with reads: if you write something like `goog.asserts.assert(rootNode.get(['a']))` to indicate that there is a truthy value at path `['a']`, this code will be stripped by the compiler because it would assume that `rootNode.get(['a'])` has no side effects. As a result, no read would be included into the transaction. One solution in this case is simply to write `var a = rootNode.get(['a']); goog.asserts.assert(a)`: the assertion will be stripped, but the call to `get` won't be.

This [demo](http://mightyclient.appspot.com/) gives another example of how transactions can be used.


## Custom Events ##

The same mechanism that is used internally for dispatching `mc.ds.ChangeEvent` events is also available for dispatching any custom events whose scope is defined by a data tree path though the `RootNode`'s method `dispatchEvent`.

Normally you would dispatch custom events for paths which never point to primitive values (when schema is enabled, this is enforced by assertions). Suppose your data looks like this:

    '1'
        'a'
            <number>
        'b'
            <number>
    '2'
        'a'
            <number>
        'b'
            <number>
    '3'
      ...

and in your data model you dynamically calculate the sum of numbers at paths `<ID>/a` and `<ID>/b` for each `ID`. If you need to track changes in that dynamic value, what you would do is listen to paths `*/a` and `*/b`, and when for example the value at `2/a` changes, dispatch a custom event for path `2/c`.

Note also that `dispatchEvent` accepts wildcard paths. Returning to the previous example, suppose your dynamically calculated value depends not just on values at `<ID>/a` and `<ID>/b`, but suppose the sum also includes a third component which is the same for all IDs. If that component changes, you would dispatch custom event for path `*/c` which will be caught by listeners to paths such as `1/c`, `2/c`, or `*/c`.

If you have a very large number of listeners, be aware that performance of the event dispatcher may be affected by the number of non-matching listeners.


## Schema ##

For the purposes of documentation and testing, the client code can optionally define a schema. The schema will be enforced by run-time assertions, but when the code is compiled with `mc.ds.ENABLE_SCHEMA` set to `false`, both schema definitions and run-time checks will be stripped away by the compiler. `mc.ds.ENABLE_SCHEMA` is set to `goog.asserts.ENABLE_ASSERTS` by default, which in turn is set to `goog.DEBUG` by default.

In the schema the following primitive elements are supported:

- `'number'`: Values that are numbers other than NaN and Infinity.

- `'boolean'`: Values that are booleans.

- `'string'`: Values that are strings.

- `'*'`: Values that are anything that can be serialized to JSON.

- `'EventPath'`: a special element which marks paths at which no value can be stored but which (are the only paths that) can be used to dispatch custom events.

In addition to these primitive value types, you can define your own types using function `mc.ds.defineType`. For instance, the following:

    mc.ds.defineType('MyType', {'a': {'b': 'number'}});

defines a custom type matched by data structures such as

    'a'
        'b'
            1

as well as an empty data structure, but not by this structure:

    'a'
        'B'
            1

or this:

    'a'
        'b'
            true

You can use one type in the definition of another type:

    mc.ds.defineType('MyOtherType', {'c': 'MyType'});

which will be matched by

    'c'
        'a'
            'b'
                1

Type definition can be recursive and calls to `mc.ds.defineType` can be made in any order. All type names live in a single namespace. Schema definition can contain wildcard keys (`'*'`), but if some object in the schema has such key, it should have no other keys.

To enable the schema checks, after instantiating the root node you need to set its type, which is done using function `mc.ds.setType`:

    var rootNode = new mc.ds.RootNode({});
    mc.ds.setType(rootNode, 'MyOtherType');

Then, if you call something like `rootNode.get(['c', 'a', 'B'])`, you will get an assertion error `Invalid path 'c/a/B': unexpected path element 'B' following 'c/a'.`


## Hooks for synchronization ##

Datasource has two features that enable synchronization of data:

- Events dispatched by associated EventTarget, which is accessible with `getEventTarget` method. These events indicate when a transaction starts and ends, and when when a read or a write is made. See `mc.ds.EventType` enum for details.

- A restriction, enforced by assertions, that data cannot be modified by callbacks to change events.
