# Client-server communication protocol: datasource #

## Tokens ##

There are two random strings that are used in client-server communication:

- Connection token: this token is generated on the client before sending a request to connect and is reset before sending a request to reconnect.

- In-memory state token: this token is generated on the server when the in-memory state is initialized and remains unchanged for as long as the in-memory state is not reset.


## Messages sent from client to server ##

### Request to connect ###

Request to connect to the datasource from a client that does not have any previous state of the data.

Format:

    {'connect': <connection token>}


### Request to reconnect ###

Request to connect to the datasource from a client that already has some previous state of the data.

Format:

    {'reconnect': [
      <the version starting from which the server is requested to
          provide updates>,
      <connection token>
    ]}


### Updates ###

Updates to be applied to the data.

Format:

    {'updates': [
      <connection token>,
      <in-memory state token if the base version (the following item) is higher
          than the version that the client knows to be committed to disk, null
          otherwise>,
      <version over which changes were made>,
      [
        // Transaction 1
        [
          // Reads
          <an object tree where at each path that has been read there is
              an object with key '.' pointing to value of null>,
          // Writes
          [
            // Update 1
            [<'/'-joined path>, <new value>],
            // Update 2
            [...],
            ...
          ]
        ],
        // Transaction 2
        [...],
        ...
      ]
    ]}


### Request to disconnect ###

Request to disconnect this client from the datasource.

Format: `{'disconnect': null}`.


## Messages sent from server to client ##

### Response to request to connect ###

In response to a request to connect, the server sends a message in the following format:

    {'connected': [
      <connection token>,
      <in-memory state token>,
      <version of the data on disk>,
      <data tree as stored on disk>
    ]}

`<data tree as stored on disk>` is a data structure containing the state of the data as stored on disk. On the server, it is the deserialized version of the `<state>` returned by `get_disk_state` method of `abstract_storage.AbstractStorage`. On the client, it is the value passed as argument to protected method `createRootNode` of `mc.ServerAgent`, and the default implementation of this method expects the value to be a tree of nested objects where at each path plus '.' key there is the corresponding value. If required, one can override `createRootNode` on the client to accept a different representation of the state.

If there are updates between the version of the data on disk and the in-memory version, the server will also send a separate message with those updates in the format described in section "Server updates".


### Response to request to reconnect ###

In response to a request to reconnect, the server sends a message in the following format:

    {'reconnected': [
      <connection token>,
      <in-memory state token>
    ]}

If there are updates between the version in the 'reconnect' message and the in-memory version, the server will also send a separate message with those updates in the format described in section "Server updates".

Additionally, the server will always send a 'Saved' nofication described in the corresponding section below.


### Response to client updates ###

After the client sends updates to the data, the server responds with information on which of the updates have been successfully applied in-memory.

Format:

    {'outcomes': [
      <connection token>,
      <token of the in-memory state to which updates have been applied, may or
          may not be the same as the token in the client's message>,
      <version in the update message sent by the client>,
      [<outcome for transaction 1>, <outcome for transaction 2>, ...]
    ]}

where the array of outcomes corresponds to the array of transactions sent by the client, and each outcome is either the version of the data after the transaction was applied, if it was applied successfully, or null otherwise.


### Server updates ###

Updates to the data not originating from this client.

Format:

    {'updates': [
      <in-memory state token>,
      <version of the data after the updates>,
      [
        // Update 1
        [<'/'-joined path>, <new value>],
        // Update 2
        ...
      ]
    ]}

Updates are consecutive, so each of the updates increments the version by 1.


### 'Saved' notification ###

Notification that all updates up to a given version have been saved to disk.

Format: `{'saved': [<in-memory state token>, <version>]}`.


### Error notification ###

This notifies the client that one of the following two things has happended:

- The server is not able to complete an operation that the client initiated, specifically the server cannot connect the client to the datasource or cannot apply the updates that the client sent. In this case the client will never receive the normal response message from the server, and for UX purposes it is useful for the client to be notified of this fact immediately.

- Because of a server error, such as perishing of the in-memory state, the client will no longer receive notifications of updates made by other clients.

When the client receives this notification, it should dispose its datasource state and attempt to reconnect.

Format: `{'error': null}`.
